/** @file dsapu.cpp
 *  @author Mic
 *  @date 2007
 *
 *  @brief Audio processing unit emulation for Dualis.
 */

#include <windows.h>
#include <commctrl.h>
#include <string.h>
#include <bass.h>
#include <mmsystem.h>
#include <mmreg.h>
#include <stdio.h>
#include "dsapu.h"
#include "zombie.h"


#define APU_VERSION 24


short waveforms[8][8] = {				/**< Rectangle wave dutycycle table */ 
	{LOW,LOW,LOW,LOW,LOW,LOW,LOW,HIGH},			// 12.5%
	{LOW,LOW,LOW,LOW,LOW,LOW,HIGH,HIGH},		// 25.0%
	{LOW,LOW,LOW,LOW,LOW,HIGH,HIGH,HIGH},		// 37.5%
	{LOW,LOW,LOW,LOW,HIGH,HIGH,HIGH,HIGH},		// 50.0%
	{LOW,LOW,LOW,HIGH,HIGH,HIGH,HIGH,HIGH},
	{LOW,LOW,HIGH,HIGH,HIGH,HIGH,HIGH,HIGH},
	{LOW,HIGH,HIGH,HIGH,HIGH,HIGH,HIGH,HIGH},
	{LOW,LOW,LOW,LOW,LOW,LOW,LOW,LOW}			//  0.0%
};


int adpcmIndexTable[8] = {-1,-1,-1,-1,2,4,6,8};
short adpcmDiffTable [89] = {
  0x0007,0x0008,0x0009,0x000A,0x000B,0x000C,0x000D,0x000E,0x0010,0x0011,0x0013,0x0015,
  0x0017,0x0019,0x001C,0x001F,0x0022,0x0025,0x0029,0x002D,0x0032,0x0037,0x003C,0x0042,
  0x0049,0x0050,0x0058,0x0061,0x006B,0x0076,0x0082,0x008F,0x009D,0x00AD,0x00BE,0x00D1,
  0x00E6,0x00FD,0x0117,0x0133,0x0151,0x0173,0x0198,0x01C1,0x01EE,0x0220,0x0256,0x0292,
  0x02D4,0x031C,0x036C,0x03C3,0x0424,0x048E,0x0502,0x0583,0x0610,0x06AB,0x0756,0x0812,
  0x08E0,0x09C3,0x0ABD,0x0BD0,0x0CFF,0x0E4C,0x0FBA,0x114C,0x1307,0x14EE,0x1706,0x1954,
  0x1BDC,0x1EA5,0x21B6,0x2515,0x28CA,0x2CDF,0x315B,0x364B,0x3BB9,0x41B2,0x4844,0x4F7E,
  0x5771,0x602F,0x69CE,0x7462,0x7FFF
};


HINSTANCE		hInst;
HWAVEOUT		hWav;
HSYNC			hSync;
HANDLE			hout;
WAVEFORMATEX	waveFormat;
WAVEHDR			waveHdr[2];
DWORD			oldVolume,dummy;
HSTREAM			hStream;
unsigned char	*IOREGS,			/**< Pointer to the ARM9 I/O registers */
                *IOREGS_ARM7,		/**< Pointer to the ARM7 I/O registers */
				*VRAM;				/**< Pointer to VRAM */
unsigned int	*cycle = NULL,
                cycles;
unsigned int	cpu_snd;
short			*stream[2];
zombie_callback zombie;
char			apuName[] = "DS APU";
char			waveOutInitError[] = "Failed to initialize vaveOut device";
char			memoryError[] = "Unable to allocate sufficient memory";
char			*lpszError;
char			szbuffer[256];
bool			audio_done;
int				buffer_played;
int				sampleRate = APU_FREQ;
tag_pcm			pcm;
tag_sndchan		channels[16];
unsigned char	**arm7_pageTable;
unsigned int	*arm7_pageMask;
bool			terminate;
bool			useSmallBuffer=false;
short			sndbias;
int				frameRate;
short			psgbuffer[32768][2];
UINT			psgtimestamp,
                psgbufferstart,
				psgbufferpos,
				psgbufferlen,
				psgdata;
UINT stream1pos;


bool __stdcall DllMain(HINSTANCE hDLL,DWORD reason,LPVOID unused)
{
	if (reason==DLL_PROCESS_ATTACH)
		hInst = hDLL;

	return true;
}



///////////////////////////////////////////////////////////////////////////////////////


/**
 *  @brief Mixes up a new batch of samples when BASS has finished the previous one.
 */
static DWORD CALLBACK stream_proc(HSTREAM handle, void *buffer, DWORD length, DWORD user) {
	if (stream1pos)
	{
		if (stream1pos > (length>>APU_CHN))
		{
				wsprintf(szbuffer, "s1 %d, l %d\n", stream1pos<<2, length);
	WriteConsole(hout,
		         szbuffer,
				 lstrlen(szbuffer),
				 (unsigned long*)&dummy,
				 NULL);
			memcpy(buffer,stream[1],length);
			memcpy(&stream[1][0],&stream[1][length>>1],((stream1pos - (length>>APU_CHN))<<APU_CHN));
			stream1pos = 0;
			return length;
		}
		else
		{
			memcpy(buffer,stream[1],stream1pos<<APU_CHN);
			pcm.len = (length>>(APU_CHN)) - stream1pos;
			stream[0] = (short*)buffer;
			stream[0] = &stream[0][stream1pos<<1];
			buffer_played = 0;
			stream1pos = 0;
			apu_mix();
			return length;
		}
	}

	pcm.len = length>>(APU_CHN);
	stream[0] = (short*)buffer;
	buffer_played = 0;
	apu_mix();

	return pcm.len << APU_CHN; 
}


static void CALLBACK sync_proc(HSYNC handle, DWORD channel, DWORD data, DWORD user) {
	wsprintf(szbuffer, "Sync occured on channel %d: %d\n", channel, data);
	WriteConsole(hout,
		         szbuffer,
				 lstrlen(szbuffer),
				 (unsigned long*)&dummy,
				 NULL);
}


DWORD buflen;

/**
 *  @brief Initializes APU emulation.
 *
 *  @param commlink Callback to the main message hub
 *
 *  @return true on success, otherwise false
 */
bool __stdcall apu_init(zombie_callback commlink) {
	int i,j;

	zombie = commlink;

	audio_done = false;

	frameRate = 60;

	if (BASS_Init(1,
		          APU_FREQ,
				  BASS_DEVICE_LATENCY | (((APU_CHN == 1) ? BASS_DEVICE_MONO : 0)),
				  0,
				  0) == FALSE)
		return false;

	if (useSmallBuffer) {
		buflen = BASS_GetConfig(BASS_CONFIG_UPDATEPERIOD); 
		BASS_INFO info;
		BASS_GetInfo(&info);						// Retrieve device info
		buflen = info.minbuf*4;						// Add the 'minbuf'
		if (buflen < 200) buflen = 200;				// Don't go lower than 200 ms
		BASS_SetConfig(BASS_CONFIG_BUFFER,buflen);	// Set the length

			wsprintf(szbuffer, "Using audio buffer length %d\n", buflen);
	WriteConsole(hout,
		         szbuffer,
				 lstrlen(szbuffer),
				 (unsigned long*)&dummy,
				 NULL);
	}

	stream[1] = (short*)LocalAlloc(LPTR, APU_FREQ*APU_CHN*2*2);

	if ((hStream = BASS_StreamCreate(sampleRate,
		                             APU_CHN,
									 0,
									 (STREAMPROC*)stream_proc,
									 NULL)) == 0)
		return false;

	hSync = BASS_ChannelSetSync(hStream,BASS_SYNC_STALL,0, (SYNCPROC*)sync_proc,0);
 
	hout = GetStdHandle(STD_OUTPUT_HANDLE);
	return true;

}



/**
 *  @brief Resets APU emulation
 *
 *  @return true on success, otherwise false
 */
bool __stdcall apu_reset() {
	int i;

	// Ask the GPU for a pointer to VRAM
	VRAM = (unsigned char*)zombie(ASK, GPU_MEMORY, (void*)0L, (void*)0L);
	if ((int)VRAM == 0L)
		return false;

	// Dualis keeps the I/O registers in the same chunk of
	// memory as VRAM.
	IOREGS		= &VRAM[IOREGS_OFFSET];
	IOREGS_ARM7 = &VRAM[IOREGS_ARM7_OFFSET];

	IOREGS_ARM7[0x501] = 0;
	
	// Reset all channels
	for (i=0; i<16; i++) {
		channels[i].pos = channels[i].end = 0;
		channels[i].enabled = false;
		channels[i].delta = 0;
		channels[i].timestamp = 0;

		*(unsigned short*)(IOREGS_ARM7 + 0x40A + (i<<4)) = 0;
	}

	terminate      = false;
	buffer_played  = 0;
	sndbias        = 0;
	psgbufferpos   = 0;
	psgbufferlen   = 0;
	psgbufferstart = 0;
	psgdata        = 0;

	stream1pos = 0;

	wsprintf(szbuffer, "Setting buffer length %d ms\n", buflen);
	WriteConsole(hout,
		         szbuffer,
				 lstrlen(szbuffer),
				 (unsigned long*)&dummy,
				 NULL);

	wsprintf(szbuffer, "Mixing at %d Hz\n", sampleRate);
	WriteConsole(hout,
		         szbuffer,
				 lstrlen(szbuffer),
				 (unsigned long*)&dummy,
				 NULL);

	BASS_ChannelStop(hStream);



	BASS_ChannelPlay(hStream,TRUE);

	return true;
}


/**
 *  @brief Shuts down APU emulation.
 *
 *  @return
 */
void __stdcall apu_close() {
	terminate = true;

	BASS_ChannelStop(hStream);
	BASS_StreamFree(hStream);
	BASS_Free();

	if (stream[1])
		LocalFree((HLOCAL)stream[1]);
}



int __stdcall apu_config(HWND hwnd) {
	return 0; 
}


/**
 *  @brief Handles mixing of PSG channels
 *
 *  @param forceSamples Force a specific number of samples to be mixed
 *
 *  @return
 */
static void psg_mix(UINT forcesamples) {
	int i,n;
	UINT timediff,nsamples,mixtime;
	int sampleL,sampleR;


	if (psgdata == 0) return;
	if ((IOREGS_ARM7[0x501]&0x80)==0)	// Check sound master enable bit
		return;

	mixtime = *cycle;
	
	if (forcesamples != 0) {
		nsamples = forcesamples;
		mixtime += nsamples * (ARM7_FREQ / sampleRate);

	} else {
		if (mixtime >= psgtimestamp)
			timediff = mixtime - psgtimestamp;
		else if ((psgtimestamp >= 0x20000000) && (mixtime < 0x20000000))
			timediff = mixtime + 0x20000000 - psgtimestamp;
		else
			timediff = 0;
		nsamples = (sampleRate * (timediff>>10)) >> 15;
	}

	if (nsamples>0)
		psgtimestamp = mixtime;

	for (n=0; n<nsamples; n++) {
		sampleL = sampleR = 0;
		for (i=8; i<16; i++) {
			if (channels[i].enabled && (channels[i].fmt == FMT_PSG)) {
				// Rectangle waves
				if (i<14) {
					sampleL += (int)(waveforms[channels[i].duty][(channels[i].pos>>20)&7] * channels[i].left)>>14;
					sampleR += (int)(waveforms[channels[i].duty][(channels[i].pos>>20)&7] * channels[i].right)>>14;

				// Noise	
				} else {
					if (((channels[i].pos+channels[i].delta3)&0xFFF00000) != (channels[i].pos&0xFFF00000)) {
						if (channels[i].tap&1) {
							channels[i].noise = LOW;
							channels[i].tap = (channels[i].tap>>1)^0x6000;
						} else {
							channels[i].noise = HIGH;
							channels[i].tap >>= 1;
						}
					}
					sampleL += (int)(channels[i].noise * channels[i].left)>>14;
					sampleR += (int)(channels[i].noise * channels[i].right)>>14;
				}
				channels[i].pos += channels[i].delta3;
			}
		}
		psgbuffer[psgbufferpos][0] = sampleL;
		psgbuffer[psgbufferpos][1] = sampleR;
		psgbufferpos++;
		psgbufferpos &= 32767;
	}
	psgbufferlen += nsamples;
}


static int clamp(int val, int min, int max) {
	if (val < min) return min;
	else if (val > max) return max;
	return val;
}


/**
 *  @brief Mixes audio channels and writes the samples to a buffer.
 *
 *  @return
 */
void apu_mix() {
	int          i,j,k,l,r;
	int          sampleL,sampleR;
	unsigned int pos,posfrac,addr,addr2;
	short int    smp,smp2,diff;
	char         smp3,smp4;

	if (terminate) return;

	if ((IOREGS_ARM7[0x501] & 0x80) == 0) {		// Check sound master enable bit 
		// Clear the buffer
		memset(stream[buffer_played], 0, pcm.len*2);
		return;

	} else {
		if (psgdata != 0)
			psg_mix(0);

		if ((psgbufferlen < pcm.len) && (psgdata != 0)) {
			pcm.len = psgbufferlen;
		}


		for (i=0; i<pcm.len; i++) {
			sampleL = sampleR = k = 0;
			for (j=0; j<16; j++) {
				if (channels[j].enabled) {
					l = channels[j].left; r = channels[j].right;

					// PCM8 
					if (channels[j].fmt == FMT_PCM8) {
						addr = channels[j].sad + (channels[j].pos >> 9);
						addr2 = addr;
						if (((channels[j].pos + (channels[j].delta&0xE00)+0x200)>>9) < channels[j].end) {
							addr2 += (channels[j].delta>>9)+1; 
						}
						if (arm7_pageTable[addr>>20]) {
							posfrac = channels[j].pos & 0x1FF;
							smp3 = *(char*)(arm7_pageTable[addr>>20]  + (addr  & arm7_pageMask[addr>>20]));
							smp4 = *(char*)(arm7_pageTable[addr2>>20] + (addr2 & arm7_pageMask[addr2>>20]));
							__asm {
								movsx ebx,byte ptr smp3
								movsx ecx,byte ptr smp4
								mov eax,posfrac
								mov edi,eax
								xor eax,0x1FF
								imul ebx			// eax = smp3 * (posfrac^0x1ff)
								imul ecx,edi		// ecx = smp4 * posfrac
								add eax,ecx
								sar eax,9
								mov ebx,eax
								imul dword ptr l
								imul ebx,dword ptr r
								sar eax,6
								sar ebx,6
								add sampleL,eax
								add sampleR,ebx
							}
						}

					// PCM16 
					} else if (channels[j].fmt == FMT_PCM16) {
						addr = channels[j].sad + ((channels[j].pos >> 9)<<1);
						addr2 = addr;
						if (((channels[j].pos + (channels[j].delta&0xE00)+0x200)>>9) < channels[j].end) {
							addr2 += ((channels[j].delta>>9)+1)<<1; 
						}

						if (arm7_pageTable[addr>>20]) {
							posfrac = channels[j].pos & 0x1FF;
							smp  = *(short*)(arm7_pageTable[addr>>20]  + (addr  & arm7_pageMask[addr>>20]));
							smp2 = *(short*)(arm7_pageTable[addr2>>20] + (addr2 & arm7_pageMask[addr2>>20]));
							__asm {
								movsx ebx,word ptr smp
								movsx ecx,word ptr smp2
								mov eax,posfrac
								mov edi,eax
								xor eax,0x1FF
								imul ebx			// eax = smp  * (posfrac^0x1ff)
								imul ecx,edi		// ecx = smp2 * posfrac
								add eax,ecx
								sar eax,9
								mov ebx,eax
								imul dword ptr l
								imul ebx,dword ptr r
								sar eax,14
								sar ebx,14
								add sampleL,eax
								add sampleR,ebx
							}

						/*smp = *(short*)(arm7_pageTable[addr>>20] + (addr & arm7_pageMask[addr>>20]));
							sampleL += (int)(smp*channels[j].left)>>14;
							sampleR += (int)(smp*channels[j].right)>>14;*/
						}


					} else if (channels[j].fmt == FMT_ADPCM) {
						if (channels[j].pos == 0) {
							addr = channels[j].sad;
							channels[j].adpcmSmp = *(short*)(arm7_pageTable[addr>>20]  + (addr  & arm7_pageMask[addr>>20]));
							channels[j].adpcmStartSmp = channels[j].adpcmSmp;
							addr += 2;
							channels[j].adpcmIndex = *(char*)(arm7_pageTable[addr>>20]  + (addr  & arm7_pageMask[addr>>20]));
							channels[j].adpcmIndex &= 0x7F;
							channels[j].adpcmStartIndex = channels[j].adpcmIndex;
							channels[j].pos += 8<<9;
						}
						for (pos = channels[j].pos; pos <= (channels[j].pos+channels[j].delta); pos += 0x200)
						{
						addr = channels[j].sad + (pos>>10); //(channels[j].pos >> 10);
						smp3 = *(char*)(arm7_pageTable[addr>>20]  + (addr  & arm7_pageMask[addr>>20]));
						if (pos&0x200)  //(channels[j].pos & 0x200) //1)
							smp3 = (smp3>>4)&0xF;
						else
							smp3 &= 0xF;
						diff = ((smp3 & 7)*2+1) * adpcmDiffTable[channels[j].adpcmIndex] / 8;
						if (smp3 & 8) diff = -diff;
						channels[j].adpcmSmp = clamp(channels[j].adpcmSmp+diff,-0x8000,0x7FFF);
						channels[j].adpcmIndex = clamp(channels[j].adpcmIndex+adpcmIndexTable[smp3&7],0,88);
						}
						sampleL += (channels[j].adpcmSmp*l)/16384;
						sampleR += (channels[j].adpcmSmp*r)/16384;
	

					} else continue;

					channels[j].pos += channels[j].delta;
					if ((channels[j].pos>>9) >= channels[j].end) {
						if (channels[j].looptype == 1) {
							channels[j].pos = channels[j].looppoint + (((channels[j].pos>>9) - channels[j].end)<<9);
							channels[j].end = (channels[j].looppoint>>9) + channels[j].nsamples;

						} else { //else if (channels[j].looptype == 2) {
							channels[j].pos = 0;
							channels[j].enabled = false;
							IOREGS_ARM7[0x403 + (j<<4)] &= 0x7F;
						}
					}
				}
			}

			if (psgbufferlen>0) {
				sampleL += (int)psgbuffer[psgbufferstart][0];
				sampleR += (int)psgbuffer[psgbufferstart][1];
				psgbufferstart = (psgbufferstart+1)&32767;
				psgbufferlen--;
			}

			// Apply bias and saturation 
			sampleL += sndbias;
			if (sampleL > 32767)
				sampleL = 32767;
			else if (sampleL < -32768)
				sampleL = -32768;
			

			sampleR += sndbias;
			if (sampleR > 32767)
				sampleR = 32767;
			else if (sampleR < -32768)
				sampleR = -32768;

			stream[0][i+i]   = sampleL;
			stream[0][i+i+1] = sampleR;
		}
	}
}


void __stdcall apu_write(unsigned int offset, unsigned int data) {
	int chn,reg;
	__int64 freq;

	offset &= 0xFFF;

	//wsprintf(szbuffer,"APU: [%x] = %x\n",offset,data);
	//WriteConsole(hout,szbuffer,lstrlen(szbuffer),(unsigned long*)&dummy,NULL);

	if (offset<0x500) {
		chn = (offset>>4)&0xF;
		switch (offset&0xF) {
			case 0:				// SOUND*CNT
			case 1:
			case 2:
			case 3:
				if (channels[chn].fmt == 3)
					psg_mix(0);

				if (IOREGS_ARM7[offset+3]&0x80) {
					if (!channels[chn].enabled) {
						channels[chn].fmt = (soundFormat_t)(((*(int*)(IOREGS_ARM7 + 0x400 + (chn<<4)))>>29)&3);
						//if (channels[chn].fmt != 3)
							channels[chn].pos = 0;
						channels[chn].looppoint = (*(unsigned short*)(IOREGS_ARM7 + 0x40A + (chn<<4)) << 9);
						channels[chn].looptype = (((*(int*)(IOREGS_ARM7 + 0x400 + (chn<<4)))>>27)&3); 
						channels[chn].duty = (((*(int*)(IOREGS_ARM7 + 0x400 + (chn<<4)))>>24)&7); 
						channels[chn].enabled = true;
						channels[chn].left = 0x7f- (IOREGS_ARM7[0x400 + (chn<<4) + 2]&0x7f);
						channels[chn].right = (IOREGS_ARM7[0x400 + (chn<<4) + 2]&0x7f);

						channels[chn].vol = ((IOREGS_ARM7[0x500]&0x7f));
						channels[chn].vol *= ((IOREGS_ARM7[0x400 + (chn<<4)]&0x7f));
						channels[chn].vol >>= 7;

						channels[chn].left *= channels[chn].vol;
						channels[chn].right *= channels[chn].vol;

						channels[chn].noise = HIGH;
						channels[chn].tap = HIGH;

						channels[chn].end = channels[chn].nsamples;

						channels[chn].nsamples = channels[chn].nwords * ((channels[chn].fmt==0)?4:2);
						wsprintf(szbuffer,"Channel %d volume=%d, pos=%d, lt=%d, fmt=%d, end=%d, freq=%d, d=%d, lr = %d,%d\n",
							chn,channels[chn].vol,channels[chn].pos,channels[chn].looptype,
							channels[chn].fmt,channels[chn].end,channels[chn].freq>>4,channels[chn].delta,
							channels[chn].left,channels[chn].right);
						WriteConsole(hout,szbuffer,lstrlen(szbuffer),(unsigned long*)&dummy,NULL);
					} else {
						channels[chn].looptype = (((*(int*)(IOREGS_ARM7 + 0x400 + (chn<<4)))>>27)&3); 
						channels[chn].left = 0x7f- (IOREGS_ARM7[0x400 + (chn<<4) + 2]&0x7f);
						channels[chn].right = (IOREGS_ARM7[0x400 + (chn<<4) + 2]&0x7f);
						channels[chn].duty = (((*(int*)(IOREGS_ARM7 + 0x400 + (chn<<4)))>>24)&7); 

						channels[chn].fmt = (soundFormat_t)(((*(int*)(IOREGS_ARM7 + 0x400 + (chn<<4)))>>29)&3);
						channels[chn].vol = ((IOREGS_ARM7[0x500]&0x7f));
						channels[chn].vol *= ((IOREGS_ARM7[0x400 + (chn<<4)]&0x7f));
						channels[chn].vol >>= 7;

						channels[chn].noise = HIGH;
						channels[chn].tap = HIGH;

						//if (channels[chn].fmt == 3)
						//	channels[chn].pos = 0;
						//channels[chn].end = channels[chn].nsamples;

						channels[chn].left *= channels[chn].vol;
						channels[chn].right *= channels[chn].vol;

						/*wsprintf(szbuffer,"Channel %d already enabled\n",
							chn);
						WriteConsole(hout,szbuffer,lstrlen(szbuffer),(unsigned long*)&dummy,NULL);
					*/}
					if (channels[chn].fmt == 3)
						psgdata |= (1<<chn);
				} else {
					channels[chn].enabled = false;
					if (channels[chn].fmt == 3)
						psgdata &= ~(1<<chn);
				}
				break;
			
			case 4:				// SOUND*SAD
				channels[chn].sad = *(unsigned int*)(IOREGS_ARM7 + 0x404 + (chn<<4));
				channels[chn].sad &= 0x07FFFFFF;

				/*wsprintf(szbuffer,"Setting source address %X for channel %d\n",
					channels[chn].sad,chn);
				WriteConsole(hout,szbuffer,lstrlen(szbuffer),(unsigned long*)&dummy,NULL);*/
				break;

			case 8:				// SOUND*TMR
				freq = 16777216*16 / (0x10000 - *(unsigned short*)(IOREGS_ARM7 + 0x408 + (chn<<4)));
				if (channels[chn].fmt == 3)
					psg_mix(0);
				channels[chn].delta3 = (freq<<16)/sampleRate;
				channels[chn].delta2 = (freq<<5)/sampleRate;
				channels[chn].delta = (channels[chn].delta2*frameRate)/60;
				channels[chn].freq = freq;

				/*wsprintf(szbuffer,"Setting frequency %d for channel %d, delta=%d\n",
					freq>>4,chn,channels[chn].delta);
				WriteConsole(hout,szbuffer,lstrlen(szbuffer),(unsigned long*)&dummy,NULL);*/
				break;
			
			case 10:			// SOUND*PNT
				channels[chn].looppoint = (*(unsigned short*)(IOREGS_ARM7 + 0x40A + (chn<<4)) << 9);
				//channels[chn].end = channels[chn].looppoint + (channels[chn].nsamples<<9);
				break;

			case 12:			// SOUND*LEN
				channels[chn].nwords = *(unsigned int*)(IOREGS_ARM7 + 0x40C + (chn<<4));
				channels[chn].nwords &= 0x003FFFFF;
				if (channels[chn].fmt == FMT_PCM8)
					channels[chn].nsamples = channels[chn].nwords * 4;
				else if (channels[chn].fmt == FMT_PCM16)
					channels[chn].nsamples = channels[chn].nwords * 2;
				else if (channels[chn].fmt == FMT_ADPCM)
					channels[chn].nsamples = channels[chn].nwords * 8;

				//channels[chn].end = channels[chn].looppoint + (channels[chn].nsamples<<9);
				/*wsprintf(szbuffer,"Channel %d volume=%d, len=%d, pos=%d, lt=%d\n",
					chn,channels[chn].vol,channels[chn].len,channels[chn].pos,channels[chn].looptype);
				WriteConsole(hout,szbuffer,lstrlen(szbuffer),(unsigned long*)&dummy,NULL);*/
				break;
		}
	}

	if (offset == 0x504) {
		sndbias = *(short*)(IOREGS_ARM7 + 0x504);
		sndbias = (sndbias - 0x200) * 0x20;
	}

	if ((IOREGS_ARM7[0x501] & 0x80) == 0)	// Check sound master enable bit
		return;


	return;
}



void __stdcall apu_update() {
	/*pcm.len = 2;
	stream[0] = (short*)&stream[1][stream1pos<<1];
	buffer_played = 0;
	apu_mix();
	stream1pos += 2;*/
}


/**
 *  @brief Returns information about the plugin
 */
int __stdcall apu_get(int what) {
	switch (what) {
		case APU_WRITE_CALLBACK:
			return (int)apu_write;
		case APU_MIX_CALLBACK:
			return (int)apu_mix;
		case PLUGIN_NAME:
			return (int)&apuName;
		case PLUGIN_VERSION:
			return APU_VERSION;
		case LAST_ERROR:
			return (int)lpszError;
		default:
			break;
	}
	return 0;
}


void __stdcall apu_set(int what, void* param1, void* param2) {
	switch (what) {
		case CPU_CYCLE_PTR:
			if ((int)param2 == 1)
				cycle = (unsigned int *)param1;
			break;

		case MMU_PAGE_TABLE:
			arm7_pageTable = (unsigned char**)param1;
			break;

		case MMU_PAGE_MASKS:
			arm7_pageMask = (unsigned int*)param1;
			break;

		case APU_PAUSE:
			if ((int)param1==1) {
				BASS_ChannelStop(hStream);
			} else {
				BASS_ChannelPlay(hStream,TRUE);
			}

		case APU_SYNC:
			//frameRate = (int)param1;
			for (int i=0; i<16; i++) {
				//channels[i].delta = (channels[i].delta2*frameRate)/60;
			}
			break;

		case PLUGIN_SETTING:
			switch ((int)param1) {
				case 0x10:
					if ((int)param2 == 0) {
						useSmallBuffer = true;
					} else { 
						useSmallBuffer = false;
					}
					break;
				case 0x11:
					sampleRate = (int)param2;
			}
			break;

		default:
			break;
	}
}

