/** @file arm7_read.cpp
 *  @author Mic
 *  @date 2005-2006
 *
 *  @brief Functions for reading from the ARM7 memory space
 */

#include "arm7_read.h"


/**
 *  @brief Reads 8 bits from an arbitrary address (generic).
 *
 *  @param address Address to read from (absolute).
 *
 *  @return Data read (or random value).
 */
unsigned char RDHDECL arm7_read_byte_gen(unsigned int address)
{

	if (arm7_pageTable[address>>20])
		return *(unsigned char*)(arm7_pageTable[address>>20] + (address&0xFFFFF));

	return rand()&0xFF;
}


/**
 *  @brief Reads 8 bits from the DTCM memory space.
 *  @warning ARM7 does not have access to DTCM.
 *
 *  @param address Address to read from (absolute).
 *
 *  @return Random data.
 */
unsigned char RDHDECL arm7_read_byte_dtcm(unsigned int address)
{
	/*if ((address>=dtcmBase) && (address<dtcmBase+0x4000))
		return *(unsigned char*)(DTCM + (address-dtcmBase));
	if (arm9_pageTable[address>>20])
		return *(unsigned char*)(arm9_pageTable[address>>20] + (address&0xFFFFF));*/

	return rand()&0xFF;
}


// ITCM read
//
unsigned char RDHDECL arm7_read_byte_itcm(unsigned int address)
{
	/*if ((address<0x8000))
		return *(unsigned char*)(ITCM + address);
	if ((address>=dtcmBase) && (address<dtcmBase+0x4000))
		return *(unsigned char*)(DTCM + (address-dtcmBase));*/
	return rand()&0xFF;
}


/**
 *  @brief Reads 8 bits from the I/O memory space.
 *  @todo Handle all necessary registers.
 *
 *  @param address Address to read from (absolute).
 *
 *  @return Data read (or zero).
 */
unsigned char __fastcall arm7_read_byte_io(unsigned int address)
{
	int	offset,page;
	unsigned int tlong;

	/*if (address==0x4100000) {
		ipc_fifo_recv(0);
		return 0;
	}*/

	switch (address&0xFFFFF)
	{
		case P1:
			return keys&0xFF;
		case EXTKEYIN:
			return extkeys&0xFF;

		/*case TM0D:
			if (timer[0].enabled)
			{
				tlong = *arm9_cycle;
				if (tlong < timer[0].tstart)
					tlong -= (timer[0].tstart-0x40000000);
				else
					tlong -= timer[0].tstart;
				timer[0].tcurr = tlong >> prescalar[IOREGS[TM0CNT]&3];
			}
			return timer[0].tcurr;

		case TM1D:
			if (timer[1].countup)
				return timer[1].te;
			if (timer[1].enabled)
			{
				tlong = *arm9_cycle;
				if (tlong < timer[1].tstart)
					tlong -= (timer[1].tstart-0x40000000);
				else
					tlong -= timer[1].tstart;
				timer[1].tcurr = tlong >> prescalar[IOREGS[TM1CNT]&3];
			}
			return timer[1].tcurr;

		case TM2D:
			if (timer[2].countup)
				return timer[2].te;
			if (timer[2].enabled)
			{
				tlong = *arm9_cycle;
				if (tlong < timer[2].tstart)
					tlong -= (timer[2].tstart-0x40000000);
				else
					tlong -= timer[2].tstart;
				timer[2].tcurr = tlong >> prescalar[IOREGS[TM2CNT]&3];
			}
			return timer[2].tcurr;

		case TM3D:
			if (timer[3].countup)
				return timer[3].te;
			if (timer[3].enabled)
			{
				tlong = *arm9_cycle;
				if (tlong < timer[3].tstart)
					tlong -= (timer[3].tstart-0x40000000);
				else
					tlong -= timer[3].tstart;
				timer[3].tcurr = tlong >> prescalar[IOREGS[TM3CNT]&3];
			}
			return timer[3].tcurr;*/

		case IF_:
			return irqFlags7&0xFF;

		// SPICNT
		case 0x1C0:
		// RTCCNT
		case 0x138:
			/*wsprintf(buffer,"(9) Reading from %08X, at %X\n",
				address,arm7_regs[15]);
			WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);*/
			break;

		case 0x180:
			/*wsprintf(buffer,"(9) Reading from %08X, at %X, %X\n",
				address,arm9_regs[15],
				*(unsigned short int*)(arm9_pageTable[0x40] + (address&0xFFFFF)));
			WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);*/
			return *(unsigned char*)(arm7_pageTable[0x40] + (address&0xFFFFF));

		case 0x300:
			return 1;

		default:
			return *(unsigned char*)(arm7_pageTable[0x40] + (address&0xFFFFF));
	}
}


unsigned char __fastcall arm7_read_byte_wifi(unsigned int address) {
	if ((address&0xFFFFF)<0x10000)
		return *(unsigned char*)(arm7_pageTable[0x48] + (address&0xFFFF));
	return 0;
}


// ROM read
//
unsigned char __fastcall arm7_read_byte_rom(unsigned int address) {
	int page,offset=address&0xFFFFF;

	if ((address-0x8000000) >= arm7_pageLength[8]+arm7_pageLength[9]) {
		page = arm7_regs[15]-4;
		offset = (page&0x0FFFFF);
		return *(unsigned char*)(arm7_pageTable[page>>20] + offset);
	}
	
	if (arm7_pageTable[address>>20])
		return *(unsigned char*)(arm7_pageTable[address>>20] + offset);

	return rand()&0xFF;

}


unsigned char __fastcall arm7_read_byte_ptok(unsigned int address)
{
	return *(unsigned char*)(arm7_pageTable[address>>20] + (address&0xFFFFF));
}


/* IWRAM read */
unsigned char __fastcall arm7_read_byte_wram(unsigned int address) {
	return *(unsigned char*)(arm7_pageTable[address>>20] + (address & arm7_pageMask[address>>20]));
}


unsigned char __fastcall arm7_read_byte_ewram(unsigned int address) {
	return *(unsigned char*)(EIWRAM + ((address)&0xFFFF));
}


/* Palette read */
unsigned char __fastcall arm7_read_byte_plt(unsigned int address) {
	return *(unsigned char*)(PLT_VRAM + (address&0xFFF));
}


/* OAM read */
unsigned char __fastcall arm7_read_byte_oam(unsigned int address)
{
	return *(unsigned char*)(OAM + (address&0x7FF));
}


/* SRAM read */
unsigned char __fastcall arm7_read_byte_sram(unsigned int address)
{
	return *(unsigned char*)(arm7_pageTable[0xA] + (address&0xFFFF));
}


// LCDC read
unsigned char __fastcall arm7_read_byte_lcdc(unsigned int address) {
	int chunk;
	address &= 0xFFFFF; if (address>0xA3FFF) address-=0xA3FFF;
	chunk = ((address>>14)|0x200)&0x3FF;
	if (vramChunk[chunk])
		return *(unsigned char*)(vramChunk[chunk] + (address&0x3FFF));
	return 0;
}


// VRAM read
unsigned char __fastcall arm7_read_byte_vram(unsigned int address) {
	int chunk = (address>>14)&0x3FF;
	if (vramChunk[chunk])
		return *(unsigned char*)(vramChunk[chunk] + (address&0x3FFF));
	return 0;
}


unsigned char __fastcall arm7_read_byte_na(unsigned int address)
{
	return rand()&0xFF;
}


unsigned char __fastcall arm7_read_byte_zero(unsigned int address) {
	return 0;
}


///////////////////////////////////////////////////////////////////////////////////


// Generic read
//
unsigned short int RDHDECL arm7_read_word_gen(unsigned int address) {
	dummy = address;

	if (arm7_pageTable[address>>20])
		return *(unsigned short int*)(arm7_pageTable[address>>20] + (address&0xFFFFF));

	return rand()&0xFFFF;
}


// DTCM read
//
unsigned short int RDHDECL arm7_read_word_dtcm(unsigned int address) {
	if ((address>=dtcmBase) && (address<dtcmBase+0x4000))
		return *(unsigned short int*)(DTCM + (address-dtcmBase));
	if (arm7_pageTable[address>>20])
		return *(unsigned short int*)(arm7_pageTable[address>>20] + (address&0xFFFFF));

	return rand()&0xFFFF;
}


// ITCM read
//
unsigned short int RDHDECL arm7_read_word_itcm(unsigned int address)
{
	if ((address<0x8000))
		return *(unsigned short int*)(ITCM + address);
	if ((address>=dtcmBase) && (address<dtcmBase+0x4000))
		return *(unsigned short int*)(DTCM + (address-dtcmBase));
	return rand()&0xFFFF;
}


// I/O read
//
unsigned short int __fastcall arm7_read_word_io(unsigned int address)
{
	int	offset,page;
	unsigned int tlong;
	unsigned short int *pagew;

	if (address==0x4100000) {
		ipc_fifo_recv(0);
		return 0;
	}

	switch (address&0xFFFFF)
	{
		case P1:
			return keys;
		case EXTKEYIN:
			return extkeys;

		case TM0D:
			if (timer[1][0].enabled)
			{
				tlong = (*arm7_cycle)<<1;
				if (tlong < timer[1][0].tstart)
					tlong -= (timer[1][0].tstart-0x40000000);
				else
					tlong -= timer[1][0].tstart;
				timer[1][0].tcurr = tlong >> (prescalar[IOREGS_ARM7[TM0CNT]&3]+1);
			}
			return timer[1][0].tcurr;

		case TM1D:
			if (timer[1][1].countup)
				return timer[1][1].te;
			if (timer[1][1].enabled)
			{
				tlong = (*arm7_cycle)<<1;
				if (tlong < timer[1][1].tstart)
					tlong -= (timer[1][1].tstart-0x40000000);
				else
					tlong -= timer[1][1].tstart;
				timer[1][1].tcurr = tlong >> (prescalar[IOREGS_ARM7[TM1CNT]&3]+1);
			}
			return timer[1][1].tcurr;

		case TM2D:
			if (timer[1][2].countup)
				return timer[1][2].te;
			if (timer[1][2].enabled)
			{
				tlong = (*arm7_cycle<<1);
				if (tlong < timer[1][2].tstart)
					tlong -= (timer[1][2].tstart-0x40000000);
				else
					tlong -= timer[1][2].tstart;
				timer[1][2].tcurr = tlong >> (prescalar[IOREGS_ARM7[TM2CNT]&3]+1);
			}
			return timer[1][2].tcurr;

		case TM3D:
			if (timer[1][3].countup)
				return timer[1][3].te;
			if (timer[1][3].enabled)
			{
				tlong = (*arm7_cycle)<<1;
				if (tlong < timer[1][3].tstart)
					tlong -= (timer[1][3].tstart-0x40000000);
				else
					tlong -= timer[1][3].tstart;
				timer[1][3].tcurr = tlong >> (prescalar[IOREGS_ARM7[TM3CNT]&3]+1);
			}
			return timer[1][3].tcurr;

		case IF_:
			return irqFlags7&0xFFFF;

		//case 0x136:
			//if ((IOREGS_ARM7[0x135]&0xFF)==0x80) {
		//		return IOREGS_ARM7[0x136];
			//} else
			//	return 0;

		// SPICNT
		//case 0x1C0:
		//case 0x1C2:
			/*wsprintf(buffer,"(7) Reading %04X from %08X, at %X\n",
				*(unsigned short int*)(arm7_pageTable[0x40] + 0x1C2),address,arm7_regs[15]);
			WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);*/
		//	return *(unsigned short int*)(arm7_pageTable[0x40] + 0x1C2);

		// RTCCNT
		case 0x138:
		//case 0x184:
			wsprintf(buffer,"(7) Reading from %08X, at %X\n",
				address,arm7_regs[15]);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			return 0;
			break;

		case 0x600:
			gpu_get(GPU_3D_COMMAND);
			return *(unsigned short int*)(IOREGS + 0x600);

		case 0x300:
			return 1;

		case 0x10:
		case 0x12:
		case 0x14:
		case 0x16:
		case 0x18:
		case 0x1A:
		case 0x1C:
		case 0x1E:
			page = arm7_regs[15];
			pagew = (unsigned short int*)arm7_pageTable[page>>20];
			offset = (page&0x0FFFFF)>>1;
			return pagew[offset];

		default:
			/*wsprintf(buffer,"(7) Reading from %08X, at %X\n",
				address,arm7_regs[15]);
			WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);*/
			return *(unsigned short int*)(arm7_pageTable[0x40] + (address&0xFFFFF));
	}
}


unsigned short int __fastcall arm7_read_word_wifi(unsigned int address) {
	if ((address&0xFFFFF)<0x10000)
		return *(unsigned short int*)(arm7_pageTable[0x48] + (address&0xFFFF));
	return 0;
}


// ROM read
//
unsigned short int __fastcall arm7_read_word_rom(unsigned int address) {
	int page,offset=address&0xFFFFF;
	unsigned short int *pagew;

	if ((address-0x8000000) >= arm7_pageLength[0x8])
	{
		page = arm7_regs[15]-4;
		pagew = (unsigned short int*)arm7_pageTable[page>>20];
		offset = (page&0x0FFFFF)>>1;
		return pagew[offset];
	}
	
	if (arm7_pageTable[address>>20])
		return *(unsigned short int*)(arm7_pageTable[address>>20] + offset);

	return rand()&0xFFFF;

}


unsigned short int __fastcall arm7_read_word_ptok(unsigned int address)
{
	return *(unsigned short int*)(arm7_pageTable[address>>20] + (address&0xFFFFF));
}


// IWRAM read
//
unsigned short int __fastcall arm7_read_word_wram(unsigned int address) {
/*	switch (IOREGS[0x247]&3) {
		case 0:
			return *(unsigned short int*)(EIWRAM + (address&0xFFFF));
		case 1:
			return *(unsigned short int*)(IWRAM + ((address)&0x3FFF));
		case 2:
			return *(unsigned short int*)(IWRAM + 0x4000 + (address&0x3FFF));
		case 3:
			return *(unsigned short int*)(IWRAM + ((address)&0x7FFF));
	}
	return 0;*/
	return *(int*)(arm7_pageTable[address>>20] + (address & arm7_pageMask[address>>20]));

	//return *(unsigned short int*)(IWRAM + ((address)&0x7FFF));
}


unsigned short int __fastcall arm7_read_word_ewram(unsigned int address) {
	return *(unsigned short int*)(EIWRAM + ((address)&0xFFFF));
}


// Palette read
unsigned short int __fastcall arm7_read_word_plt(unsigned int address)
{
	return *(unsigned short int*)(PLT_VRAM + (address&0xFFF));
}


// OAM read
unsigned short int __fastcall arm7_read_word_oam(unsigned int address)
{
	return *(unsigned short int*)(OAM + (address&0x7FF));
}


// LCDC read
unsigned short int __fastcall arm7_read_word_lcdc(unsigned int address)
{
	int chunk;
	address &= 0xFFFFF; if (address>0xA3FFF) address-=0xA3FFF;
	chunk = ((address>>14)|0x200)&0x3FF;
	if (vramChunk[chunk])
		return *(unsigned short int*)(vramChunk[chunk] + (address&0x3FFF));
	return 0;
}


// VRAM read
unsigned short int __fastcall arm7_read_word_vram(unsigned int address)
{
	int chunk = (address>>14)&0x3FF;
	if (vramChunk[chunk])
		return *(unsigned short int*)(vramChunk[chunk] + (address&0x3FFF));
	return 0;
}


unsigned short int __fastcall arm7_read_word_na(unsigned int address)
{
	return rand()&0xFFFF;
}


unsigned short int __fastcall arm7_read_word_zero(unsigned int address)
{
	return 0;
}




unsigned short int __fastcall arm7_read_word_ipc(unsigned int address)
{
	int offset = address&0xFFFFF;

	/*if (offset == ipc_x)
		return (touch->x*14);
	if (offset == ipc_y)
		return (touch->y*8917)>>9;

	if (offset == ipc_z)
		return ((touch->z&0x40)|(btnXY[0])|(btnXY[1]<<1))^0x37F;*/

	if (arm7_pageTable[ipc_page])
		return *(unsigned short int*)(arm7_pageTable[ipc_page] + offset);

	return rand()&0xFFFF;
}


unsigned short int __fastcall arm7_read_word_ipc_mirror(unsigned int address)
{
	int offset = address&0xFFFFF;

	if (offset == 0xFF004)
		return (touch->x*14);
	if (offset == 0xFF006)
		return (touch->y*8917)>>9;

	if (offset == 0xFF014)
		return ((touch->z&0x40)|(btnXY[0])|(btnXY[1]<<1))^0x37F;

	if (arm7_pageTable[address>>20])
		return *(unsigned short int*)(arm7_pageTable[address>>20] + offset);

	return rand()&0xFFFF;
}


unsigned short int __fastcall arm7_read_word_ipc_v2(unsigned int address)
{
	int offset = address&0xFFFFF;

	/*if (offset == ipc_x)
		return (touch->x*14);
	if (offset == ipc_y)
		return ((touch->y-16)*8917)>>9;

	if (offset == ipc_x+4)
		return (touch->x);
	if (offset == ipc_y+4)
		return (touch->y-16);
	
	if (offset == ipc_z + 4)
		return ((touch->z&0x40)|(btnXY[0])|(btnXY[1]<<1))^0x37F;*/

	if (arm7_pageTable[ipc_page])
		return *(unsigned short int*)(arm7_pageTable[ipc_page] + offset);

	return rand()&0xFFFF;
}


unsigned short int __fastcall arm7_read_word_ipc_mirror_v2(unsigned int address)
{
	int offset = address&0xFFFFF;

	/*if (offset == 0xFF004)
		return (touch->x*14);
	if (offset == 0xFF006)
		return ((touch->y-16)*8917)>>9;

	if (offset == 0xFF008)
		return (touch->x);
	if (offset == 0xFF00A)
		return (touch->y-16);

	if (offset == 0xFF014 + 4)
		return ((touch->z&0x40)|(btnXY[0])|(btnXY[1]<<1))^0x37F;*/

	if (arm7_pageTable[address>>20])
		return *(unsigned short int*)(arm7_pageTable[address>>20] + offset);

	return rand()&0xFFFF;
}

/////////////////////////////////////////////////////////////////////////////////////////

int __fastcall arm7_read_dword_gen(unsigned int address)
{
	int n;
	dummy = address;

	__asm
	{
		mov ebx,address
		mov edx,arm7_pageTableAddr
	and ebx,0xFFFFFFF
		mov edi,arm7_pageLengthAddr
		mov esi,ebx
		mov ecx,arm7_pageMaskAddr
		shr ebx,20			// page
		and esi,0x0FFFFF	// offset
		mov eax,ebx
		shr eax,4

		jmp __rd_normal

		// Unaligned read, rotate data
__rd_unaligned:
		mov ecx,esi
		and esi,0xFFFFFFFC		// Clear lower 2 bits of offset
		and ecx,3
		add esi,[edx+ebx*4]		// ADD arm7_pageTable[page]
		shl ecx,3
		mov eax,[esi]
		ror eax,cl				// Rotate value by (Offset & 3)*8
		jmp __rd_done

		// Return "random" data
__rd_random:
		mov ecx,ebp
		mov eax,address
		ror eax,cl
		jmp __rd_done

__rd_normal:
		test esi,3
		jnz __rd_unaligned
		and esi,[ecx+ebx*4]		// AND with pageMask
		mov edi,esi
		add esi,[edx+ebx*4]		// ADD arm7_pageTable[page]
		cmp esi,edi
		je __rd_done
		mov eax,[esi]

__rd_done:
		mov n,eax
	}

	return n;
}


int __fastcall arm7_read_dword_dtcm(unsigned int address) {
	int n = 0;
	return n;
}


int __fastcall arm7_read_dword_itcm(unsigned int address)
{
	int n = 0;
	return n;
}


int __fastcall arm7_read_dword_io(unsigned int address) {
	int n;

	//	wsprintf(buffer,"ARM7 reading word from %08X\n",address);
	//	WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

	if (address==0x4100000) {
		return ipc_fifo_recv(1);
	}

	if (address == 0x04100010) {
		wsprintf(buffer,"Reading from %08X at %08X\n",
			address,arm9_regs[15]);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
	}

	__asm
	{
		mov ebx,address
		mov edx,arm7_pageTableAddr
		mov edi,arm7_pageLengthAddr
		mov esi,ebx
		mov ecx,arm7_pageMaskAddr
		shr ebx,20			// page
		and esi,0x0FFFFF	// offset
		mov eax,ebx
		shr eax,4

		cmp esi,P1
		jne __rdio_check_extkeys
		mov ebx,[IOREGS + P1CNT]
		movzx eax,keys
		shl ebx,16
		or eax,ebx
		jmp __rd_done
__rdio_check_extkeys:
		cmp esi,EXTKEYIN
		jne __rd_normal
		mov ebx,[IOREGS_ARM7 + 0x138]
		movzx eax,extkeys
		shl ebx,16
		or eax,ebx
		jmp __rd_done

		// Unaligned read, rotate data
__rd_unaligned:
		mov ecx,esi
		and esi,0xFFFFFFFC		// Clear lower 2 bits of offset
		and ecx,3
		add esi,[edx+ebx*4]		// ADD arm7_pageTable[page]
		shl ecx,3
		mov eax,[esi]
		ror eax,cl				// Rotate value by (Offset & 3)*8
		jmp __rd_done

		// Return "random" data
__rd_random:
		mov ecx,ebp
		mov eax,address
		ror eax,cl
		jmp __rd_done

__rd_normal:
		test esi,3
		jnz __rd_unaligned
		and esi,[ecx+ebx*4]		// AND with pageMask (was eax*4)
		add esi,[edx+ebx*4]		// ADD arm7_pageTable[page]
		mov eax,[esi]

__rd_done:
		mov n,eax
	}

	return n;
}


/**
 *  @brief Reads 32 bits from the WiFi memory space.
 *
 *  @param address Address to read from (absolute).
 *
 *  @return int Data read (or zero).
 */
int __fastcall arm7_read_dword_wifi(unsigned int address) {
	if ((address&0xFFFFF)<0x10000)
		return *(int*)(arm7_pageTable[0x48] + (address&0xFFFF));
	return 0;
}


/**
 *  @brief Reads 32 bits from the ROM memory space.
 *
 *  @param address Address to read from (absolute).
 *
 *  @return int Data read (or zero).
 */
int __fastcall arm7_read_dword_rom(unsigned int address) {
	int page,offset;
	int *pagew;

	if ((address&0xFFFFF) >= arm7_pageLength[0x8])
	{
		page = arm7_regs[15]-4;
		pagew = (int*)arm7_pageTable[page>>20];
		offset = (page&0x0FFFFF)>>2;
		return pagew[offset];
	}
	
	if (arm7_pageTable[address>>20])
		return *(int*)(arm7_pageTable[address>>20] + (address&0xFFFFF));

	return rand()&0xFFFFFFFF;

}


int __fastcall arm7_read_dword_ptok(unsigned int address) {
	return *(int*)(arm7_pageTable[address>>20] + (address&0xFFFFF));
}


int __fastcall arm7_read_dword_wram(unsigned int address) {
	return *(int*)(arm7_pageTable[address>>20] + (address & arm7_pageMask[address>>20]));
}


int __fastcall arm7_read_dword_ewram(unsigned int address) {
	/*if ((address >= 0x3807a10)&&(address <= 0x3807a14)) {
		wsprintf(buffer,"Reading %08X from %08X, PC=%08X\n",*(int*)(EIWRAM + ((address)&0xFFFF)),address,arm7_regs[15]);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
	}*/

	return *(int*)(EIWRAM + ((address)&0xFFFF));
}


int __fastcall arm7_read_dword_plt(unsigned int address)
{
	return *(int*)(PLT_VRAM + (address&0xFFF));
}


int __fastcall arm7_read_dword_oam(unsigned int address)
{
	return *(int*)(OAM + (address&0x7FF));
}


int __fastcall arm7_read_dword_lcdc(unsigned int address)
{
	int n;

	__asm
	{
		mov ebx,address
		and ebx,0xFFFFF
		cmp ebx,0xA3FFF
		jbe __rd_address_ok
		sub ebx,0xA3FFF
__rd_address_ok:
		mov edx,ebx
		lea esi,vramChunk
		shr edx,14
		and ebx,0x3FFF
		//or edx,0x200
		xor eax,eax
		and edx,0x3FF
		or eax,[esi + edx*4]
		jz __rd_random
		
		jmp __rd_normal

		// Unaligned read, rotate data
__rd_unaligned:
		mov ecx,ebx
		and ebx,0xFFFFFFFC		// Clear lower 2 bits of offset
		and ecx,3
		shl ecx,3
		mov esi,[eax+ebx]
		ror esi,cl				// Rotate value by (Offset & 3)*8
		jmp __rd_done

		// Return "random" data
__rd_random:
		mov ecx,ebp
		mov eax,address
		ror eax,cl
		jmp __rd_done

__rd_normal:
		test esi,3
		jnz __rd_unaligned
		mov esi,[eax+ebx]

__rd_done:
		mov n,esi
	}

	return n;
}


int __fastcall arm7_read_dword_vram(unsigned int address)
{
	int n;

	__asm
	{
		mov ebx,address
		mov edx,ebx
		lea esi,vramChunk
		shr edx,14
		and ebx,0x3FFF
		//or edx,0x200
		xor eax,eax
		and edx,0x3FF
		or eax,[esi + edx*4]
		jz __rd_random
		
		jmp __rd_normal

		// Unaligned read, rotate data
__rd_unaligned:
		mov ecx,ebx
		and ebx,0xFFFFFFFC		// Clear lower 2 bits of offset
		and ecx,3
		shl ecx,3
		mov esi,[eax+ebx]
		ror esi,cl				// Rotate value by (Offset & 3)*8
		jmp __rd_done

		// Return "random" data
__rd_random:
		mov ecx,ebp
		mov eax,address
		ror eax,cl
		jmp __rd_done

__rd_normal:
		test esi,3
		jnz __rd_unaligned
		mov esi,[eax+ebx]

__rd_done:
		mov n,esi
	}

	return n;
}


int __fastcall arm7_read_dword_na(unsigned int address)
{
	return rand()&0xFFFFFFFF;
}


int __fastcall arm7_read_dword_zero(unsigned int address)
{
	return 0;
}


int __fastcall arm7_read_dword_ipc(unsigned int address)
{
	int offset = address&0xFFFFF;

	if (offset == ipc_x)
		return (touch->x*14);
	if (offset== ipc_y)
		return (touch->y*8917)>>9;
	if (offset == ipc_z)
		return (touch->z&0x40)^0x40;

	if (arm7_pageTable[ipc_page])
		return *(int*)(arm7_pageTable[ipc_page] + offset);

	return rand()&0xFFFFFFFF;
}