/** @file fat.h
 *  @author Mic
 *	@date 2006
 *  @brief FAT emulation declarations for Dualis.\n
 * Structures taken from Michael Chisholm's FAT library
 */   

#ifndef __FAT_H__
#define __FAT_H__

// 

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;

#define ATTRIB_DIR 0x10
#define ATTRIB_LFN 0x0F

#define FILE_FREE 0xE5

#define __PACKED
 #pragma pack(1)


/** 
 *  @brief Boot sector structure - must be packed
 */
typedef struct BOOT_RECORD
{
	__PACKED	u8	jmpBoot[3];
	__PACKED	u8	OEMName[8];
	// BIOS Parameter Block
	__PACKED	u16	bytesPerSector;
	__PACKED	u8	sectorsPerCluster;
	__PACKED	u16	reservedSectors;
	__PACKED	u8	numFATs;
	__PACKED	u16	rootEntries;
	__PACKED	u16	numSectorsSmall;
	__PACKED	u8	mediaDesc;
	__PACKED	u16	sectorsPerFAT;
	__PACKED	u16	sectorsPerTrk;
	__PACKED	u16	numHeads;
	__PACKED	u32	numHiddenSectors;
	__PACKED	u32	numSectors;
	struct  
	{
		// Ext BIOS Parameter Block for FAT16
		__PACKED	u8	driveNumber;
		__PACKED	u8	reserved1;
		__PACKED	u8	extBootSig;
		__PACKED	u32	volumeID;
		__PACKED	u8	volumeLabel[11];
		__PACKED	u8	fileSysType[8];
		// Bootcode
		__PACKED	u8	bootCode[448];
		__PACKED	u16 signature;
	} fat16;
} BOOT_RECORD;


/** 
 *  @ Directory entry struct - must be packed
 */
typedef struct DIR_ENT
{
	__PACKED	u8	name[8];
	__PACKED	u8	ext[3];
	__PACKED	u8	attrib;
	__PACKED	u8	reserved;
	__PACKED	u8	cTime_ms;
	__PACKED	u16	cTime;
	__PACKED	u16	cDate;
	__PACKED	u16	aDate;
	__PACKED	u16	startClusterHigh;
	__PACKED	u16	mTime;
	__PACKED	u16	mDate;
	__PACKED	u16	startCluster;
	__PACKED	u32	fileSize;
} DIR_ENT;

#endif
