// DS GPU emulator
// Mic, 2004/2007
//
// Extern symbols are found mainly in dsgpu_asm.asm
//

#include "stdafx.h"
#include "colorfx.h"
#include "zombie.h"

#include <commctrl.h>
#include <string.h>
#include <stdio.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "d3dsetup.h"
#include <d3d8types.h>
#include "resource.h"
#include "CMatrixStack.h"


#define GPU_VERSION	31

#define JUMP	0
#define VECTOR	1

#define VRAMCNT_A			0x240
#define VRAMCNT_B			0x241
#define VRAMCNT_C			0x242
#define VRAMCNT_D			0x243
#define VRAMCNT_E			0x244
#define VRAMCNT_F			0x245
#define VRAMCNT_G			0x246
#define VRAMCNT_H			0x248
#define VRAMCNT_I			0x249


#define TEXCACHE_SIZE 6


void draw_scanline(unsigned char*,unsigned char*,int);
void finish_screen_d3d();
void finish_screen_ogl();
void finish_screen_gdi();
void reset_screen();
BOOL CALLBACK DlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam);


// Background layer struct
typedef struct {
	int num,prio;
}tag_bg;


typedef struct {
	GLenum id;
	unsigned int addr,fmt,color,width,height,numcolors;
	GLfloat map_r[256],map_g[256],map_b[256],map_a[256];
}tag_texture;


typedef void(*draw_bg_callback)(int);


HANDLE					hout;
HINSTANCE				hInst;
HWND					hwnd,hwnd_s;
HDC						dc,dc_s;
RECT					rect;
HGLRC					hrc;
PIXELFORMATDESCRIPTOR	pfd;
int						pixformat;
GLuint					texID[16];
int						texIdNum;
tag_texture				textureCache[TEXCACHE_SIZE];

CMatrixStack			*mtxStack[4];


char					dataLatch;
short int				*colorLUT;
int						filter = GL_LINEAR;
int						width,visibleLines,totalLines,nameTblHeight;
int						hbiAction,hbiAddress,vbiAction,vbiAddress;
int						yscroll;
int						vramPtr,addressLatch;
int						spriteDbg[4];
int						htile,patLine,patLineInv; //,tileShift;
int						mosaicX,mosaicY;
int						addressShift;
int						dummy;
extern "C" int			newframe;
unsigned int			frameInterval;
draw_bg_callback		draw_bg_callbacks[6][4],draw_bg;

unsigned int			textBgMasks[4][2] = {{0x1f,0x1f},{0x3f,0x1f},{0x1f,0x3f},{0x3f,0x3f}};
unsigned int			textBgShifts[4] = {5,6,5,6};
unsigned int			textScrMasks[4][2] = {{0x0,0x0},{0x100,0x0},{0x0,0x100},{0x100,0x100}};
//unsigned int			textScrMasks[4][2] = {{0xff,0xff},{0x1ff,0xff},{0xff,0x1ff},{0x1ff,0x1ff}};
unsigned int			textScrShifts[2][4] = {{0,3,0,3},{0,3,0,4}};
unsigned int			rsBgMasks[4][2] = {{0x0f,0x0f},{0x1f,0x1f},{0x3f,0x3f},{0x7f,0x7f}};
unsigned int			rsBgShifts[4] = {4,5,6,7};
float					mtx_4x4[4][4] = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
float					mtx_4x3[4][4] = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
float					mtx_3x3[4][4] = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};

DWORD					lastFrameTime;

bool					hblankEnabled,vblankEnabled;
bool					limitFps = true;
bool					enableObj = true;
bool					enableLayer[2][5] = {{true,true,true,true,true},{true,true,true,true,true}};
bool					forceLayer[5] = {false,false,false,false,false};
bool					emulate3D,block3Dcommands;

char					gpuName[] = "DS GPU";
char					szMMXError[] = "This program requires MMX to run";
char					szMemoryError[] = "Unable to allocate sufficient memory";
char					*lpszError;

int						frameskip = 0,frames;
int						displayScreens;
int						lastbg;
unsigned int			BG2X[2],BG2Y[2],BG2PA[2],BG2PB[2],BG2PC[2],BG2PD[2];
unsigned int			BG3X[2],BG3Y[2],BG3PA[2],BG3PB[2],BG3PC[2],BG3PD[2];
unsigned int			bg2x[2],bg2y[2],bg3x[2],bg3y[2];
extern "C" unsigned int x2,y2,deltax,deltay;
extern "C" unsigned int objx2,objy2,objdx,objdy,objdmx,objdmy;
extern "C" unsigned int xclip,clippedWidth;
extern "C" unsigned int tileOffset;
extern "C" unsigned int core;
unsigned int			midpoint[2];
unsigned int			painv,pbinv,pcinv,pdinv;
vram_bank				*vramBank;
unsigned char			*vramBase[9];
unsigned int linesCaptured;

//unsigned int			rs_inverse[0x8000];
char					buffer[400];
extern "C" unsigned int objwidth[16];
extern "C" unsigned int objheight[16];
int						log2_lut[32];

typedef struct {
	int cmd,parm;
} t_3d_command;

t_3d_command			commandList[512];
int						commandListPos;

int						renderer = 0;

////////////////////////////////////////////////////////////////////////////////////////////////

// Used by the OpenGL renderer
//float					vertexes[4][2] = {{1.0f,1.0f},{-1.0f,1.0f},{-1.0f,-0.99f},{1.0f,-0.99f}};
float					vertexes[4][2] = {{1.0f,1.0f},{-1.0f,1.0f},{-1.0f,0.0f},{1.0f,0.0f}};
float					vertexes_s[4][2] = {{1.0f,0.0f},{-1.0f,0.0f},{-1.0f,-1.0f},{1.0f,-1.0f}};
float					vertexes_3[4][2] = {{1.0f,0.2f},{-1.0f,0.2f},{-1.0f,0.0f},{1.0f,0.0f}};
float					vertexes_single[4][2] = {{1.0f,1.0f},{-1.0f,1.0f},{-1.0f,-1.0f},{1.0f,-1.0f}};
#define _S 1.0f
#define _T (0.375f)
float					texCoords[4][2] = {{_S,0.0f},{0.0f,0.0f},{0.0f,_T},{_S,_T}};
float					texCoords_s[4][2] = {{_S,_T},{0.0f,_T},{0.0f,_T*2.0f},{_S,_T*2.0f}};
float					texCoords3D[4][2] = {{_S,0.75f},{0.0f,0.75f},{0.0f,0.0f},{_S,0.0f}};

t_matrix_4x4			mtx_current[4];
t_matrix_4x4			mtx_trans = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
t_matrix_4x4			mtx_scale = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
int						map3d_mtxMode[4] = {GL_PROJECTION,0,GL_MODELVIEW,GL_TEXTURE};
int						map3d_begend[4] = {GL_TRIANGLES,GL_QUADS,GL_TRIANGLE_STRIP,GL_QUAD_STRIP};
int						map3d_cull[4] = {GL_FRONT_AND_BACK,GL_FRONT,GL_BACK,0};
int						framecnt = 0;
int						mtx_stat[3][3] = {{0,13,1<<13},{0,0,0},{0,8,31<<8}};
int						mtx_mode=2;
GLfloat					lightv[4];
t_3d_command			fifocommands[16];
int						fifopos = -1,fifolen=0,fifoparms=0;
float					polyalpha=1.0;
GLfloat					currColor[4];
bool					vtx_begin = false;
bool					lightAttr[4] = {false,false,false,false};
GLenum					polymode = GL_FILL;
int						activeTexture;

////////////////////////////////////////////////////////////////////////////////////////////////

// Used by the D3D renderer
typedef float			customVertex[6];
#define _V2 0.373046875f
#define _V3 0.375f
#define _V4 0.748046875f

customVertex			mainCore_topScreen[12] = {
	{0,0,0, 1, 0,0},
	{255.5f,0,0, 1, 1,0},
	{255.5f,191.5f,0, 1, 1,_V3},
	{0,191.5f,0, 1, 0,_V3},
	{0,0,0, 1, 0,0},
	{255.5f,191.5f,0, 1, 1,_V3},	//
	{0,192,0, 1, 0,_V3},
	{255.5f,192,0, 1, 1,_V3},
	{255.5f,383.5f,0, 1, 1,_V4},
	{0,383.5f,0, 1, 0,_V4},
	{0,192,0, 1, 0,_V3},
	{255.5f,383.5f,0, 1, 1,_V4},
};

#define _V5 0.75f

customVertex			mainCore_bottomScreen[12] = {
	{0,192,0, 1, 0,0},
	{255.5f,192,0, 1, 1,0},
	{255.5f,383.5f,0, 1, 1,_V3},
	{0,383.5f,0, 1, 0,_V3},
	{0,192,0, 1, 0,0},
	{255.5f,383.5f,0, 1, 1,_V3},	//
	{0,0,0, 1, 0,_V3},
	{255.5f,0,0, 1, 1,_V3},
	{255.5f,191.5f,0, 1, 1,_V5},
	{0,191.5f,0, 1, 0,_V5},
	{0,0,0, 1, 0,_V3},
	{255.5f,191.5f,0, 1, 1,_V5},
};

IDirect3DVertexBuffer8	*lpVertexBuf1,*lpVertexBuf2,*lpVertexBufCurr;
IDirect3DTexture8		*lpTexture;
D3DSURFACE_DESC			d3ddesc;
char					*lpVertexBufferData;

////////////////////////////////////////////////////////////////////////////////////////////////

int						colormode = 0;
int						colormask[2][3] = {{0x001F,0x03E0,0x7C00}, {0x001F,0x03E0,0x7C00}};

int						sizemult = 1;

tag_bg					layers[5],sorted_layers[5];	// 4 BG layers plus the OAM layer

cpu_irq_callback		cpu_irq,arm7_irq;
cpu_set_io_handler_callback cpu_set_io_handler;
//mmu_translate_address_callback mmu_translate_address;
mmu_read_word_callback	mmu_read_word,arm7_read_word;
mmu_irq_callback		mmu_irq;
mmu_get_callback		mmu_get;
zombie_callback			zombie;

__int64					fifteen = 15;
__int64					masklo = 0x7BDE7BDE7BDE7BDE;  // 011110 11110 11110 -> 001111 01111 01111, 0011 1101 1110 1111
__int64					coladd = 0x35AD35AD35AD35AD;


__int64					fubar;

int						plus1=1,minus1=-1;
LARGE_INTEGER			perfFreq,perfCnt1,perfCnt2;

palette_dib				palDib;

bool useLighting[4];

// Used to copy the screen buffer to the window
struct bi {
	BITMAPINFOHEADER bih;
	RGBQUAD mask[4];
}bi;




void draw_text_bg(int);
void draw_rotscale_bg(int);
void draw_ext_rs_bg(int);
void draw_rs_bg(int);


// DLL entry point
bool __stdcall DllMain(HINSTANCE hDLL,DWORD reason,LPVOID unused) {
	if (reason==DLL_PROCESS_ATTACH)
		hInst = hDLL;

	return true;
}



void kill_gl() {
	if (hrc!=NULL) {
		wglMakeCurrent(NULL,NULL);
		wglDeleteContext(hrc);
		glDeleteTextures(TEXCACHE_SIZE+2,&texID[0]);
	}
	glIsInit = false;
	glOK = false;
}


bool setup_gl(HDC hDC) {
	int i;

	hrc = NULL;
	glIsInit = true;
	glOK = true;

	// Setup OpenGL
	pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER | PFD_DRAW_TO_WINDOW;
	pfd.dwLayerMask = PFD_MAIN_PLANE;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 16;
	pfd.cDepthBits = 16;
	pfd.cAccumBits = 0;
	pfd.cStencilBits = 0;
	if ((pixformat = ChoosePixelFormat(hDC,&pfd))==0)
		return false;

	SetPixelFormat(hDC,pixformat,&pfd);
	hrc = wglCreateContext(hDC);
	if (wglMakeCurrent(hDC,hrc)==FALSE)
		return false;

	glDisable(GL_CULL_FACE);
	glDisable(GL_DEPTH_TEST);

	glGenTextures(TEXCACHE_SIZE+2,&texID[0]);
	for (i=0; i<TEXCACHE_SIZE; i++)
		textureCache[i].id = texID[i+2];

	glBindTexture(GL_TEXTURE_2D,texID[0]);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB5_A1, 256, 512, 0, GL_RGBA, glFormat, vscreen);
	if (glGetError()!=GL_NO_ERROR)
		return false;

	unsigned char *temp = (unsigned char*)LocalAlloc(LPTR,256*256*4);
	glBindTexture(GL_TEXTURE_2D,texID[1]);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 256, 256, 0, GL_RGBA, GL_UNSIGNED_BYTE, temp);
	if (glGetError()!=GL_NO_ERROR)
		return false;
	LocalFree(temp);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter);
  	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter);

	glBindTexture(GL_TEXTURE_2D,texID[0]);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter);
  	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter);

	return true;
}



bool __stdcall gpu_init(zombie_callback commlink) {
	int i;
	zombie = commlink;

	zombie(TELL,ZOMBIE_DEBUG_MSG,"GPU init",0);

	scanline = 0;
	width = 256;
	visibleLines = 192;
	totalLines = 263;
	nameTblHeight = 240;
	spriteLimit = 128;
	frameInterval = 1090;
	glFormat = GL_UNSIGNED_SHORT_5_5_5_1;

	// Check for available CPU extensions
	__asm {
		pusha
		mov eax,1
		cpuid
		mov [i],edx
		popa
	}
	hasMMX  = ((i&0x0800000)!=0);
	hasSSE  = ((i&0x2000000)!=0);
	hasSSE2 = ((i&0x4000000)!=0);
	hasCMOV = ((i&0x0008000)!=0);
	//hasSSE2 = false;
	if ((!hasMMX)||(!hasCMOV)) {
		lpszError = szMMXError;
		return false;
	}

	hblankEnabled = true;
	vblankEnabled = true;
	emulate3D = true;

	VRAM = (unsigned char*)LocalAlloc(LPTR,0xA4000 + 0x8000 + 0x8000);

	VRAM_A = VRAM;
	VRAM_B = VRAM_A + 0x20000;
	VRAM_C = VRAM_B + 0x20000;
	VRAM_D = VRAM_C + 0x20000;
	VRAM_E = VRAM_D + 0x20000;
	VRAM_F = VRAM_E + 0x10000;
	VRAM_G = VRAM_F + 0x4000;
	VRAM_H = VRAM_G + 0x4000;
	VRAM_I = VRAM_H + 0x8000;
	vramBase[0] = VRAM_A;
	vramBase[1] = VRAM_B;
	vramBase[2] = VRAM_C;
	vramBase[3] = VRAM_D;
	vramBase[4] = VRAM_E;
	vramBase[5] = VRAM_F;
	vramBase[6] = VRAM_G;
	vramBase[7] = VRAM_H;
	vramBase[8] = VRAM_I;

	vscreen_size = width * visibleLines * 2;

	// Allocate four virtual screens and align them on 16-byte boundaries
	vscreenmem = (unsigned char*)LocalAlloc(LPTR,vscreen_size*8 + 16);
	vscreen = (unsigned char*)(((unsigned int)vscreenmem & 0xFFFFFFF0)+0x10);
	vscreen_s = vscreen + vscreen_size;
	vscreen2 = vscreen_s + vscreen_size;
	vscreen2_s = vscreen2 + vscreen_size;

	framebuffer = (unsigned char*)LocalAlloc(LPTR, 256*384*3);


	if ((vscreenmem == NULL)||(VRAM==NULL)||(framebuffer==NULL))
	{
		lpszError = szMemoryError;
		return false;
	}


	colorLUT = (short*)&VRAM[PALETTE_OFFSET];

	log2_lut[1] = 0;
	log2_lut[2] = 1;
	log2_lut[4] = 2;
	log2_lut[8] = 3;
	log2_lut[16] = 4;

	draw_bg_callbacks[0][0] = draw_text_bg;
	draw_bg_callbacks[0][1] = draw_text_bg;
	draw_bg_callbacks[0][2] = draw_text_bg;
	draw_bg_callbacks[0][3] = draw_text_bg;

	draw_bg_callbacks[1][0] = draw_text_bg;
	draw_bg_callbacks[1][1] = draw_text_bg;
	draw_bg_callbacks[1][2] = draw_text_bg;
	draw_bg_callbacks[1][3] = draw_rs_bg;

	draw_bg_callbacks[2][0] = draw_text_bg;
	draw_bg_callbacks[2][1] = draw_text_bg;
	draw_bg_callbacks[2][2] = draw_rotscale_bg;
	draw_bg_callbacks[2][3] = draw_rotscale_bg;

	draw_bg_callbacks[3][0] = draw_text_bg;
	draw_bg_callbacks[3][1] = draw_text_bg;
	draw_bg_callbacks[3][2] = draw_text_bg;
	draw_bg_callbacks[3][3] = draw_ext_rs_bg;

	draw_bg_callbacks[4][0] = draw_text_bg;
	draw_bg_callbacks[4][1] = draw_text_bg;
	draw_bg_callbacks[4][2] = draw_rotscale_bg;
	draw_bg_callbacks[4][3] = draw_ext_rs_bg;

	draw_bg_callbacks[5][0] = draw_text_bg;
	draw_bg_callbacks[5][1] = draw_text_bg;
	draw_bg_callbacks[5][2] = draw_ext_rs_bg;
	draw_bg_callbacks[5][3] = draw_ext_rs_bg;


	mtxStack[0] = new CMatrixStack(1);
	mtxStack[2] = new CMatrixStack(32);
	mtxStack[3] = new CMatrixStack(8);

	// Get a handle to STDOUT (for WriteConsole)
	hout = GetStdHandle(STD_OUTPUT_HANDLE);

	// Ask for the handle of the video output window
	hwnd = (HWND)zombie(ASK,ZOMBIE_WINDOW_HANDLE,(void*)2,0);
	dc = GetDC(hwnd);

	displayScreens = 3;		// Display both screens

	useGL = true;
	useD3D = false;

	glFormatBackup = glFormat;

	setup_gl(dc);
	glViewport(0,0,256,192+192);
	glClear(GL_COLOR_BUFFER_BIT);

	if (!create_d3d_device(hwnd,256,384,15,false))
		return true;
	if (lpD3DDevice->CreateTexture(256, 512, 0, 0, D3DFMT_X1R5G5B5, D3DPOOL_MANAGED, &lpTexture) != D3D_OK)
		if (lpD3DDevice->CreateTexture(256, 512, 0, 0, D3DFMT_A1R5G5B5, D3DPOOL_MANAGED, &lpTexture) != D3D_OK)
			return true;

	lpTexture->GetLevelDesc(0,&d3ddesc);
	if ((d3ddesc.Format != D3DFMT_X1R5G5B5)&&
		(d3ddesc.Format != D3DFMT_A1R5G5B5))
		return true;

	if (lpD3DDevice->CreateVertexBuffer(12*sizeof(customVertex),0,
                                    D3DFVF_XYZRHW + D3DFVF_TEX1,
                                    D3DPOOL_MANAGED, &lpVertexBuf1) != D3D_OK)
	{
		//wsprintf(buffer,"Failed to create vertex buffer\n");
		//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
		return true;
	}
	lpD3DDevice->CreateVertexBuffer(12*sizeof(customVertex),0,
                                    D3DFVF_XYZRHW + D3DFVF_TEX1,
                                    D3DPOOL_MANAGED, &lpVertexBuf2);

	// Lock the vertex buffer.
	lpVertexBuf1->Lock(0, 0, (BYTE**)&lpVertexBufferData, NULL);
	memcpy(lpVertexBufferData,&mainCore_topScreen,12*sizeof(customVertex));
	lpVertexBuf1->Unlock();

	// Lock the vertex buffer.
	if (lpVertexBuf2->Lock(0, 0, (BYTE**)&lpVertexBufferData, NULL) != D3D_OK)
	{
		//wsprintf(buffer,"Failed to lock vertex buffer\n");
		//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
		return true;
	}
	memcpy(lpVertexBufferData,&mainCore_bottomScreen,12*sizeof(customVertex));
	lpVertexBuf2->Unlock();

	// Set our created texture to the current one.
	lpD3DDevice->SetTexture(0, lpTexture);

	lpD3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

	// Set up the texture parameters.
	lpD3DDevice->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_POINT);
	lpD3DDevice->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
	lpD3DDevice->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_SELECTARG1 );
	lpD3DDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );

	lpD3DDevice->Clear(0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,0xFFFFFFFF,1.0f,0);
	lpD3DDevice->SetVertexShader(D3DFVF_XYZRHW + D3DFVF_TEX1);

	d3dOK = true;

	return true;
}



int dsVPX1=0,dsVPY1=0,dsVPX2=256,dsVPY2=192;	// Viewpoint

bool __stdcall gpu_reset() {
	int i;

	zombie(TELL,ZOMBIE_DEBUG_MSG,"GPU reset",0L);

	// Ask for the CPU IRQ callback
	cpu_irq = (cpu_irq_callback)zombie(ASK,CPU_IRQ_CALLBACK,0L,0L);
	arm7_irq = (cpu_irq_callback)zombie(ASK,CPU_IRQ_CALLBACK,0L,(void*)1L);
	mmu_irq = (mmu_irq_callback)zombie(ASK,MMU_IRQ_CALLBACK,0L,0L);
	mmu_get = (mmu_get_callback)zombie(ASK,MMU_GET_CALLBACK,0L,0L);
	mmu_read_word = (mmu_read_word_callback)zombie(ASK,MMU_READ_WORD_CALLBACK,0L,0L);
	arm7_read_word = (mmu_read_word_callback)zombie(ASK,MMU_READ_WORD_CALLBACK,(void*)1L,0L);
	vramBank = (vram_bank*)zombie(ASK,MMU_VRAM_BANKS,0L,0L);

	commandListPos = 0;

	//memset(VRAM,0,0xE0000);
	memset(VRAM,0,0xA4000 + 0x8000 + 0x8000); //0x88000+0x40000+0x20000+0xA0000+0x20000);
	memset(vscreen,0xFF,vscreen_size*4);

	for (i=0; i<128; i++) {
		VRAM[SPRITE_TABLE_OFFSET + (i<<3)] = 0;//192;
		VRAM[SPRITE_TABLE_OFFSET + 0x400 + (i<<3)] = 0;//192;
	}

	VRAM[IOREGS_OFFSET+4]     = 0;
	VRAM[IOREGS_OFFSET+6]     = 0;
	VRAM[IOREGS_OFFSET+7]     = 0;
	VRAM[IOREGS_OFFSET+0x130] = 0xFF;

	VRAM[IOREGS_ARM7_OFFSET+4]     = 0;
	VRAM[IOREGS_ARM7_OFFSET+6]     = 0;
	VRAM[IOREGS_ARM7_OFFSET+7]     = 0;
	VRAM[IOREGS_ARM7_OFFSET+0x130] = 0xFF;

	IOREGS      = &VRAM[IOREGS_OFFSET];
	IOREGS_ARM7 = &VRAM[IOREGS_ARM7_OFFSET];

	IOREGS[0] = IOREGS[1] = 0;
	IOREGS[0x1000] = IOREGS[0x1001] = 0;
	IOREGS[0x50] = IOREGS[0x51] = 0;
	IOREGS[0x1050] = IOREGS[0x1051] = 0;

	IOREGS[8] = 0;
	IOREGS[10] = 0;
	IOREGS[12] = 0;
	IOREGS[14] = 0;

	*(int*)(IOREGS + 0x10) = 0;
	*(int*)(IOREGS + 0x14) = 0;
	*(int*)(IOREGS + 0x18) = 0;
	*(int*)(IOREGS + 0x1C) = 0;
	*(int*)(IOREGS + 0x1010) = 0;
	*(int*)(IOREGS + 0x1014) = 0;
	*(int*)(IOREGS + 0x1018) = 0;
	*(int*)(IOREGS + 0x101C) = 0;

	*(int*)(IOREGS + 0x20) = 0x100;
	*(int*)(IOREGS + 0x24) = 0x1000000;
	*(int*)(IOREGS + 0x30) = 0x100;
	*(int*)(IOREGS + 0x34) = 0x1000000;

	for (i=0; i<5; i++)	{
		layers[i].num = i;
		layers[i].prio = i;
	}

    *((int*)&bi.mask[0]) = 0x001F; 	// 00000000 000rrrrr
    *((int*)&bi.mask[1]) = 0x03E0; 	// 000000gg ggg00000
    *((int*)&bi.mask[2]) = 0x7C00;	// 0bbbbb00 00000000

	scanline = 0;
	frames = 0;
	newframe = 0;
	lastbg = -1;

	/* Call glEnd in case emulation was rest in the middle of a
	   VTX_BEGIN/END */
	glEnd();
	while (glGetError()!=GL_NO_ERROR) {}

	/* Reset texture cache */
	for (i=0; i<TEXCACHE_SIZE; i++)
		textureCache[i].fmt = 0;
	activeTexture = 0;

	/* Reset projection matrices */
	glMatrixMode(GL_PROJECTION);
	for (i=0; i<32; i++)
		glPopMatrix();
	glLoadIdentity();

	/* Reset modelview matrices */
	glMatrixMode(GL_MODELVIEW);
	for (i=0; i<32; i++)
		glPopMatrix();
	glLoadIdentity();

	/* Reset lighting */
 	glDisable(GL_LIGHTING);
 	glDisable(GL_LIGHT0);
 	glDisable(GL_LIGHT1);
 	glDisable(GL_LIGHT2);
 	glDisable(GL_LIGHT3);
	glDisable(GL_CULL_FACE);
	useLighting[0] = false;
	useLighting[1] = false;
	useLighting[2] = false;
	useLighting[3] = false;
	lightAttr[0] = lightAttr[1] = lightAttr[2] = lightAttr[3] = false;

	glColor4f(1,1,1,1);
	polyalpha = 1.0f;
	vtx_begin = false;
	polymode = GL_FILL;
	block3Dcommands = false;

	GetClientRect(hwnd,&rect);
	glViewport(dsVPX1,dsVPY1,dsVPX2,dsVPY2); 

	while (glGetError()!=GL_NO_ERROR) {}

	mtxStack[0]->flush();
	mtxStack[2]->flush();
	mtxStack[3]->flush();

	QueryPerformanceFrequency(&perfFreq);
	QueryPerformanceCounter(&perfCnt1);
	return true;
}


void __stdcall gpu_close() {
	destroy_d3d_device();
	kill_gl();
	LocalFree(vscreenmem);
	LocalFree(VRAM);
	LocalFree(framebuffer);
	delete mtxStack[0];
	delete mtxStack[2];
	delete mtxStack[3];
}


int __stdcall gpu_config(HWND hwnd) {
	return DialogBoxParam(hInst,
           MAKEINTRESOURCE(IDD_DIALOG1), hwnd,
           DlgProc, NULL);
}





void bgr555_to_bgr888(char *src, char *dest,int width,int height,bool flip) {
	int x,y,srcinc;
	unsigned short rgb555;

	srcinc = width*2;

	// Flip the image, for DIB compability
	if (flip) {
		src += (height-1)*width*2;
		srcinc = -srcinc;
	}

	// Yes, this sucks
	for (y=0; y<height; y++) {
		for (x=0; x<width<<1; x+=2)	{
			rgb555 = *(unsigned short*)(&src[x]);
			*(dest++) = ((rgb555>>7)&0xf8) + (rgb555>>13);
			*(dest++) = ((rgb555>>2)&0xf8) + ((rgb555>>8)&3);
			*(dest++) = ((rgb555<<3)&0xf8) + ((rgb555>>3)&3);
		}
		src += srcinc;
	}
}


void bgra1555_to_bgra8888(char *src, char *dest,int width,int height,bool flip) {
	int x,y,srcinc;
	unsigned short rgb555;

	srcinc = width*2;

	// Flip the image, for DIB compability
	if (flip) {
		src += (height-1)*width*2;
		srcinc = -srcinc;
	}

	// Yes, this sucks
	for (y=0; y<height; y++) {
		for (x=0; x<width<<1; x+=2)	{
			rgb555 = *(unsigned short*)(&src[x]);
			*(dest++) = ((rgb555>>7)&0xf8) + (rgb555>>13);
			*(dest++) = ((rgb555>>2)&0xf8) + ((rgb555>>8)&3);
			*(dest++) = ((rgb555<<3)&0xf8) + ((rgb555>>3)&3);
			*(dest++) = (rgb555&0x8000)?0xff:0x00;
		}
		src += srcinc;
	}
}


void rgb555_to_bgr888(char *src,char *dest,int width,int height,bool flip) {
	int x,y,srcinc;
	unsigned short rgb555;

	srcinc = width*2;

	// Flip the image, for DIB compability
	if (flip) {
		src += (height-1)*width*2;
		srcinc = -srcinc;
	}

	// Yes, this sucks
	for (y=0; y<height; y++)
	{
		for (x=0; x<width<<1; x+=2)
		{
			rgb555 = *(unsigned short*)(&src[x]);
			*(dest++) = ((rgb555<<3)&0xf8) + ((rgb555>>3)&3);
			*(dest++) = ((rgb555>>2)&0xf8) + ((rgb555>>8)&3);
			*(dest++) = ((rgb555>>7)&0xf8) + (rgb555>>13);
		}
		src += srcinc;
	}
}


void rgba1555_to_bgra8888(char *src,char *dest,int width,int height,bool flip) {
	int x,y,srcinc;
	unsigned short rgb555;

	srcinc = width*2;

	// Flip the image, for DIB compability
	if (flip) {
		src += (height-1)*width*2;
		srcinc = -srcinc;
	}

	// Yes, this sucks
	for (y=0; y<height; y++) {
		for (x=0; x<width<<1; x+=2)	{
			rgb555 = *(unsigned short*)(&src[x]);
			if (rgb555&0x8000) {
				*(dest++) = ((rgb555<<3)&0xf8) + ((rgb555>>3)&3);
				*(dest++) = ((rgb555>>2)&0xf8) + ((rgb555>>8)&3);
				*(dest++) = ((rgb555>>7)&0xf8) + (rgb555>>13);
				*(dest++) = 0xff;
			} else {
				*(int*)dest = 0;
				dest += 4;
			}
		}
		src += srcinc;
	}
}


void rgba1555_to_bgra8888_semitrans(char *src,char *dest,int width,int height,bool flip) {
	int x,y,srcinc;
	unsigned short rgb555;

	srcinc = width*2;

	// Flip the image, for DIB compability
	if (flip) {
		src += (height-1)*width*2;
		srcinc = -srcinc;
	}

	// Yes, this sucks
	for (y=0; y<height; y++) {
		for (x=0; x<width<<1; x+=2)	{
			rgb555 = *(unsigned short*)(&src[x]);
			if ( ((rgb555&0x8000)!=0) && ((rgb555&0x7FFF)!=0) ) {
				*(dest++) = ((rgb555<<3)&0xf8) + ((rgb555>>3)&3);
				*(dest++) = ((rgb555>>2)&0xf8) + ((rgb555>>8)&3);
				*(dest++) = ((rgb555>>7)&0xf8) + (rgb555>>13);
				*(dest++) = 0xff;
			} else {
				*(int*)dest = 0;
				dest += 4;
			}
		}
		src += srcinc;
	}
}


int __stdcall gpu_get(int what) {
	int i,*gxstat=(int*)(VRAM+IOREGS_OFFSET+0x600);
	unsigned char *ptbl;

	switch (what) {
		case GPU_3D_COMMAND:
			/*for (i=0; i<3; i++)
			{
				*gxstat &= ~mtx_stat[i][2];
				*gxstat |= mtx_stat[i][0] << mtx_stat[i][1];
			}*/
			*gxstat &= ~((1<<13)|(31<<8));
			*gxstat |= (mtxStack[0]->depth<<13)|(mtxStack[2]->depth<<8);
			//*gxstat |= 0x6000000;
			break;

		case GPU_RENDERER_OK:
			return ((int)glOK)|((int)d3dOK<<1);

		case GPU_CURRENT_SCANLINE:
			return scanline;

		case GPU_SCREEN_WIDTH:
			return width;

		case GPU_SCREEN_HEIGHT:
			return visibleLines;

		case GPU_VBLANK_LINES:
			return totalLines-visibleLines;

		case GPU_HBLANK_PIXELS:
			return 68;

		case GPU_NUMBER_COLORS:
			return 256;

		// Get the current palette as a DIB (both cores)
		case GPU_PALETTE:
			palDib.data = (char*)&VRAM[PALETTE_OFFSET];
			memset(&palDib.bih,0,sizeof(BITMAPINFOHEADER));
 			palDib.bih.biSize = sizeof(BITMAPINFOHEADER);
 			palDib.bih.biWidth = 16;
 			palDib.bih.biHeight = -64;
 			palDib.bih.biPlanes = 1;
 			palDib.bih.biBitCount = 16;
  			palDib.bih.biCompression = BI_BITFIELDS;

			*((int*)&palDib.colors[0]) = 0x001F;	// 00000000 000rrrrr
			*((int*)&palDib.colors[1]) = 0x03E0;	// 000000gg ggg00000
			*((int*)&palDib.colors[2]) = 0x7C00;	// 0bbbbb00 00000000

			return (int)&palDib;

		case GPU_FRAMEBUFFER_MAIN:
			return (int)vscreen;

		// Get the current framebuffer (both cores)
		case GPU_FRAMEBUFFER_BGR24:
			if (useGL) {
				if (!vtx_begin) {
					glGetIntegerv(GL_READ_BUFFER,(GLint*)&i);
					glReadBuffer(GL_BACK);
					glReadPixels(0,0,256,384,GL_BGR_EXT,GL_UNSIGNED_BYTE,framebuffer);
					glReadBuffer((GLenum)i);
					return (int)framebuffer;
				} else {
					return 0;
				}
			} else if (useD3D)
			{
				rgb555_to_bgr888((char*)vscreen,(char*)framebuffer,256,384,true);
				return (int)framebuffer;
			} else
			{
				bgr555_to_bgr888((char*)vscreen,(char*)framebuffer,256,384,true);
				return (int)framebuffer;
			}
			return 0;
			break;

		// Get the current framebuffer (main core only)
		case GPU_FRAMEBUFFER_MAIN_BGR24:
			if (useGL) {
				if (!vtx_begin) {
					glGetIntegerv(GL_READ_BUFFER,(GLint*)&i);
					glReadBuffer(GL_BACK);
					if ((VRAM[IOREGS_OFFSET + 0x304 + 1]&0x80) == 0)
						glReadPixels(0,0,256,192,GL_BGR_EXT,GL_UNSIGNED_BYTE,framebuffer);
					else
						glReadPixels(0,192,256,192,GL_BGR_EXT,GL_UNSIGNED_BYTE,framebuffer);
					glReadBuffer((GLenum)i);
					return (int)framebuffer;
				} else {
					return 0;
				}
			} else if (useD3D) {
				if ((VRAM[IOREGS_OFFSET + 0x304 + 1]&0x80) == 0)
					rgb555_to_bgr888((char*)(vscreen+192*256*2),(char*)framebuffer,256,192,true);
				else
					rgb555_to_bgr888((char*)vscreen,(char*)framebuffer,256,192,true);
				return (int)framebuffer;
			} else {
				if ((VRAM[IOREGS_OFFSET + 0x304 + 1]&0x80) == 0)
					bgr555_to_bgr888((char*)(vscreen+192*256*2),(char*)framebuffer,256,192,true);
				else
					bgr555_to_bgr888((char*)vscreen,(char*)framebuffer,256,192,true);
				return (int)framebuffer;
			}
			return 0;
			break;

		// Get the current framebuffer (sub core only)
		case GPU_FRAMEBUFFER_SUB_BGR24:
			if (useGL) {
				glReadBuffer(GL_BACK);
				if ((VRAM[IOREGS_OFFSET + 0x304 + 1]&0x80) == 0)
					glReadPixels(0,192,256,192,GL_BGR_EXT,GL_UNSIGNED_BYTE,framebuffer);
				else
					glReadPixels(0,0,256,192,GL_BGR_EXT,GL_UNSIGNED_BYTE,framebuffer);
				glReadBuffer(GL_FRONT);
				return (int)framebuffer;
			} else if (useD3D) {
				if ((VRAM[IOREGS_OFFSET + 0x304 + 1]&0x80) != 0)
					rgb555_to_bgr888((char*)(vscreen+192*256*2),(char*)framebuffer,256,192,true);
				else
					rgb555_to_bgr888((char*)vscreen,(char*)framebuffer,256,192,true);
				return (int)framebuffer;
			} else {
				if ((VRAM[IOREGS_OFFSET + 0x304 + 1]&0x80) != 0)
					bgr555_to_bgr888((char*)(vscreen+192*256*2),(char*)framebuffer,256,192,true);
				else
					bgr555_to_bgr888((char*)vscreen,(char*)framebuffer,256,192,true);
				return (int)framebuffer;
			}
			return 0;
			break;

		case GPU_MEMORY:
			wsprintf(buffer,"GPU: Returning VRAM pointer (0x%08X)",VRAM);
			zombie(TELL,ZOMBIE_DEBUG_MSG,buffer,0);
			return (int)VRAM;

		case GPU_MAIN_OBJ:
		case GPU_SUB_OBJ:
			// Grab the first correctly setup bank
			for (i=0; i<9; i++)	{
				if (vramBank[i].type == MAIN_OBJ+(what-GPU_MAIN_OBJ))
					return (int)vramBase[i];
			}
			return -1;

		case PLUGIN_NAME:
			return (int)&gpuName;

		case PLUGIN_VERSION:
			return GPU_VERSION;

		case LAST_ERROR:
			return (int)lpszError;

		// Debug stuff
		case 0x700:
			//wsprintf(buffer,"%d, %d, %d, %d, %d, %d\n",objx2,objy2,midpoint[0],midpoint[1],spriteDbg[1],objdy);
			wsprintf(buffer,"x=%d, y=%d, pa-pd = %d, %d, %d, %d\n",xtile,hofs,objdx,objdy,objdmx,objdmy);
			//wsprintf(buffer,"%d, %d, %d\n",spriteDbg[0],spriteDbg[1],spriteDbg[2]);
			//wsprintf(buffer,"%d, %d, %d\n",hTileMask,vTileMask,tileShift);
			//wsprintf(buffer,"x=%d, y=%d, pa-pd = %d, %d, %d, %d\n",x2,y2,BG3PA,BG3PB,BG3PC,BG3PD);
			//wsprintf(buffer,"Blended 8 pixels in %d clocks (%d)\n",(unsigned int)(fubar&0xFFFFFFFF),blended2);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			return (int)&buffer[0];
			break;

		default:
			break;
	}

	return 0;
}


unsigned char *get_texture_palette_base() {
	for (int i=4; i<=6; i++) {
		if (vramBank[i].type == TEXPLTT) {
			//wsprintf(buffer,"Using bank %d for texture palettes\n",i),
			//WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);
			return vramBase[i];
		}
	}
	return NULL;
}


bool check_gl_error(char *cp) {
	GLenum errCode;
	errCode = glGetError();
	switch (errCode) {
		case GL_NO_ERROR:
			return false;
		case GL_INVALID_ENUM:
			wsprintf(buffer,"GL_INVALID_ENUM (%s)",cp);
			break;
		case GL_INVALID_VALUE:
			wsprintf(buffer,"GL_INVALID_VALUE (%s)",cp);
			break;
		case GL_INVALID_OPERATION:
			wsprintf(buffer,"GL_INVALID_OPERATION (%s)",cp);
			break;
		case GL_STACK_OVERFLOW:
			wsprintf(buffer,"GL_STACK_OVERFLOW (%s)",cp);
			break;
		case GL_STACK_UNDERFLOW:
			wsprintf(buffer,"GL_STACK_UNDERFLOW (%s)",cp);
			break;
		case GL_OUT_OF_MEMORY:
			wsprintf(buffer,"GL_OUT_OF_MEMORY (%s)",cp);
			break;
		default:
			wsprintf(buffer,"Unknown error (%s)",cp);
			break;

	}
	MessageBox(NULL,buffer,"OpenGL error",MB_OK|MB_ICONERROR);
	return true;
}


UINT texPalOfs;

void set_ogl_texture(int fmt) {
	int i,j,u,v,offset;
	bool reUse = false;
	unsigned short *palData;
	unsigned char c,*p;
	
	//wsprintf(buffer,"Set texture: fmt=%d, texofs=%X\n",fmt,texPalOfs);
	//WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);

	if (((fmt>>26)&7) == 0) {
		activeTexture = -1;
		return;
	}
	
	for (i=0; i<TEXCACHE_SIZE; i++) {
		if ((textureCache[i].fmt == fmt) &&
			(textureCache[i].numcolors == 32768)) {	// Check if this texture already exists
			activeTexture = i;
			glBindTexture(GL_TEXTURE_2D,textureCache[i].id);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,
					((fmt&0x10000)?GL_REPEAT:GL_CLAMP));
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, 
					((fmt&0x20000)?GL_REPEAT:GL_CLAMP));
			return;
		}
		if (textureCache[i].fmt == 0)		// Do we have an unused cache slot ?
			break;
	}

	if (i == TEXCACHE_SIZE) {		// All slots taken
		for (j=0; j<TEXCACHE_SIZE; j++) {
			if ((textureCache[j].color  == ((fmt>>26)&7)) &&
				(textureCache[j].width  == (8 << ((fmt>>20)&7))) &&
				(textureCache[j].height == (8 << ((fmt>>23)&7)))) {
				reUse = true;
				i = j;
				break;
			}
		}
		if (j == TEXCACHE_SIZE) {
			//return;
			//glDeleteTextures(1,&textureCache[0].id);
			//glGenTextures(1,&textureCache[0].id);
			i = 2;
		}
	}
	
	glBindTexture(GL_TEXTURE_2D,textureCache[i].id);

	if (!reUse) {
		textureCache[i].width  = 8 << ((fmt>>20)&7);
		textureCache[i].height = 8 << ((fmt>>23)&7);
		textureCache[i].color = ((fmt>>26)&7);
		if (textureCache[i].color==1)
			textureCache[i].numcolors = 32;
		else if (textureCache[i].color==2)
			textureCache[i].numcolors = 4;
		else if (textureCache[i].color==3)
			textureCache[i].numcolors = 16;
		else if (textureCache[i].color==4)
			textureCache[i].numcolors = 256;
		else if (textureCache[i].color==6)
			textureCache[i].numcolors = 8;
		else if (textureCache[i].color==7)
			textureCache[i].numcolors = 32768;
		else
			return;
	}

	textureCache[i].fmt = fmt;
	
	activeTexture = i;


	if (textureCache[i].numcolors == 32768) {
		if (fmt & (1<<29)) 
			rgba1555_to_bgra8888_semitrans((char*)vramBase[0] + (fmt&0xFFFF)*8,
							 (char*)framebuffer,
							 textureCache[i].width,textureCache[i].height,
							 false);
		else
			rgba1555_to_bgra8888((char*)vramBase[0] + (fmt&0xFFFF)*8,
							 (char*)framebuffer,
							 textureCache[i].width,textureCache[i].height,
							 false);

		if (reUse) {
			glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
						  textureCache[i].width, textureCache[i].height,
						  GL_RGBA, GL_UNSIGNED_BYTE, framebuffer);
		} else {
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB5_A1,
						  textureCache[i].width, textureCache[i].height,
						  0, GL_RGBA, GL_UNSIGNED_BYTE, framebuffer);
		}
	} else if (textureCache[i].numcolors == 4) {
		offset = 0;
		p = &framebuffer[0];
		palData = (unsigned short*)(get_texture_palette_base() + texPalOfs*8);
		if ((int)palData == texPalOfs*8) return;

		for (v=0; v<textureCache[i].height; v++) {
			for (u=0; u<textureCache[i].width; u++) {
				c = ((*(unsigned char*)(vramBase[0] + (fmt&0xFFFF)*8 + (offset>>2))) >> ((u&3)<<1)) & 3;
				*(p++) = ((palData[c])&0x1F)*8; 
				*(p++) = ((palData[c]>>5)&0x1F)*8; 
				*(p++) = ((palData[c]>>10)&0x1F)*8;
				*(p++) = 0xff;
				offset++;
			}
		}
		if (reUse) {
			glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
						  textureCache[i].width, textureCache[i].height,
						  GL_RGBA, GL_UNSIGNED_BYTE, framebuffer);
		} else {
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB5_A1,
						  textureCache[i].width, textureCache[i].height,
						  0, GL_RGBA, GL_UNSIGNED_BYTE, framebuffer);
		}

	} else if (textureCache[i].numcolors == 8) {
		offset = 0;
		p = &framebuffer[0];
		palData = (unsigned short*)(get_texture_palette_base() + texPalOfs*16);
		if ((int)palData == texPalOfs*16) return;

		for (v=0; v<textureCache[i].height; v++) {
			for (u=0; u<textureCache[i].width; u++) {
				c = *(unsigned char*)(vramBase[0] + (fmt&0xFFFF)*8 + offset);
				*(p++) = ((palData[c&7])&0x1F)*8; 
				*(p++) = ((palData[c&7]>>5)&0x1F)*8; 
				*(p++) = ((palData[c&7]>>10)&0x1F)*8;
				*(p++) = ((c&0xF8)==0xF8)?0xff:(c&0xF8);
				offset++;
			}
		}
		if (reUse) {
			glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
						  textureCache[i].width, textureCache[i].height,
						  GL_RGBA, GL_UNSIGNED_BYTE, framebuffer);
		} else {
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB5_A1,
						  textureCache[i].width, textureCache[i].height,
						  0, GL_RGBA, GL_UNSIGNED_BYTE, framebuffer);
		}

	} else if (textureCache[i].numcolors == 16) {
		offset = 0;
		p = &framebuffer[0];
		palData = (unsigned short*)(get_texture_palette_base() + texPalOfs*16);
		if ((int)palData == texPalOfs*16) return;

		for (v=0; v<textureCache[i].height; v++) {
			for (u=0; u<textureCache[i].width; u++) {
				c = ((*(unsigned char*)(vramBase[0] + (fmt&0xFFFF)*8 + (offset>>1))) >> ((u&1)<<2)) & 15;
				if ((fmt&(1<<29))&&(c==0)) {
		//wsprintf(buffer,"Transparent pixel, reuse=%d\n",reUse);
		//WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);
					*(int*)p = 0;
					p += 4;
				} else {
					*(p++) = ((palData[c])&0x1F)*8; 
					*(p++) = ((palData[c]>>5)&0x1F)*8; 
					*(p++) = ((palData[c]>>10)&0x1F)*8;
					*(p++) = 0xff;
				}
				offset++;
			}
		}

		/*for (v=0; v<textureCache[i].height; v++) {
			for (u=0; u<textureCache[i].width; u++) {
				framebuffer[offset] =  
					((*(char*)(vramBase[0] + (fmt&0xFFFF)*8 + (offset>>1))) >> ((u&1)<<2)) & 15;
				offset++;
			}
		}*/
		if (reUse) {
			glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
						  textureCache[i].width, textureCache[i].height,
						  GL_RGBA, GL_UNSIGNED_BYTE, framebuffer);
		} else {
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB5_A1,
						  textureCache[i].width, textureCache[i].height,
						  0, GL_RGBA, GL_UNSIGNED_BYTE, framebuffer);
		}

	} else if (textureCache[i].numcolors == 32) {
		offset = 0;
		p = &framebuffer[0];
		palData = (unsigned short*)(get_texture_palette_base() + texPalOfs*16);
		if ((int)palData == texPalOfs*16) return;

		for (v=0; v<textureCache[i].height; v++) {
			for (u=0; u<textureCache[i].width; u++) {
				c = *(unsigned char*)(vramBase[0] + (fmt&0xFFFF)*8 + offset);
				*(p++) = ((palData[c&0x1F])&0x1F)*8; 
				*(p++) = ((palData[c&0x1F]>>5)&0x1F)*8; 
				*(p++) = ((palData[c&0x1F]>>10)&0x1F)*8;
				*(p++) = ((c&0xE0)==0xE0)?0xff:(c&0xE0);
				offset++;
			}
		}
		if (reUse) {
			glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
						  textureCache[i].width, textureCache[i].height,
						  GL_RGBA, GL_UNSIGNED_BYTE, framebuffer);
		} else {
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB5_A1,
						  textureCache[i].width, textureCache[i].height,
						  0, GL_RGBA, GL_UNSIGNED_BYTE, framebuffer);
		}

	} else if (textureCache[i].numcolors == 256) {
		offset = 0;
		p = &framebuffer[0];
		palData = (unsigned short*)(get_texture_palette_base() + texPalOfs*16); //textureCache[activeTexture].numcolors*2);
		if ((int)palData == texPalOfs*16) return;

		for (v=0; v<textureCache[i].height; v++) {
			for (u=0; u<textureCache[i].width; u++) {
				c = *(unsigned char*)(vramBase[0] + (fmt&0xFFFF)*8 + offset);
				if ((fmt&(1<<29))&&(c==0)) {
					*(int*)p = 0;
					p += 4;
				} else {
					*(p++) = ((palData[c])&0x1F)*8; 
					*(p++) = ((palData[c]>>5)&0x1F)*8; 
					*(p++) = ((palData[c]>>10)&0x1F)*8;
					*(p++) = 0xff;
				}
				offset++;
			}
		}
		if (reUse) {
			glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
						  textureCache[i].width, textureCache[i].height,
						  GL_RGBA, GL_UNSIGNED_BYTE, framebuffer);
		} else {
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB5_A1,
						  textureCache[i].width, textureCache[i].height,
						  0, GL_RGBA, GL_UNSIGNED_BYTE, framebuffer);
		}
	}

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,
				((fmt&0x10000)?GL_REPEAT:GL_CLAMP));
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, 
				((fmt&0x20000)?GL_REPEAT:GL_CLAMP));
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

}



void process_3d_commands() {
	static bool zcoord = false;
	static bool pendingTexUpload=false;
	static int texfmt=0;
	static float x,y,z;
	static float trans[3],scale[3];
	static int tCount=0,sCount=0,mmCountC=0,mmCountR=0,mlCountC=0,mlCountR=0;
	float f,g,h;
	static float rot=0.0f;
	int i,j,k,cmd;
	unsigned char *temp;
	unsigned int parm;
	bool inVtxList;


	if (block3Dcommands) {
		for (i=0; i<commandListPos; i++) {
			if (commandList[i].cmd == 0x540)
				block3Dcommands = false;
		}
		return;
	}

	if (VRAM[IOREGS_OFFSET+0x603]&8)
		return;

	for (i=0; i<commandListPos; i++) {
		if (useGL) {
			//wsprintf(buffer,"3D command 0x%x = 0x%08X\n",commandList[i].cmd,commandList[i].parm),
			//WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);
			cmd = commandList[i].cmd;

			if (commandList[i].cmd==0x400) {
				//wsprintf(buffer,"3d fifo: %08X\n",commandList[i].parm);
				//WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);
				if (fifopos == -1) {
					fifolen = 4;
					j = 0;
					parm = commandList[i].parm;
					for (k=0; k<4; k++) {
						fifocommands[j].cmd = ((parm&0xFF)<<2) + 0x400;
						if (fifocommands[j].cmd == 0x48C) {
							fifocommands[j+1].cmd = fifocommands[j].cmd;
							j++;
						} else if ((fifocommands[j].cmd == 0x46C)||
						 	       (fifocommands[j].cmd == 0x470)) {
							fifocommands[j+1].cmd = fifocommands[j].cmd;
							fifocommands[j+2].cmd = fifocommands[j].cmd;
							j += 2;
						}
						j++;
						parm >>= 8;
					}
					fifolen = j;
					fifoparms = j;
					if (fifocommands[fifolen-1].cmd == 0x504)
						fifoparms--;
					fifopos = 0;
					return;
				} else {
					if (fifocommands[fifopos].cmd == 0x504)
						fifopos++;
					fifocommands[fifopos++].parm = commandList[i].parm;
					return;
					/*commandList[i].cmd = fifocommands[fifopos].cmd;
					wsprintf(buffer,"Pos %d, cmd %08X\n",fifopos,commandList[i].cmd);
					WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);*/

				}
			}

			if (!vtx_begin) { //(vtx_begin) {
				//if ((commandList[i].cmd != 0x488) && (commandList[i].cmd != 0x48C)) {
			//		wsprintf(buffer,"3d command: %X %08X\n",commandList[i].cmd&0xFFF,commandList[i].parm);
			//		WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);
				//}
			}

			switch(commandList[i].cmd) {
				/*case 0x400:
					wsprintf(buffer,"3d fifo: %08X\n",commandList[i].parm);
					WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);
					if (fifopos == -1) {
						fifocommands[0].cmd = ((commandList[i].parm&0xFF)<<2) + 0x400;
						fifocommands[1].cmd = (((commandList[i].parm>>8)&0xFF)<<2) + 0x400;
						fifocommands[2].cmd = (((commandList[i].parm>>16)&0xFF)<<2) + 0x400;
						fifocommands[3].cmd = (((commandList[i].parm>>24)&0xFF)<<2) + 0x400;
						break;
					} else {
						fifocommands[fifopos].parm = commandList[i].parm;
					}
					fifopos++;
					break;*/

				// MTX_MODE
				case 0x440:
					mtx_mode = commandList[i].parm&3;
					if (mtx_mode == 1) mtx_mode = 2;
					break;

				// MTX_PUSH
				case 0x444:
					mtxStack[mtx_mode]->push(&mtx_current[mtx_mode]);
					break;

				// MTX_POP
				case 0x448:
					mtxStack[mtx_mode]->pop(&mtx_current[mtx_mode],commandList[i].parm&0x1F);
					break;

				// MTX_STORE
				case 0x44C:
					mtxStack[mtx_mode]->store(&mtx_current[mtx_mode],commandList[i].parm);
					break;

				// MTX_RESTORE
				case 0x450:
					mtxStack[mtx_mode]->restore(&mtx_current[mtx_mode],commandList[i].parm);
					break;

				case 0x454:
					//WriteConsole(hout,"LoadIdentity\n",14,(LPDWORD)&dummy,NULL);
					//glLoadIdentity();
					memcpy(&mtx_current[mtx_mode],&identity_4x4,sizeof(t_matrix_4x4));
					break;

				// MTX_LOAD_4X4
				case 0x458:
					f = (float)commandList[i].parm / 4096.0f;
					if (f >= 524288.0f)
						f -= 1048576.0f;

					mtx_current[mtx_mode][mlCountR][mlCountC] = f;
					if ((++mlCountC)==4) {
						mlCountC = 0;
						if ((++mlCountR)==4) {
							mlCountR = 0;
					/*if ((mtx_mode==0)&&(framecnt<4))
					{
						WriteConsole(hout,"Mtx\n",5,(LPDWORD)&dummy,NULL);
						for (j=0; j<4; j++)
						{
							wsprintf(buffer,"[%d %d %d %d]\n",(int)(mtx_current[0][j][0]*100.0f),
								(int)(mtx_current[0][j][1]*100.0f),
								(int)(mtx_current[0][j][2]*100.0f),
								(int)(mtx_current[0][j][3]*100.0f));

							WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);
						}
						WriteConsole(hout,"\n",2,(LPDWORD)&dummy,NULL);
					}	*/		
						}
					}
					break;

				// MTX_LOAD_4X3
				case 0x45C:
					f = (float)commandList[i].parm / 4096.0f;
					if (f >= 524288.0f) f -= 1048576.0f;

					//mtx_4x4[mlCountR][mlCountR] = f;
					mtx_current[mtx_mode][mlCountR][mlCountC] = f;
					if ((++mlCountC)==3) {
						mtx_current[mtx_mode][mlCountR][mlCountC] = 
							identity_4x4[mlCountR][mlCountC];
						mlCountC = 0;
						if ((++mlCountR)==4) {
							mlCountR = 0;
							/*mtx_current[mtx_mode][3][0] = 0.0f;
							mtx_current[mtx_mode][3][1] = 0.0f;
							mtx_current[mtx_mode][3][2] = 0.0f;
							mtx_current[mtx_mode][3][3] = 1.0f;*/
							/*mtx_current[mtx_mode][0][3] = 0.0f;
							mtx_current[mtx_mode][1][3] = 0.0f;
							mtx_current[mtx_mode][2][3] = 0.0f;
							mtx_current[mtx_mode][3][3] = 1.0f;*/
							//memcpy(&mtx_current[mtx_mode],&mtx_4x3,sizeof(t_matrix_4x4));
						}
					}
					break;

				// MTX_MULT_4X4
				case 0x460:
					f = (float)commandList[i].parm / 4096.0f;
					if (f >= 524288.0f)
						f -= 1048576.0f;

					mtx_4x4[mlCountR][mlCountC] = f;
					if ((++mlCountC)==4) {
						mlCountC = 0;
						if ((++mlCountR)==4) {
							mlCountR = 0;
							glMatrixMode(GL_MODELVIEW);
							glLoadMatrixf((GLfloat*)&mtx_current[mtx_mode][0][0]);
							glMultMatrixf((GLfloat*)&mtx_4x3[0][0]);
							glGetFloatv(GL_MODELVIEW_MATRIX,(GLfloat*)&mtx_current[mtx_mode][0][0]);
						}
					}
					break;

				// MTX_MULT_4X3
				case 0x464:
					f = (float)commandList[i].parm / 4096.0f;
					if (f >= 524288.0f) f -= 1048576.0f;
					mtx_4x3[mlCountR][mlCountC] = f;
					if ((++mlCountC)==3) {
						mtx_current[mtx_mode][mlCountR][mlCountC] = 
							identity_4x4[mlCountR][mlCountC];
						mlCountC = 0;
						if ((++mlCountR)==4) {
							mlCountR = 0;
							/*mtx_4x3[3][0] = 0.0f;
							mtx_4x3[3][1] = 0.0f;
							mtx_4x3[3][2] = 0.0f;
							mtx_4x3[3][3] = 1.0f;*/
							glMatrixMode(GL_MODELVIEW);
							glLoadMatrixf((GLfloat*)&mtx_current[mtx_mode][0][0]);
							glMultMatrixf((GLfloat*)&mtx_4x3[0][0]);
							glGetFloatv(GL_MODELVIEW_MATRIX,(GLfloat*)&mtx_current[mtx_mode][0][0]);
						}
					}
					/*if (check_gl_error("Mult_4x3")) {
						zombie(TELL,ZOMBIE_HALT,0,0);
						return;
					}*/
					break;

				// MTX_MULT_3X3
				case 0x468:
					f = (float)commandList[i].parm / 4096.0f;
					if (f >= 524288.0f) f -= 1048576.0f;
					mtx_3x3[mmCountR][mmCountC] = f;

					if ((++mmCountC)==3) {
						mmCountC = 0;
						if ((++mmCountR)==3) {
							mmCountR = 0;
						/*for (j=0; j<3; j++)
						{
							wsprintf(buffer,"[%d %d %d]\n",(int)(mtx_3x3[j][0]*100.0f),
								(int)(mtx_3x3[j][1]*100.0f),
								(int)(mtx_3x3[j][2]*100.0f));

							WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);
						}
						WriteConsole(hout,"\n",2,(LPDWORD)&dummy,NULL);*/

							glMatrixMode(GL_MODELVIEW);
							glLoadMatrixf((GLfloat*)&mtx_current[mtx_mode][0][0]);
							glMultMatrixf((GLfloat*)&mtx_3x3[0][0]);
							glGetFloatv(GL_MODELVIEW_MATRIX,(GLfloat*)&mtx_current[mtx_mode][0][0]);

						}
					}
					/*if (check_gl_error("Mult_3x3")) {
						zombie(TELL,ZOMBIE_HALT,0,0);
						return;
					}*/
					break;

				case 0x46C:
					if (vtx_begin) break;
					scale[sCount] = (float)commandList[i].parm / 4096.0;
					if (scale[sCount] >= 524288.0f)
						scale[sCount] -= 1048576.0f;
					if ((++sCount) == 3) {
						sCount = 0;
						glMatrixMode(GL_MODELVIEW);
						mtx_scale[0][0] = scale[0];
						mtx_scale[1][1] = scale[1];
						mtx_scale[2][2] = scale[2];
						glLoadMatrixf((GLfloat*)&mtx_current[mtx_mode][0][0]);
						glScalef(scale[0],scale[1],scale[2]);
						glGetFloatv(GL_MODELVIEW_MATRIX,(GLfloat*)&mtx_current[mtx_mode][0][0]);
					}
					break;

				case 0x470:
					if (vtx_begin) break;
					trans[tCount] = (float)commandList[i].parm / 4096.0;
					if (trans[tCount] >= 524288.0f)
						trans[tCount] -= 1048576.0f;
					if ((++tCount) == 3) {
						tCount = 0;

						//wsprintf(buffer,"trans by: %d %d %d\n",(int)(trans[0]*100.0f),(int)(trans[1]*100.0f),(int)(trans[2]*100.0f));
						//WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);

						glMatrixMode(GL_MODELVIEW);
						glLoadMatrixf((GLfloat*)&mtx_current[mtx_mode][0][0]);
						glTranslatef(trans[0],trans[1],trans[2]);
						glGetFloatv(GL_MODELVIEW_MATRIX,(GLfloat*)&mtx_current[mtx_mode][0][0]);
					}
					/*if (check_gl_error("Translate")) {
						zombie(TELL,ZOMBIE_HALT,0,0);
						return;
					}*/
					break;

				// COLOR
				case 0x480:
					glColor4ub((commandList[i].parm&0x1F)<<3,
						       (commandList[i].parm&0x3E0)>>2,
							   (commandList[i].parm&0x7C00)>>7,
							   (unsigned char)(polyalpha*255.0f));
					/*wsprintf(buffer,"color %d %d %d\n",
							(commandList[i].parm&0x1F)<<3,
						       (commandList[i].parm&0x3E0)>>2,
							   (commandList[i].parm&0x7C00)>>7);
						WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);*/
					break;

				// NORMAL
				case 0x484:
					if ((f = (float)(commandList[i].parm&0x3FF) / 512.0f) >= 1.0f) 
						f -= 2.0f;
					if ((g = (float)((commandList[i].parm>>10)&0x3FF) / 512.0f) >= 1.0f) 
						g -= 2.0f;
					if ((h = (float)((commandList[i].parm>>20)&0x3FF) / 512.0f) >= 1.0f) 
						h -= 2.0f;
					glNormal3f(f, g, h);
					break;

				// TEXCOORD
				case 0x488:
					if (activeTexture != -1) {
						g = (float)(commandList[i].parm>>16) / 16.0f;
						if (g>=2048.0f) g -= 4096.0f;
						f = (float)(commandList[i].parm&0xFFFF) / 16.0f;
						if (f>=2048.0f) f -= 4096.0f;

						if (textureCache[activeTexture].width != 0) {
							glTexCoord2f((f/((float)textureCache[activeTexture].width)),
										 (g/((float)textureCache[activeTexture].height)));
							/*wsprintf(buffer,"Texcoord %04x,%04x (%d,%d)\n",
								commandList[i].parm&0xFFFF,
								commandList[i].parm>>16,
								textureCache[activeTexture].width,
								textureCache[activeTexture].height);
							WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);*/
						}
					}
					break;

				// VTX16
				case 0x48C:
					if (zcoord)	{
						z = (float)commandList[i].parm / 4096.0f;
						if (z>=8.0f) z -= 16.0f;
						glVertex3f(x, y, z);
					} else {
						y = (float)(commandList[i].parm>>16) / 4096.0f;
						if (y>=8.0f) y -= 16.0f;
						x = (float)(commandList[i].parm&0xFFFF) / 4096.0f;
						if (x>=8.0f) x -= 16.0f;
					}
					zcoord = !zcoord;
					break;

				// VTX10
				case 0x490:
					x = (float)(commandList[i].parm&0x3FF) / 64.0f;
					y = (float)((commandList[i].parm>>10)&0x3FF) / 64.0f;
					z = (float)((commandList[i].parm>>20)&0x3FF) / 64.0f;
					if (x >= 8.0f) x -= 16.0f;
					if (y >= 8.0f) y -= 16.0f;
					if (z >= 8.0f) z -= 16.0f;
					glVertex3f(x, y, z);
					break;

				// VTX_XY
				case 0x494:
					y = (float)(commandList[i].parm>>16) / 4096.0f;
					if (y>=8.0f) y -= 16.0f;
					x = (float)(commandList[i].parm&0xFFFF) / 4096.0f;
					if (x>=8.0f) x -= 16.0f;
					glVertex3f(x, y, z);
					break;
				// VTX_XZ
				case 0x498:
					z = (float)(commandList[i].parm>>16) / 4096.0f;
					if (z>=8.0f) z -= 16.0f;
					x = (float)(commandList[i].parm&0xFFFF) / 4096.0f;
					if (x>=8.0f) x -= 16.0f;
					glVertex3f(x, y, z);
					break;
				// VTX_YZ
				case 0x49C:
					z = (float)(commandList[i].parm>>16) / 4096.0f;
					if (z>=8.0f) z -= 16.0f;
					y = (float)(commandList[i].parm&0xFFFF) / 4096.0f;
					if (y>=8.0f) y -= 16.0f;
					glVertex3f(x, y, z);
					break;
				
				// VTX_DIFF
				case 0x4A0:
					f = (float)(commandList[i].parm&0x3FF) / 4096.0f;
					if (f>=0.125f) f -= 0.250f;
					g = (float)((commandList[i].parm>>10)&0x3FF) / 4096.0f;
					if (g>=0.125f) g -= 0.250f;
					h = (float)((commandList[i].parm>>20)&0x3FF) / 4096.0f;
					if (h>=0.125f) h -= 0.250f;
					x += f; y += g; z += h;
					glVertex3f(x, y, z);
					break;

				// POLYGON_ATTRIBUTE
				case 0x4A4:
					if (vtx_begin) break;
					polyalpha = (float)((commandList[i].parm>>16)&31);
					polyalpha /= 31.0f;
					if (polyalpha != 0.0f) {
						glGetFloatv(GL_CURRENT_COLOR,&currColor[0]);
						currColor[3] = polyalpha;
						glColor4fv(&currColor[0]);
						glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
						polymode = GL_FILL;
					} else {
						polyalpha = 1.0f;
						glGetFloatv(GL_CURRENT_COLOR,&currColor[0]);
						currColor[3] = polyalpha;
						glColor4fv(&currColor[0]);
						glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
						polymode = GL_LINE;
					}
					lightAttr[0] = ((commandList[i].parm&1)!=0);
					lightAttr[1] = ((commandList[i].parm&2)!=0);
					lightAttr[2] = ((commandList[i].parm&4)!=0);
					lightAttr[3] = ((commandList[i].parm&8)!=0);
					
					useLighting[0] = useLighting[0] && lightAttr[0];
					useLighting[1] = useLighting[1] && lightAttr[1];
					useLighting[2] = useLighting[2] && lightAttr[2];
					useLighting[3] = useLighting[3] && lightAttr[3];

					if ((commandList[i].parm&0xC0)!=0xC0) {
						glEnable(GL_CULL_FACE);
						glCullFace(map3d_cull[(commandList[i].parm>>6)&3]);
					} else {
						glDisable(GL_CULL_FACE);
					}
					break;

				// TEXIMG_PARAM
				case 0x4A8:
					if (vtx_begin) break;
					//wsprintf(buffer,"3d fifo[%03X]: %08X\n",commandList[i].cmd,commandList[i].parm);
					//WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);
					if (pendingTexUpload) {
						set_ogl_texture(texfmt);
						pendingTexUpload = true;
					} else {
						pendingTexUpload = true;
					}
					texfmt = commandList[i].parm;
					break;

				// TEXPLTT_BASE
				case 0x4AC:
					if (vtx_begin) break;
					texPalOfs = commandList[i].parm&0x1FFF;
					if (pendingTexUpload) {
						set_ogl_texture(texfmt);
						pendingTexUpload = false;
					}
					//wsprintf(buffer,"3d fifo[%03X]: %08X, col=%d\n",commandList[i].cmd,commandList[i].parm,textureCache[activeTexture].numcolors);
					//WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);
					//if (activeTexture==-1) break;
					/*if ((textureCache[activeTexture].numcolors == 4) ||
						(textureCache[activeTexture].numcolors == 16) ||
						(textureCache[activeTexture].numcolors == 256)) {
						for (j=0; j<textureCache[activeTexture].numcolors; j++) {
							textureCache[activeTexture].map_r[j] = 
								(*(unsigned short*)(vramBase[4] + 
								                    (commandList[i].parm&0x1FFF)*textureCache[activeTexture].numcolors*2 +
								                   j*2)) & 0x1F;
							textureCache[activeTexture].map_r[j] /= 31.0f;
							textureCache[activeTexture].map_g[j] = 
								((*(unsigned short*)(vramBase[4] + 
								                     (commandList[i].parm&0x1FFF)*textureCache[activeTexture].numcolors*2 +
								                   j*2))>>5) & 0x1F;
							textureCache[activeTexture].map_g[j] /= 31.0f;
							textureCache[activeTexture].map_b[j] = 
								((*(unsigned short*)(vramBase[4] +
								                     (commandList[i].parm&0x1FFF)*textureCache[activeTexture].numcolors*2 +
								                   j*2))>>10) & 0x1F;
							textureCache[activeTexture].map_b[j] /= 31.0f;
							textureCache[activeTexture].map_a[j] = 1.0f;
						}
						glPixelMapfv(GL_PIXEL_MAP_I_TO_R,textureCache[activeTexture].numcolors,&textureCache[activeTexture].map_r[0]);
						glPixelMapfv(GL_PIXEL_MAP_I_TO_G,textureCache[activeTexture].numcolors,&textureCache[activeTexture].map_g[0]);
						glPixelMapfv(GL_PIXEL_MAP_I_TO_B,textureCache[activeTexture].numcolors,&textureCache[activeTexture].map_b[0]);
						glPixelMapfv(GL_PIXEL_MAP_I_TO_A,textureCache[activeTexture].numcolors,&textureCache[activeTexture].map_a[0]);
					}*/
					break;

				// DIFFUSE_AMBIENT
				case 0x4C0:
					if (vtx_begin) break;
					lightv[0] = (float)(commandList[i].parm&0x1F) / 31.0f;
					lightv[1] = (float)((commandList[i].parm>>5)&0x1F) / 31.0f;
					lightv[2] = (float)((commandList[i].parm>>10)&0x1F) / 31.0f;
					lightv[3] = 1.0f;
					glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,&lightv[0]);
					lightv[0] = (float)((commandList[i].parm>>16)&0x1F) / 31.0f;
					lightv[1] = (float)((commandList[i].parm>>21)&0x1F) / 31.0f;
					lightv[2] = (float)((commandList[i].parm>>26)&0x1F) / 31.0f;
					lightv[3] = 1.0f;
					glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,&lightv[0]);
					break;

				// SPECULAR_EMISSION
				case 0x4C4:
					if (vtx_begin) break;
					lightv[0] = (float)(commandList[i].parm&0x1F) / 31.0f;
					lightv[1] = (float)((commandList[i].parm>>5)&0x1F) / 31.0f;
					lightv[2] = (float)((commandList[i].parm>>10)&0x1F) / 31.0f;
					lightv[3] = 1.0f;
					glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,&lightv[0]);
					lightv[0] = (float)((commandList[i].parm>>16)&0x1F) / 31.0f;
					lightv[1] = (float)((commandList[i].parm>>21)&0x1F) / 31.0f;
					lightv[2] = (float)((commandList[i].parm>>26)&0x1F) / 31.0f;
					lightv[3] = 1.0f;
					glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,&lightv[0]);
					break;

				// LIGHT_VEC
				case 0x4C8:
					if (vtx_begin) break;
					if ((f = (float)(commandList[i].parm&0x3FF) / 512.0f) >= 1.0f) 
						f -= 2.0f;
					if ((g = (float)((commandList[i].parm>>10)&0x3FF) / 512.0f) >= 1.0f) 
						g -= 2.0f;
					if ((h = (float)((commandList[i].parm>>20)&0x3FF) / 512.0f) >= 1.0f) 
						h -= 2.0f;
					lightv[0] = f; lightv[1] = g; lightv[2] = h;
					glLightfv(GL_LIGHT0+((commandList[i].parm>>30)&3),GL_SPOT_DIRECTION,lightv);
					useLighting[((commandList[i].parm>>30)&3)] = true;
					break;

				// LIGHT_COLOR
				case 0x4CC:
					if (vtx_begin) break;
					lightv[0] = (float)(commandList[i].parm&0x1F) / 31.0f;
					lightv[1] = (float)((commandList[i].parm>>5)&0x1F) / 31.0f;
					lightv[2] = (float)((commandList[i].parm>>10)&0x1F) / 31.0f;
					lightv[3] = 1.0f;
					glLightfv(GL_LIGHT0+((commandList[i].parm>>30)&3),GL_AMBIENT,lightv);
					glLightfv(GL_LIGHT0+((commandList[i].parm>>30)&3),GL_DIFFUSE,lightv);
					glLightfv(GL_LIGHT0+((commandList[i].parm>>30)&3),GL_SPECULAR,lightv);
					useLighting[((commandList[i].parm>>30)&3)] = true;
					break;

				// SHININESS
				case 0x4D0:
					break;

				// BEGIN_VTXS
				case 0x500:
					glMatrixMode(GL_PROJECTION);
					glLoadMatrixf(&mtx_current[0][0][0]);
					//glLoadIdentity();

					//wsprintf(buffer,"3d fifo[%03X]: %08X\n",commandList[i].cmd,commandList[i].parm);
					//WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);

					if (pendingTexUpload) {
						set_ogl_texture(texfmt);
						pendingTexUpload = false;
					}
					glEnable(GL_BLEND);
					if (polyalpha == 1.0f) {
						glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
						//glBlendFunc(GL_SRC_ALPHA,GL_DST_ALPHA);
					} else {
						glBlendFunc(GL_SRC_ALPHA,GL_ONE);
					}

					glMatrixMode(GL_MODELVIEW);
					glLoadMatrixf(&mtx_current[2][0][0]);
					//glLoadIdentity();
					//glDisable(GL_CULL_FACE);

					/*if (framecnt<4) {
						WriteConsole(hout,"Mod\n",5,(LPDWORD)&dummy,NULL);
						glGetFloatv(GL_MODELVIEW_MATRIX,&mtx_scale[0][0]);
						for (j=0; j<4; j++)
						{
							wsprintf(buffer,"[%d %d %d %d]\n",(int)(mtx_scale[j][0]*100.0f),
								(int)(mtx_scale[j][1]*100.0f),
								(int)(mtx_scale[j][2]*100.0f),
								(int)(mtx_scale[j][3]*100.0f));

							WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);
						}
						WriteConsole(hout,"\n",2,(LPDWORD)&dummy,NULL);
					}					

					glLoadIdentity();*/
					
					if ((VRAM[IOREGS_OFFSET + 0x240] == 0x83) &&
						(activeTexture != -1) &&
						(textureCache[activeTexture].fmt != 0)) {
						//wsprintf(buffer,"Enabling texture %d\n",activeTexture);
						//WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);

						glBindTexture(GL_TEXTURE_2D,textureCache[activeTexture].id);
						glEnable(GL_TEXTURE_2D);
					} /*else {
						glDisable(GL_TEXTURE_2D);
					}*/

					useLighting[0] = useLighting[0] && lightAttr[0];
					useLighting[1] = useLighting[1] && lightAttr[1];
					useLighting[2] = useLighting[2] && lightAttr[2];
					useLighting[3] = useLighting[3] && lightAttr[3];

					/*if (useLighting[0]||useLighting[1]||useLighting[2]||useLighting[3])
						glEnable(GL_LIGHTING);
					if (useLighting[0])
						glEnable(GL_LIGHT0);
					if (useLighting[1])
						glEnable(GL_LIGHT1);
					if (useLighting[2])
						glEnable(GL_LIGHT2);
					if (useLighting[3])
						glEnable(GL_LIGHT3);*/

					//glEnable(GL_DEPTH_TEST);
					vtx_begin = true;
					if ((polyalpha>0.0f)&&(polyalpha<1.0f))
						glEnable(GL_BLEND);
					glBegin(map3d_begend[commandList[i].parm]);
					/*if (check_gl_error("Begin")) {
						zombie(TELL,ZOMBIE_HALT,0,0);
						return;
					}*/
					break;

				case 0x504:
					//WriteConsole(hout,"glEnd\n",7,(LPDWORD)&dummy,NULL);
					if (vtx_begin) {
						vtx_begin = false;
						glEnd();
						if (check_gl_error("End")) {
							zombie(TELL,ZOMBIE_HALT,0,0);
							return;
						}
					}
					glDisable(GL_BLEND);
					glDisable(GL_LIGHTING);
					glDisable(GL_LIGHT0);
					glDisable(GL_LIGHT1);
					glDisable(GL_LIGHT2);
					glDisable(GL_LIGHT3);
					if (VRAM[IOREGS_OFFSET + 0x240] == 0x83)
						glDisable(GL_TEXTURE_2D);
					break;

				case 0x540:
					VRAM[IOREGS_OFFSET+0x603] |= 8;
					//wsprintf(buffer,"vtx_begin=%d, i=%d. commandListPos=%d\n",
					//	(int)vtx_begin,i,commandListPos);
					//WriteConsole(hout,buffer,strlen(buffer),(LPDWORD)&dummy,NULL);
					if (vtx_begin) {
						glEnd();
						while (glGetError()!=GL_NO_ERROR) {}
						vtx_begin = false;
					}
					glFlush();
 					glBindTexture(GL_TEXTURE_2D,texID[1]);		
					// Copy our viewport to the bg03d texture (from 0,0 to 255,255... no border)
					glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, 256, 256, 0);
					if (check_gl_error("Swapbuffers")) {
						zombie(TELL,ZOMBIE_HALT,0,0);
						return;
					}
					commandListPos = 0;
					framecnt++;
					return;

				case 0x580:
					dsVPX1 = (commandList[i].parm)&0xFF;
					dsVPY1 = (commandList[i].parm>>8)&0xFF;
					dsVPX2 = (commandList[i].parm>>16)&0xFF;
					dsVPY2 = (commandList[i].parm>>24)&0xFF;
					glViewport(dsVPX1,dsVPY1,dsVPX2,dsVPY2);
					//WriteConsole(hout,"Viewport\n",10,(LPDWORD)&dummy,NULL);
					break;

				case 0x600:
					break;

				default:
					wsprintf(buffer,"3d fifo[%03X]: %08X\n",commandList[i].cmd,commandList[i].parm);
					WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);
					break;
			}
			//if (cmd==0x400) fifopos++;
		}
	}

	commandListPos = 0;
}


void __stdcall gpu_set(int what,int param1,int param2) {
	static bool zcoord = false;
	static float x,y,z;
	static float trans[3],scale[3];
	static int tCount=0,sCount=0,mmCountC=0,mmCountR=0,mlCountC=0,mlCountR=0;
	float f;
	static float rot=0.0f;
	int i,j,k;
	bool inVtxList;

	switch (what) {
		case GPU_3D_COMMAND:
			if (emulate3D==false)
				return;
			/*if (commandListPos < 512) {
				commandList[commandListPos].cmd = param1;
				commandList[commandListPos].parm = param2;
				commandListPos++;
			}*/
			commandList[0].cmd = param1;
			commandList[0].parm = param2;
			commandListPos = 1;
			process_3d_commands();
			if ((fifopos == fifoparms)&&(fifolen != -1)) {
				memcpy(commandList,fifocommands,sizeof(t_3d_command)*fifolen);
				commandListPos = fifolen;
				
				/*inVtxList = false;
				j = 0;
				for (i=0; i<commandListPos; i++) {
					if (commandList[i].cmd == 0x500) {
						inVtxList = true;
						j = i;
					} else if (commandList[i].cmd == 0x504) {
						inVtxList = false;
					} else if ((commandList[i].cmd == 0x46C)||(commandList[i].cmd == 0x470)) {
						if (inVtxList) {
							for (k=commandListPos-1; k>=j; k--) {
								commandList[k+3] = commandList[k];
							}
							for (k=0; k<3; k++) {
								commandList[j+k] = commandList[i+k+3];
							}
							for (k=i+3; k<commandListPos; k++) {
								commandList[k] = commandList[k+3];
							}
							j += 3;
						}
					}
				}*/
				process_3d_commands();
				commandListPos = 1;
				fifopos = -1;
				fifolen = -1;
			}
			return;


		case CPU_IRQ_CALLBACK:
			cpu_irq = (cpu_irq_callback)param1;
			break;
		case MMU_READ_WORD_CALLBACK:
			mmu_read_word = (mmu_read_word_callback)param1;
			break;
		case GPU_VBLANK_ACTION:
			vbiAction = param1;
			vbiAddress = param2;
			break;
		case GPU_HBLANK_ACTION:
			hbiAction = param1;
			hbiAddress = param2;
			break;
		case PLUGIN_SETTING:
			switch (param1)
			{
				case 0:
					frameskip = param2;
					break;
				case 1:
					limitFps = param2;
					break;
				case 2:
					enableObj = param2;
					break;
				case 4:
					if (!useGL && (param2!=0))
						setup_gl(dc);
					useGL = param2;
					break;
				case 5:
					if (renderer == 2)
						glFormatBackup = param2;
					else
						glFormat = param2;
					break;
				case 6:
					filter = (param2==0)?GL_NEAREST:GL_LINEAR;
				  	if (useGL && glOK) {
						glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter);
				  		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter);
					}
					break;
				case 7:
					if ((int)param2 > 0)
						frameInterval = (0x10000 / (unsigned int)param2) - 2;
					break;
				case 8:
					enableLayer[(param2>>8)&1][param2&7] = param2>>16;
					break;

				case 20:
					renderer = param2&3;
					block3Dcommands = true;		/* Make sure no new 3D commands are issued
												   this frame. */
					if (renderer==0) {
						if (!useGL) {
							glEnd();
							while (glGetError()!=GL_NO_ERROR) {}
						}
						glFormat = glFormatBackup;
						useGL = true;
						useD3D = false;
					} else if (renderer==1)	{
						glFormat = GL_UNSIGNED_SHORT_5_5_5_1; //glFormatBackup;
						useGL = false;
						useD3D = true;
					} else	{
						glFormatBackup = glFormat;
						glFormat = 0;
						useGL = false;
						useD3D = false;
					}
					break;

				case 25:
					displayScreens = (int)param2;
					break;

				case 26:
					emulate3D = ((int)param2 != 0);
					break;

				default:
					break;
			}
			break;

		case GPU_FRAME_RATE:
			frameInterval = (0x10000 / (unsigned int)param1) - 2;
			break;

		case ZOMBIE_WINDOW_HANDLE:
			hwnd = (HWND)param1;
			break;
		default:
			break;
	}
}


void __stdcall gpu_hblank() {
	VRAM[IOREGS_OFFSET+4] |= 0x2;
	VRAM[IOREGS_ARM7_OFFSET+4] |= 0x2;

	/*	wsprintf(buffer,"line %d: DISPSTAT=%d, IE=%d, IF=%d\n",
			           scanline,VRAM[IOREGS_OFFSET+4]&0x10,
					   IOREGS[0x210],
					   IOREGS[0x214]);
		WriteConsole(hout,buffer,strlen(buffer),(unsigned long*)&dummy,NULL);*/

	if (hblankEnabled &&
		((VRAM[IOREGS_OFFSET+4] & 0x10) != 0) &&
		((IOREGS[0x210] & 0x02) != 0))
	{
		if ((mmu_read_word(0x4000214)&2)==0)
		//if (mmu_get(MMU_IRQ_OK)&2)
		{
			IOREGS[0x214] |= 2;
			zombie(TELL,MMU_IRQ,(void*)2L,0L);
			cpu_irq(0x00000018);


		}
	}

	if (hblankEnabled &&
		((VRAM[IOREGS_ARM7_OFFSET+4] & 0x10) != 0) &&
		((int)arm7_irq!=NULL) &&
		((IOREGS_ARM7[0x210] & 0x02) != 0)) {
		if ((arm7_read_word(0x4000214)&2)==0) {
			//	wsprintf(buffer,"HBlank IRQ on ARM7");
			//	WriteConsole(hout,buffer,strlen(buffer),(unsigned long*)&dummy,NULL);
			IOREGS_ARM7[0x214] |= 2;
			zombie(TELL,MMU_IRQ,(void*)2L,(void*)1L);
			arm7_irq(0x00000018);
		}
	}

	mmu_irq(2);	// Signal a hblank to the MMU

	scanline++;
	if (scanline < 0xC1) {
		if (((VRAM[IOREGS_OFFSET]&8)==0) || (!emulate3D))
			draw_scanline(VRAM,vscreen,0);
		draw_scanline(VRAM_s,vscreen_s,1);
	}

	scanline--;
}



void __stdcall gpu_advance_line() {
	VRAM[IOREGS_OFFSET+4] &= 0xF9;	// Clear hblank bit and V-count match bit
	VRAM[IOREGS_ARM7_OFFSET+4] &= 0xF9;


	scanline++;
	VRAM[IOREGS_OFFSET+6] = (unsigned char)scanline;
	VRAM[IOREGS_ARM7_OFFSET+6] = (unsigned char)scanline;

	if (scanline==visibleLines+1) {
		VRAM[IOREGS_OFFSET+4]      |= 1;
		VRAM[IOREGS_ARM7_OFFSET+4] |= 1;
		if (vblankEnabled &&
			((VRAM[IOREGS_OFFSET+4] & 0x8) != 0) &&
			((IOREGS[0x210] & 0x1) != 0))
		{
			//zombie(TELL,ZOMBIE_DEBUG_MSG,"Wake up",0);
			zombie(TELL,CPU_WAKE_UP,(void*)1,(long)0);
			if ((mmu_read_word(0x4000214)&1)==0)
			////if (mmu_get(MMU_IRQ_OK)&1)
			{
				if (logIRQ)
					zombie(TELL,ZOMBIE_DEBUG_MSG,"VBlank IRQ",0);
				IOREGS[0x214] |= 1;
				zombie(TELL,MMU_IRQ,(void*)1,(long)0);
				cpu_irq(0x00000018);
			}
		}
		if (vblankEnabled &&
			((int)arm7_irq!=NULL) &&
			((VRAM[IOREGS_ARM7_OFFSET+4] & 0x8) != 0) &&
			((IOREGS_ARM7[0x210] & 0x1) != 0)) {
			//zombie(TELL,CPU_WAKE_UP,(void*)1,(void*)1);
			if ((arm7_read_word(0x4000214)&1)==0) {
				if (logIRQ)
					zombie(TELL,ZOMBIE_DEBUG_MSG,"VBlank IRQ",0);
				//wsprintf(buffer,"VBlank IRQ on ARM7");
				//WriteConsole(hout,buffer,strlen(buffer),(unsigned long*)&dummy,NULL);

				IOREGS_ARM7[0x214] |= 1;
				zombie(TELL,MMU_IRQ,(void*)1,(void*)1);
				arm7_irq(0x00000018);
			}
		}

		//io_base = VRAM + IOREGS_OFFSET;
		//process_3d_commands();

		mmu_irq(1);
	}
	else if (scanline==totalLines)
	{
		//wsprintf(buffer,"-----------------\n");
		//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

		if (renderer==1)
			finish_screen_d3d();
		else if (renderer==0)
			finish_screen_ogl();
		else
			finish_screen_gdi();

		reset_screen();
		VRAM[IOREGS_OFFSET+6] = 0;
		VRAM[IOREGS_ARM7_OFFSET+6] = 0;
		scanline = 0;
		VRAM[IOREGS_OFFSET+4] &= 0xFE;
		VRAM[IOREGS_ARM7_OFFSET+4] &= 0xFE;

		if (linesCaptured==192) {
			VRAM[IOREGS_OFFSET+0x67] &= 0x7F;	// Clear the capture enable bit
			VRAM[IOREGS_ARM7_OFFSET+0x67] &= 0x7F;	// Clear the capture enable bit
		}
		linesCaptured = 0;

		if (VRAM[IOREGS_OFFSET+5] == 0)	{
			VRAM[IOREGS_OFFSET+4] |= 4;
			if (((VRAM[IOREGS_OFFSET+4] & 0x20) != 0) &&
				((IOREGS[0x210] & 0x4) != 0)) {
				if ((mmu_read_word(0x4000214)&4)==0) {
					if (logIRQ)
						zombie(TELL,ZOMBIE_DEBUG_MSG,"VCount IRQ",0);
					zombie(TELL,MMU_IRQ,(void*)4,0);
					IOREGS[0x214] |= 4;
					cpu_irq(0x00000018);
					mmu_irq(4);
				}
			}
		}
		if (VRAM[IOREGS_ARM7_OFFSET+5] == 0)	{
			VRAM[IOREGS_ARM7_OFFSET+4] |= 4;
			if (((VRAM[IOREGS_ARM7_OFFSET+4] & 0x20) != 0) &&
				((int)arm7_irq!=NULL) &&
				((IOREGS_ARM7[0x210] & 0x4) != 0)) {
				if ((arm7_read_word(0x4000214)&4)==0) {
					if (logIRQ)
						zombie(TELL,ZOMBIE_DEBUG_MSG,"VCount IRQ",0);
				//wsprintf(buffer,"VCount IRQ on ARM7");
				//WriteConsole(hout,buffer,strlen(buffer),(unsigned long*)&dummy,NULL);
					zombie(TELL,MMU_IRQ,(void*)4,(void*)1);
					IOREGS_ARM7[0x214] |= 4;
					arm7_irq(0x00000018);
					mmu_irq(4);
				}
			}
		}

	} else
	{
		// Check for VCOUNT match
		if (VRAM[IOREGS_OFFSET+5] == (unsigned char)scanline)
		{
			VRAM[IOREGS_OFFSET+4] |= 4;
			if (((VRAM[IOREGS_OFFSET+4] & 0x20) != 0) &&
				((IOREGS[0x210] & 0x4) != 0))
			{
				if ((mmu_read_word(0x4000214)&4)==0)
				{
					if (logIRQ)
						zombie(TELL,ZOMBIE_DEBUG_MSG,"VCount IRQ",0);
					zombie(TELL,MMU_IRQ,(void*)4,0);
					IOREGS[0x214] |= 4;
					cpu_irq(0x00000018);
					//if ((int)arm7_irq) arm7_irq(0x18);
					mmu_irq(4);
				}
			}
		}
	}
//zombie(TELL,ZOMBIE_DEBUG_MSG,"Done",0);
}





// Shell sort based on insertion sort
void sort_layers(int items)
{
	int i,j,gap,first,last;
	tag_bg tempi,tempj;

	for (i=0; i<items; i++)
		sorted_layers[i] = layers[i];

	last = items;
	gap = (last/3)+1;
	while (true)
	{
		first = gap+1;
		for (i=first; i<=last; i++)
		{
			tempi = sorted_layers[i-1];
			j = i-gap;
			while (true)
			{
				tempj = sorted_layers[j-1];
				if ((tempi.prio < tempj.prio) ||
				    ((tempi.prio == tempj.prio) && (tempi.num < tempj.num)))
				{
					j += gap;
					break;
				}
				sorted_layers[j+gap-1] = tempj;
				if (j<=gap) break;
				j -= gap;
			}
			sorted_layers[j-1] = tempi;
		}
		if (gap == 1) return;
		else gap = (gap/3)+1;
	}
}




void setup_color_effects(int bg)
{
	int srcbg,dstbg;
	int currBgMask,lastBgMask;
	char *srcpal;

	if (bg==-1) bg=4;

	// Point the current palette to the regular palette
	currPalette = (bg!=4) ? pal_base : pal_base+0x200;

	didLumaOp = false;

	// Set red/blue/green shift amounts depending on what rendering mode and
	// texture format is used.
	*(int*)&SHAMT4 = 0;
	*(int*)&SHAMT5 = 10;
	*(int*)&SHAMT6 = 5;
	if (useGL && glOK && (glFormat==GL_UNSIGNED_SHORT_5_5_5_1))
	{
		*(int*)&SHAMT1 = 11;
		*(int*)&SHAMT2 = 1;
		*(int*)&SHAMT3 = 6;
	} else if (useD3D)
	{
		*(int*)&SHAMT1 = 10;
		*(int*)&SHAMT2 = 0;
		*(int*)&SHAMT3 = 5;
	} else
	{
		*(int*)&SHAMT1 = 0;
		*(int*)&SHAMT2 = 10;
		*(int*)&SHAMT3 = 5;
	}

	// Check if color effects are enabled
	lumachange = 0;
	ablend = io_base[0x50]&0xC0;
	objLuma = 0;

	if (ablend==0x40)
	{
		srcbg = io_base[0x50]&0x1F; 
		dstbg = io_base[0x51]&0x1F; 
		currBgMask = 1<<bg;
		lastBgMask = 1<<lastbg;
		if (( ((srcbg&currBgMask)==currBgMask) && ((dstbg&lastBgMask)==lastBgMask) ) ||
		    ( ((dstbg&currBgMask)==currBgMask) && ((srcbg&lastBgMask)==lastBgMask) )) // ||
			//( (bg==4) && ((dstbg&lastBgMask)==lastBgMask) ))
		{
			//if (bg==4)
			//	memset(alphamap,0,256*sizeof(short));

			eva = io_base[0x52]&0x1F;
			evb = io_base[0x53]&0x1F;
			ablend = 1;
			//eva = evb = 8;
			//wsprintf(buffer,"Ok to blend bg%d with bg%d. eva=%d, evb=%d\n",bg,lastbg,eva,evb);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

			//if (useGL && glOK && (glFormat==GL_UNSIGNED_SHORT_5_5_5_1))
			if (glFormat==GL_UNSIGNED_SHORT_5_5_5_1)
					flip_palette(core*2);
			return;
			//blendwith[bg] = lastbg;
		}
		//////////////////
		// Luma increase
		//////////////////
	} else if (ablend==0x80)
	{
		ablend = 0;
		if (io_base[0x50]&(1<<bg))
		{
			lumachange = 1;
			evy = io_base[0x54]&0x1F;
			if (evy>16) evy = 16;
			if (evy) {
				currPalette = (unsigned char*)&palfx[core*2][0];
				srcpal = (char*)(pal_base + ((bg==4) ? 0x200 : 0));

				didLumaOp = true;
				if (bg==4) objLuma = 1;
				/////////////////////////////////////////////////////////////////////////////////
				// Create another palette in which all colors have been added (31-itself)*EVY/16
				/////////////////////////////////////////////////////////////////////////////////
				luma_inc(currPalette,(unsigned char*)srcpal,evy,(((io_base[0x50])&0x20)==0));
				return;
			}
		}
	} else if (ablend==0xC0)
	{
		ablend = 0;
		//////////////////
		// Luma decrease
		//////////////////
		if (io_base[0x50]&(1<<bg))
		{
			lumachange = 3;
			evy = io_base[0x54]&0x1F;
			if (evy) {
				currPalette = (unsigned char*)&palfx[core*2][0];

				srcpal = (char*)(pal_base + ((bg==4) ? 0x200 : 0));

				didLumaOp = true;
				if (bg==4) objLuma = 2;

				/////////////////////////////////////////////////////////////////////////////////
				// Create another palette in which all colors have been subtracted itself*EVY/16
				/////////////////////////////////////////////////////////////////////////////////
				luma_dec(currPalette,(unsigned char*)srcpal,evy,(((io_base[0x50])&0x20)==0));
				return;
			}
		}
	}

	if (glFormat==GL_UNSIGNED_SHORT_5_5_5_1)
		if (bg==4)
		{
			flip_palette(core*2+1);
		} else
		{
			if (((io_base[0]&8)==0))
				flip_palette(core*2);
		}
	if (ablend!=1) ablend = 0;
}




inline void check_ext_pltt_banks(int bg)
{
	if (bg == -1) // -1 = OBJ
	{
		useExtPltt = io_base[3] & 0x80;
		if (useExtPltt)
		{
			useExtPltt = 0;
			if (core == 0)	// Main core
			{
				if (IOREGS[VRAMCNT_F] == 0x85)
				{
					useExtPltt = ((unsigned int)VRAM_F);
				} else if (IOREGS[VRAMCNT_G] == 0x85)
				{
					useExtPltt = ((unsigned int)VRAM_G);
				}
			} else			// Sub core
			{
				if (IOREGS[VRAMCNT_I] == 0x83)
				{
					useExtPltt = ((unsigned int)VRAM_I);
				}
			}
		}
	} else		  // BG
	{
		useExtPltt = io_base[3] & 0x40;

		// Check for extended palette banks
		if (useExtPltt)
		{
			useExtPltt = 0;
			if (core == 0)	// Main core ?
			{
				if (IOREGS[VRAMCNT_E] == 0x84)
				{
					useExtPltt = ((unsigned int)VRAM_E) + (bg*0x2000);

				} else if (IOREGS[VRAMCNT_F] == 0x84)
				{
					if (IOREGS[VRAMCNT_G] == 0x8C)
						useExtPltt = ((unsigned int)VRAM_F) + (bg*0x2000);
					else
						useExtPltt = ((unsigned int)VRAM_F) + ((bg&1)*0x2000);

				} else if (IOREGS[VRAMCNT_G] == 0x84)
				{
					useExtPltt = ((unsigned int)VRAM_G) + ((bg&1)*0x2000);

				} else if (IOREGS[VRAMCNT_F] == 0x8C)
				{
					useExtPltt = ((unsigned int)VRAM_F) + ((bg&1)*0x2000);

				} else if (IOREGS[VRAMCNT_G] == 0x8C)
				{
					useExtPltt = ((unsigned int)VRAM_G) + ((bg&1)*0x2000);
				}
			} else						// Sub core
			{
				if (IOREGS[VRAMCNT_H] == 0x82)
				{
					useExtPltt = ((unsigned int)VRAM_H) + (bg*0x2000);
				}
			}
		}
	}
}


void draw_backdrop() {
/*	if (useGL && glOK && (glFormat==GL_UNSIGNED_SHORT_5_5_5_1)) {
		__asm
		{
			pusha
			mov			edi,[vscreen_curr]
			mov			esi,[currPalette]
			mov			ebx,[scanline]
			dec			ebx
			shl			ebx,9
			add			edi,ebx
			xor			eax,eax
			mov			ax,[esi]
	mov 	ebx,eax
	mov 	ecx,eax
	shl 	ebx,11
	shr 	eax,9
	and 	ecx,003E0h
	and 	ebx,0F800h
	shl 	ecx,1
	and 	eax,0003Eh
	or 		ecx,ebx
	or 		eax,ecx
			mov			ebx,eax
			shl			ebx,16
			or			eax,ebx
			mov			ecx,128
			rep			stosd
			popa
		}
	} else {*/
		__asm
		{
			pusha
			mov			edi,[vscreen_curr]
			mov			esi,[currPalette]
			mov			ebx,[scanline]
			dec			ebx
			shl			ebx,9
			add			edi,ebx
			xor			eax,eax
			mov			ax,[esi]
			mov			ebx,eax
			shl			ebx,16
			or			eax,ebx
			mov			ecx,128
			rep			stosd
			popa
		}
	//}
	return;
}



const int bgMemBase[2] = {0x000>>2,0x200>>2};
const int objMemBase[2] = {0x400>>2,0x600>>2};

void draw_text_bg(int bg)
{
	int cntReg,ofsReg,colMode;
	int i,chrBaseDec,scrBaseDec;

	// Get the control register offset of this background
	cntReg = 0x08 + (bg<<1);
	ofsReg = 0x10 + (bg<<2);

	// Get the horizontal and vertical screen offsets
	hofs = io_base[ofsReg]&0xFF;
	hofs += (io_base[ofsReg+1]&1)<<8;
	vofs = io_base[ofsReg+2]&0xFF;
	vofs += (io_base[ofsReg+3]&1)<<8;

	// Get the CHR base, SCR base and color mode
	chrBase = (io_base[cntReg] >> 2) & 0xF;
	scrBase = io_base[cntReg+1] & 0x1F;
	colMode = io_base[cntReg] & 0x80;

	mstrChar = (io_base[3]&7)*4;
	mstrScrn = (io_base[3]&0x38)*4;

	hTileMask = (io_base[cntReg+1] >> 6) & 3;
	vTileMask = textBgMasks[hTileMask][1];
	tileShift = textBgShifts[hTileMask];
	hScrMask = textScrMasks[hTileMask][0];
	hScrShift = textScrShifts[0][hTileMask];
	vScrMask = /*0xFF + ((hTileMask&2)*0x80);*/ textScrMasks[hTileMask][1];
	vScrShift = textScrShifts[1][hTileMask];
	//hTileMask = textBgMasks[hTileMask][0];
	hTileMask = ((hTileMask&1)*0x20);

	charBank = NULL;
	screenBank = NULL;
	chrBaseDec = 0;
	scrBaseDec = 0;

	chrBase += mstrChar;
	scrBase += mstrScrn;

	/*wsprintf(buffer,"line %d: core=%d, bg=%d, scrBase=%d, charBase=%d, C.type=%d\n",
		scanline,core,bg,scrBase,chrBase,vramBank[2].type);
	WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);*/
	for (i=0; i<9; i++)
	{
		if (vramBank[i].type == MAIN_BG+core)
		{
			if ((chrBase >= vramBank[i].base-bgMemBase[core]) &&
				(chrBase < vramBank[i].base+vramBank[i].size-bgMemBase[core]))
			{
				charBank = vramBase[i];
				//break;
			} else if (chrBase >= vramBank[i].base-bgMemBase[core])
			{
				chrBaseDec += vramBank[i].size;
			}

			if (((scrBase>>3) >= vramBank[i].base-bgMemBase[core]) &&
				((scrBase>>3) < vramBank[i].base+vramBank[i].size-bgMemBase[core]))
			{
				screenBank = vramBase[i];
				if (charBank!=NULL) break;
			} else if ((scrBase>>3) >= vramBank[i].base-bgMemBase[core])
			{
				scrBaseDec += (vramBank[i].size<<3);
			}
		}
	}
	if ((charBank==NULL)||(screenBank==NULL))
	{
		if (newframe == 0) {
			flip_palette(core*2);
			draw_backdrop();
		}
		newframe++;
		return;
	}

	/*wsprintf(buffer,"line %d: screenBank = %X, charBank = %X, VRAM_A = %X\n",
		scanline,screenBank,charBank,VRAM_A);
	wsprintf(buffer,"core: %d, bg=%d, line %d: hofs=%d, vofs=%d\n",core,bg,scanline-1,hofs,vofs);
	WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);*/

	check_ext_pltt_banks(bg);
	setup_color_effects(bg);

	lastbg = bg;

	chrBase -= mstrChar;
	scrBase -= mstrScrn;
	chrBase -= chrBaseDec;
	scrBase -= scrBaseDec;
	if ((chrBase<0) || (scrBase<0)) return;
	bgcnt &= 0xE0C3;
	bgcnt |= (chrBase << 2) | (scrBase << 8);

	/*wsprintf(buffer,"core=%d line %d:bg=%d, chrBase=%d, useExtPltt=%X, chrBank=%X, scrBank=%X, bankA=%X\n",
		core,scanline,bg,chrBase,useExtPltt,charBank,screenBank,VRAM_A);
	WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);*/
	//return;

///////////////////////////
// Handle 256x1 color BGs
///////////////////////////
	if (colMode == 0x80)
	{
		if (useExtPltt != 0)
		{
			flip_palette(core*2);
			bg_line_text_bottom();

		} else
		{
		__asm
		{
			pusha
			mov		eax,[scanline]
			dec		eax
			mov		edi,[vscreen_curr]
			shl		eax,9
			add		edi,eax				// edi = vscreen+scanline*512

			mov		esi,[screenBank]
			mov		eax,[scrBase]
			add eax,[mstrScrn]
			shl		eax,11				// Select 2k screen base block
			mov		ebx,[scanline]
			dec		ebx
			add		ebx,[vofs]

			mov		edx,ebx
			and		edx,[vScrMask]
			shr		edx,3
			mov		ecx,[hTileMask]
			shr		ecx,5
			add		ecx,6
			shl		edx,cl				// ebx = (scanline/8)*64
			add		esi,edx
			add		esi,eax
			mov		ecx,ebx

			shr		ecx,3
			and		ecx,0x1F
			shl		ecx,6
			add		esi,ecx
			mov		ecx,ebx

			and		ecx,7				// ecx = (scanline+yscroll)&7  (tile row)

			mov		ebx,ecx
			xor		ebx,7				// "Inverted" tile row (used for vflipping)
			shl		ebx,3
			shl		ecx,3				// Multiply by 4 (number of bytes per row)
			mov		[tileOffset+4],ebx
			mov		[tileOffset],ecx

			mov		[lineaddr],edi

			mov		edx,[charBank]
			mov		eax,[chrBase]
			add eax,[mstrChar]
			shl		eax,14
			add		edx,eax
			mov		[dummy],edx

			/*mov edx,[charBank] //[VRAM_curr]
			mov ebx,ecx
			mov eax,[chrBase]
			xor ebx,7				// "Inverted" tile row (used for vflipping)
			shl eax,14				// Select 16k chr base block
			shl ebx,3
			shl ecx,3				// Multiply by 8 (number of bytes per row)
			add edx,eax				// edx = VRAM + chrBase
			mov [tileOffset+4],ebx
			mov [dummy],edx
			mov [tileOffset],ecx

			mov [lineaddr],edi		// Save destination address for later use*/

			// Is this the bottom-most layer?
			test	dword ptr [newframe],0xF
			jnz		__m0_not_first_bg

__m0_first_bg:

			push	ebp
			mov		ebp,[currPalette]

			mov		eax,[hofs]
			mov		ebx,eax
			shr		eax,3				// eax = xscroll/8
			mov		[xtile],eax
			/*mov ebx,eax
			and eax,[hTileMask] //0x1F			// AND with the number of tiles per row
			and ebx,0x1F
			shl eax,5
			add eax,ebx
			mov bx,[esi+eax*2]		// Get tile data*/
			lea		ebx,[eax*8]
			and		eax,0x1F
			and		ebx,[hScrMask]
			mov		ecx,[hScrShift]
			shl		ebx,cl
			lea		eax,[ebx+eax*2]
			xor		ebx,ebx				// ebx = 0
			mov		bx,[esi+eax]		// Get tile data
			inc		dword ptr [xtile]	// Next tile..
			mov		eax,ebx
			and		ebx,0x3FF			// ebx = tile number
			shr		eax,11				// Get vflip flag
			shl		ebx,6				// ebx *= 8*8
			and		eax,1
			add		ebx,[dummy] 		// ebx += VRAM + chrBase + tile_row*8
			add		ebx,[tileOffset+eax*4]

			mov		eax,[hofs]
			//add	ebx,8
			mov		ecx,8
			and		eax,7
			add		ebx,eax
			sub		ecx,eax				// ecx = 8 - (xscroll&7)  (length of first tile)
			__m0_tile0_draw:
				xor		eax,eax			// Clear eax
				mov		al,[ebx]		// Load pixel
				mov		ax,[ebp + eax*2]	// Get 15-bit color value
				mov		[edi],ax		// Store
				add		edi,2			// Increase destination pointer
				inc		ebx
				dec		ecx
				jnz		__m0_tile0_draw

			mov		ecx,31 //29
			__m0_draw_tiles:
				/*push ecx
				mov eax,[xtile]
				mov ebx,eax
				and eax,[hTileMask] //0x1f
				and ebx,0x1F
				shl eax,5
				add eax,ebx
				mov bx,[esi+eax*2]		// Get tile data*/
				push	ecx
				mov		eax,[xtile]
				lea		ebx,[eax*8]
				and		eax,0x1F
				and		ebx,[hScrMask]
				mov		ecx,[hScrShift]
				shl		ebx,cl
				lea		eax,[ebx+eax*2]
				xor		ebx,ebx				// ebx = 0
				mov		bx,[esi+eax]		// Get tile data

				mov		eax,ebx
				mov		edx,ebx
				shr		eax,11				// Get vflip flag
				inc		dword ptr [xtile]
				and		eax,1
				and		edx,0x400			// Get hflip flag
				and		ebx,0x3FF
				shr		edx,10
				shl		ebx,6
				//mov	ebx,2112
				add		ebx,[tileOffset+eax*4]
				add		ebx,[dummy]
				mov		eax,1
				lea		ebx,[ebx+edx*8]
				sub		eax,edx
				sub		ebx,edx				// if (hflip==1) ebx+=7
				sub		eax,edx
				mov		edx,eax				// edx = (hflip==1) ? -1 : 1

				mov		ecx,8
				__m0_draw:
					xor		eax,eax
					mov		al,[ebx]
					mov		ax,[ebp + eax*2]
					mov		[edi],ax
					add		edi,2
					add		ebx,edx
					dec		ecx
					jnz		__m0_draw
				pop		ecx
				dec		ecx
				jnz		__m0_draw_tiles

__m0_tile31:
				mov eax,[xtile]

				lea ebx,[eax*8]
				and eax,0x1F
				and ebx,[hScrMask]
				mov ecx,[hScrShift]
				shl ebx,cl
				lea eax,[ebx+eax*2]
				xor ebx,ebx				// ebx = 0
				mov bx,[esi+eax]		// Get tile data

				mov eax,ebx				// Get vflip flag
				and ebx,0x3FF
				shr eax,11
				shl ebx,6
				and eax,1
				add ebx,[dummy] //edx
				add ebx,[tileOffset+eax*4]

				mov ecx,[hofs]
				and ecx,7
				jz __m0_tile31_end
				__m0_tile31_draw:
					xor eax,eax
					mov al,[ebx]
					mov ax,[ebp + eax*2]
					mov [edi],ax
					add edi,2
					inc ebx
					dec ecx
				jnz __m0_tile31_draw
__m0_tile31_end:
			jmp __m0_done

/////////////////////////////////////////
// Handle semi-transparent layers
/////////////////////////////////////////
__m0_not_first_bg:
			push ebp
			mov ebp,[currPalette]

			test [ablend],1
			jnz __m0_prepare_blend

			mov eax,[hofs]
			xor ebx,ebx
			shr eax,3
			mov [xtile],eax
		/*	mov ebx,eax
			and eax,[hTileMask]
			and ebx,0x1F
			shl eax,5
			add eax,ebx
			mov bx,[esi+eax*2]*/
				lea ebx,[eax*8]
				and eax,0x1F
				and ebx,[hScrMask]
				mov ecx,[hScrShift]
				shl ebx,cl
				lea eax,[ebx+eax*2]
				xor ebx,ebx				// ebx = 0
				mov bx,[esi+eax]		// Get tile data
			mov eax,ebx
			inc dword ptr [xtile]
			shr eax,11
			and ebx,0x3FF
			and eax,1
			shl ebx,6
			add ebx,[tileOffset+eax*4]
			add ebx,[dummy]

			mov eax,[hofs]
			mov ecx,8
			and eax,7
			add ebx,eax
			sub ecx,eax
				__m0_tile0_draw_trans:
					mov al,[ebx]
					and eax,0xFF
					jz __m0_tile0_invisible
					mov ax,[ebp + eax*2]
					mov [edi],ax
__m0_tile0_invisible:
					add edi,2
					inc ebx
					dec ecx
				jnz __m0_tile0_draw_trans

			mov ecx,31 //29
			__m0_draw_tiles_trans:
				/*push ecx
				mov eax,[xtile]
				xor ebx,ebx
				mov ebx,eax
				and eax,[hTileMask]
				and ebx,0x1F
				shl eax,5
				add eax,ebx
				mov bx,[esi+eax*2]*/
				push ecx
				mov eax,[xtile]
				lea ebx,[eax*8]
				and eax,0x1F
				and ebx,[hScrMask]
				mov ecx,[hScrShift]
				shl ebx,cl
				lea eax,[ebx+eax*2]
				xor ebx,ebx				// ebx = 0
				mov bx,[esi+eax]		// Get tile data

				mov eax,ebx
				inc dword ptr [xtile]
				shr eax,11
				mov edx,ebx
				and eax,1
				and edx,0x400			// get hflip flag
				and ebx,0x3FF
				shr edx,10
				shl ebx,6
				add ebx,[tileOffset+eax*4]
				add ebx,[dummy]
				mov eax,1
				lea ebx,[ebx+edx*8]
				sub eax,edx
				sub ebx,edx				// if (hflip==1) ebx+=7
				sub eax,edx
				mov edx,eax				// edx = (hflip==1) ? -1 : 1

				mov ecx,8
				__m0_draw_trans:
					mov al,[ebx]
					and eax,0xFF
					jz __m0_invisible
						mov ax,[ebp + eax*2]
						mov [edi],ax
					__m0_invisible:
					add edi,2
					add ebx,edx
					dec ecx
				jnz __m0_draw_trans
				pop ecx
				dec ecx
			jnz __m0_draw_tiles_trans

__m0_tile31_trans:
				mov eax,[xtile]
				/*xor ebx,ebx
				and eax,[hTileMask] //0x1F
				mov bx,[esi+eax*2]*/
				lea ebx,[eax*8]
				and eax,0x1F
				and ebx,[hScrMask]
				mov ecx,[hScrShift]
				shl ebx,cl
				lea eax,[ebx+eax*2]
				xor ebx,ebx				// ebx = 0
				mov bx,[esi+eax]		// Get tile data

				mov eax,ebx
				inc dword ptr [xtile]
				shr eax,11
				and ebx,0x3FF
				and eax,1
				shl ebx,6
				add ebx,[tileOffset+eax*4]
				add ebx,[dummy] //edx

				mov ecx,[hofs]
				and ecx,7
				jz __m0_tile31_trans_end
				__m0_tile31_draw_trans:
					mov al,[ebx]
					and eax,0xFF
					jz __m0_tile31_invisible
					mov ax,[ebp + eax*2]
					mov [edi],ax
__m0_tile31_invisible:
					add edi,2
					inc ebx
					dec ecx
				jnz __m0_tile31_draw_trans
__m0_tile31_trans_end:
				jmp __m0_done

/////////////////////////////////////////
// Handle alpha blended layers
/////////////////////////////////////////
__m0_prepare_blend:
			mov eax,[p_alphamap]
			mov [p_alphamap_int],eax

			mov eax,[hofs]
			xor ebx,ebx
			shr eax,3
			mov [xtile],eax
			/*and eax,[hTileMask] //0x1F
			mov bx,[esi+eax*2]*/
				lea ebx,[eax*8]
				and eax,0x1F
				and ebx,[hScrMask]
				mov ecx,[hScrShift]
				shl ebx,cl
				lea eax,[ebx+eax*2]
				xor ebx,ebx				// ebx = 0
				mov bx,[esi+eax]		// Get tile data
			mov eax,ebx
			inc dword ptr [xtile]
			mov [tileData],ebx
			shr eax,11
			and ebx,0x3FF
			and eax,1
			shl ebx,6
			add ebx,[tileOffset+eax*4]
			add ebx,[dummy] //edx

			mov eax,[hofs]
			push edx
			mov ecx,8
			mov edx,edi
			and eax,7
			sub ecx,eax
			add edx,[vscreen_size]
			add edx,[vscreen_size]
			lea edi,[edi+ecx*2]
			push edi
			mov edi,[p_alphamap_int]
			add [p_alphamap_int],ecx
			add ebx,eax
			test dword ptr [tileData],0x400
			jnz __m0_tile0_trans_bld_hflip
				__m0_tile0_draw_trans_bld:
					mov al,[ebx]
					and eax,0xFF
					jz __m0_tile0_invisible_bld
					mov ax,[ebp + eax*2]
					mov [edx],ax
					mov byte ptr [edi],0xFF
					jmp __m0_tile0_opaque_bld
__m0_tile0_invisible_bld:
					mov byte ptr [edi],0
__m0_tile0_opaque_bld:
					inc edi
					add edx,2
					inc ebx
					dec ecx
				jnz __m0_tile0_draw_trans_bld
				jmp __m0_tile0_trans_bld_end

				// Draw a hflipped tile
__m0_tile0_trans_bld_hflip:
				add ebx,7
				__m0_tile0_draw_trans_bld_hflip_draw:
					mov al,[ebx]
					and eax,0xFF
					jz __m0_tile0_invisible_bld_hflip
					mov ax,[ebp + eax*2]
					mov [edx],ax
					mov byte ptr [edi],0xFF
					jmp __m0_tile0_opaque_bld_hflip
__m0_tile0_invisible_bld_hflip:
					mov byte ptr [edi],0
__m0_tile0_opaque_bld_hflip:
					inc edi
					add edx,2
					dec ebx
					dec ecx
				jnz __m0_tile0_draw_trans_bld_hflip_draw

__m0_tile0_trans_bld_end:
			pop edi
			pop edx

			mov ecx,31
			__m0_draw_tiles_trans_bld:
				push ecx
				mov eax,[xtile]
				/*xor ebx,ebx
				and eax,[hTileMask] //0x1F
				mov bx,[esi+eax*2]*/
				lea ebx,[eax*8]
				and eax,0x1F
				and ebx,[hScrMask]
				mov ecx,[hScrShift]
				shl ebx,cl
				lea eax,[ebx+eax*2]
				xor ebx,ebx				// ebx = 0
				mov bx,[esi+eax]		// Get tile data

				mov eax,ebx
				inc dword ptr [xtile]
				mov [tileData],ebx
				shr eax,11
				and ebx,0x3FF
				and eax,1
				shl ebx,6
				add ebx,[tileOffset+eax*4]
				add ebx,[dummy] //edx

				mov ecx,-8
				push edx
				add edi,16
				mov edx,edi
				add edx,[vscreen_size]
				add edx,[vscreen_size]
				push edi
				mov edi,[p_alphamap_int]
				test dword ptr [tileData],0x400
				jnz __m0_draw_trans_bld_hflip
				__m0_draw_trans_bld:
					mov al,[ebx]
					and eax,0xFF
					jz __m0_invisible_bld
					mov ax,[ebp + eax*2]
					mov [edx + ecx*2],ax
					mov byte ptr [edi],0xFF
					jmp __m0_opaque_bld
__m0_invisible_bld:
					mov byte ptr [edi],0
__m0_opaque_bld:
					inc edi
					inc ebx
					inc ecx
				jnz __m0_draw_trans_bld
				jmp __m0_draw_trans_bld_next
// Draw hflipped tiles
__m0_draw_trans_bld_hflip:
				add ebx,7
				__m0_draw_trans_bld_hflip_draw:
					mov al,[ebx]
					and eax,0xFF
					jz __m0_invisible_bld_hflip
					mov ax,[ebp + eax*2]
					mov [edx + ecx*2],ax
					mov byte ptr [edi],0xFF
					jmp __m0_opaque_bld_hflip
__m0_invisible_bld_hflip:
					mov byte ptr [edi],0
__m0_opaque_bld_hflip:
					inc edi
					dec ebx
					inc ecx
				jnz __m0_draw_trans_bld_hflip_draw
__m0_draw_trans_bld_next:
				pop edi
				add dword ptr [p_alphamap_int],8
				pop edx
				pop ecx
				dec ecx
			jnz __m0_draw_tiles_trans_bld

__m0_tile31_trans_bld:
				mov eax,[xtile]
				/*xor ebx,ebx
				and eax,0x1f //[hTileMask] //0x1F
				mov bx,[esi+eax*2]*/
				lea ebx,[eax*8]
				and eax,0x1F
				and ebx,[hScrMask]
				mov ecx,[hScrShift]
				shl ebx,cl
				lea eax,[ebx+eax*2]
				xor ebx,ebx				// ebx = 0
				mov bx,[esi+eax]		// Get tile data

				mov eax,ebx
				inc dword ptr [xtile]
				mov [tileData],ebx
				shr eax,11
				and ebx,0x3FF
				and eax,1
				shl ebx,6
				add ebx,[tileOffset+eax*4]
				add ebx,[dummy] //edx

				mov ecx,[hofs]
				mov edx,edi
				and ecx,7
				jz __m0_tile31_trans_bld_end
				add edx,[vscreen_size]
				add edx,[vscreen_size]
				mov edi,[p_alphamap_int]
				test dword ptr [tileData],0x400
				jnz __m0_tile31_draw_trans_bld_hflip
				__m0_tile31_draw_trans_bld:
					mov al,[ebx]
					and eax,0xFF
					jz __m0_tile31_invisible_bld
					mov ax,[ebp + eax*2]
					mov [edx],ax
					mov byte ptr [edi],0xFF
					jmp __m0_tile31_opaque_bld
__m0_tile31_invisible_bld:
					mov byte ptr [edi],0
__m0_tile31_opaque_bld:
					inc edi
					add edx,2
					inc ebx
					dec ecx
				jnz __m0_tile31_draw_trans_bld
				jmp __m0_tile31_trans_bld_end

				// Draw a hflipped tile
__m0_tile31_draw_trans_bld_hflip:
					add ebx,7
				__m0_tile31_draw_trans_bld_hflip_draw:
					mov al,[ebx]
					and eax,0xFF
					jz __m0_tile31_invisible_bld_hflip
					mov ax,[ebp + eax*2]
					mov [edx],ax
					mov byte ptr [edi],0xFF
					jmp __m0_tile31_opaque_bld_hflip
__m0_tile31_invisible_bld_hflip:
					mov byte ptr [edi],0
__m0_tile31_opaque_bld_hflip:
					inc edi
					add edx,2
					dec ebx
					dec ecx
				jnz __m0_tile31_draw_trans_bld_hflip_draw

__m0_tile31_trans_bld_end:


__m0_done:
			pop ebp
			popa
		}
		}
	} else
///////////////////////////
// Handle 16x16 color BGs
///////////////////////////
		__asm
		{
			pusha
			mov eax,[scanline]
			dec eax
			mov edi,[vscreen_curr]
			shl eax,9
			add edi,eax				// edi = vscreen+scanline*512

			mov esi,[screenBank] //[VRAM_curr]
			mov eax,[scrBase]
			add eax,[mstrScrn]
			shl eax,11
			mov ebx,[scanline]
			dec ebx
			add ebx,[vofs]
			mov edx,ebx
			//and ebx,0xFF			// temp fix !!
			//mov ecx,[vScrShift]
			and edx,[vScrMask]
			//shl edx,cl
			shr edx,3
			mov ecx,[hTileMask]
			shr ecx,5
			add ecx,6
			shl edx,cl				// ebx = (scanline/8)*64
			add esi,edx
			add esi,eax
			mov ecx,ebx
			//add esi,ebx				// esi = nametable

			shr ecx,3
			and ecx,0x1F
			shl ecx,6
			add esi,ecx
			mov ecx,ebx

			and ecx,7

			mov ebx,ecx
			xor ebx,7				// "Inverted" tile row (used for vflipping)
			shl ebx,2
			shl ecx,2				// Multiply by 4 (number of bytes per row)
			mov [tileOffset+4],ebx
			mov [tileOffset],ecx

	mov [lineaddr],edi

			mov edx,[charBank] //[VRAM_curr]
			mov eax,[chrBase]
			add eax,[mstrChar]
			shl eax,14
			//shl ecx,2
			add edx,eax
			//add edx,ecx
			mov [dummy],edx

			//test dword ptr [bg],0xF
			//jnz __m0_not_bg0_16
			push ebp
			//mov ebp,[currPalette]

			mov eax,[hofs]
			mov ebx,eax
			shr eax,3				// eax = xscroll/8
			mov [xtile],eax
				lea ebx,[eax*8]
				and eax,0x1F
				and ebx,[hScrMask]
				mov ecx,[hScrShift]
				shl ebx,cl
				lea eax,[ebx+eax*2]
				xor ebx,ebx				// ebx = 0
				mov bx,[esi+eax]		// Get tile data

				inc dword ptr [xtile]	// Next tile..

				mov eax,ebx				// Get vflip flag
				shr eax,11
				and eax,1

			mov ebp,ebx
			mov [tileData],ebx
			and ebx,0x3FF			// ebx = tile number
			and ebp,0xF000
			shl ebx,5				// ebx *= 4*8
			shr ebp,7
			add ebx,edx				// ebx += VRAM + chrBase + tile_row*8
				add ebx,[tileOffset+eax*4]
			push esi
			add ebp,[currPalette]
			mov esi,[currPalette]
			movzx esi,word ptr [esi]
			mov eax,[p_alphamap]
			mov [p_alphamap_int],eax

			test [ablend],1
			jz __m0_16_no_ablend
				mov eax,[numBlends]
				inc eax
				add eax,eax
				mul dword ptr [vscreen_size]
				add edi,eax
__m0_16_no_ablend:

			mov eax,[hofs]
			//add ebx,8
			mov ecx,8
			and eax,7
			sub ecx,eax					// ecx = 8 - (xscroll&7)  (length of first tile)

			test dword ptr [tileData],0x400
			jnz __m0_16_tile0_hflip

			shr eax,1
			add ebx,eax
			test ecx,1
			jz __m0_16_tile0_even
			test dword ptr [newframe],0xF
			jnz __m0_16_transp_1
				xor eax,eax				// Clear eax
				mov al,[ebx]			// Load pixel
				shr eax,4
				jz __m0_16_transp_1_1
				mov ax,[ebp + eax*2]	// Get 15-bit color value
				mov [edi],ax			// Store
				add edi,2				// Increase destination pointer
				inc ebx
				dec ecx
				jmp __m0_16_tile0_even
			__m0_16_transp_1_1:
				mov [edi],si			// Store
				add edi,2				// Increase destination pointer
				inc ebx
				dec ecx
				jmp __m0_16_tile0_even

			__m0_16_transp_1:
				test [ablend],1
				jnz __m0_16_blend_1
				xor eax,eax				// Clear eax
				mov al,[ebx]			// Load pixel
				shr eax,4
				jz __m0_16_transp_2
				mov ax,[ebp + eax*2]	// Get 15-bit color value
				mov [edi],ax			// Store
				__m0_16_transp_2:
				add edi,2				// Increase destination pointer
				inc ebx
				dec ecx
				jmp __m0_16_tile0_even

			__m0_16_blend_1:
				mov esi,[p_alphamap_int]
				xor eax,eax				// Clear eax
				mov al,[ebx]			// Load pixel
				shr eax,4
				jz __m0_16_blend_2
				mov ax,[ebp + eax*2]	// Get 15-bit color value
				mov [edi],ax			// Store
				mov al,0xFF
				__m0_16_blend_2:
				mov [esi],al
				add edi,2				// Increase destination pointer
				inc dword ptr [p_alphamap_int]
				inc ebx
				dec ecx

__m0_16_tile0_even:
			//push esi
			test dword ptr [newframe],0xF
			jnz __m0_16_tile0_transp_setup
			//movzx esi,word ptr [ebp]
			__m0_16_tile0_draw:
				cmp ecx,2
				jl __m0_16_tile0_end
				xor eax,eax				// Clear eax
				mov al,[ebx]			// Load pixel
				mov edx,eax
				shr eax,4
				mov ax,[ebp + eax*2]	// Get 15-bit color value
				cmovz eax,esi
				shl eax,16
				and edx,0xF
				jz __m0_16_transp_2_1
				mov ax,[ebp + edx*2]
				mov [edi],eax			// Store
				add edi,4				// Increase destination pointer
				inc ebx
				sub ecx,2
			jmp __m0_16_tile0_draw
			__m0_16_transp_2_1:
				mov ax,si
				mov [edi],eax			// Store
				add edi,4				// Increase destination pointer
				inc ebx
				sub ecx,2
			jmp __m0_16_tile0_draw

			__m0_16_tile0_transp_setup:
				test [ablend],1
				jnz __m0_16_tile0_blend
			__m0_16_tile0_transp:
				cmp ecx,2
				jl __m0_16_tile0_end
				xor eax,eax				// Clear eax
				mov al,[ebx]			// Load pixel
				mov edx,eax
					and eax,0xF
					jz __m0_16_transp_4
					mov ax,[ebp + eax*2]
					mov [edi],ax
				__m0_16_transp_4:
					shr edx,4
					jz __m0_16_transp_5
					mov ax,[ebp + edx*2]
					mov [edi+2],ax
				__m0_16_transp_5:
				add edi,4				// Increase destination pointer
				inc ebx
				sub ecx,2
			jmp __m0_16_tile0_transp

			__m0_16_tile0_blend:
				cmp ecx,2
				jl __m0_16_tile0_end
				mov esi,[p_alphamap_int]
				add dword ptr [p_alphamap_int],2
				xor eax,eax				// Clear eax
				mov al,[ebx]			// Load pixel
				mov edx,eax
					and eax,0xF
					jz __m0_16_blend_4
					mov ax,[ebp + eax*2]
					mov [edi],ax
					mov ah,0xFF
				__m0_16_blend_4:
					mov [esi],ah
					shr edx,4
					jz __m0_16_blend_5
					mov ax,[ebp + edx*2]
					mov [edi+2],ax
					mov dh,0xFF
				__m0_16_blend_5:
				mov [esi+1],dh
				add edi,4				// Increase destination pointer
				inc ebx
				sub ecx,2
			jmp __m0_16_tile0_blend

__m0_16_tile0_hflip:

			mov eax,ecx
			shr eax,1
			add ebx,eax
			test ecx,1
			jz __m0_16_tile0_hf_even
			test dword ptr [newframe],0xF
			jnz __m0_16_hf_transp_1
				xor eax,eax				// Clear eax
				mov al,[ebx]			// Load pixel
				and eax,0xF
				jz __m0_16_hf_transp_1_1
				mov ax,[ebp + eax*2]	// Get 15-bit color value
				mov [edi],ax			// Store
				add edi,2				// Increase destination pointer
				dec ebx
				dec ecx
				jmp __m0_16_tile0_hf_even
			__m0_16_hf_transp_1_1:
				mov [edi],si			// Store
				add edi,2				// Increase destination pointer
				//dec ebx
				dec ecx
				jmp __m0_16_tile0_hf_even

			__m0_16_hf_transp_1:
				test [ablend],1
				jnz __m0_16_hf_blend_1
				xor eax,eax				// Clear eax
				mov al,[ebx]			// Load pixel
				and eax,0xF
				jz __m0_16_hf_transp_2
				mov ax,[ebp + eax*2]	// Get 15-bit color value
				mov [edi],ax			// Store
				__m0_16_hf_transp_2:
				add edi,2				// Increase destination pointer
				//dec ebx
				dec ecx
				jmp __m0_16_tile0_hf_even

			__m0_16_hf_blend_1:
				mov esi,[p_alphamap_int]
				xor eax,eax				// Clear eax
				mov al,[ebx]			// Load pixel
				and eax,0xF
				jz __m0_16_hf_blend_2
				mov ax,[ebp + eax*2]	// Get 15-bit color value
				mov [edi],ax			// Store
				mov al,0xFF
				__m0_16_hf_blend_2:
				mov [esi],al
				add edi,2				// Increase destination pointer
				inc dword ptr [p_alphamap_int]
				//dec ebx
				dec ecx

__m0_16_tile0_hf_even:
			test dword ptr [newframe],0xF
			jnz __m0_16_tile0_hf_transp_setup
			__m0_16_tile0_hf_draw:
				cmp ecx,2
				jl __m0_16_tile0_hf_end
				dec ebx
				xor eax,eax				// Clear eax
				mov al,[ebx]			// Load pixel
				mov edx,eax
				and eax,0xF
				mov ax,[ebp + eax*2]	// Get 15-bit color value
				cmovz eax,esi
				shl eax,16
				shr edx,4
				jz __m0_16_hf_transp_2_1
				mov ax,[ebp + edx*2]
				mov [edi],eax			// Store
				add edi,4				// Increase destination pointer
				//dec ebx
				sub ecx,2
			jmp __m0_16_tile0_hf_draw
			__m0_16_hf_transp_2_1:
				mov ax,si
				mov [edi],eax			// Store
				add edi,4				// Increase destination pointer
				//dec ebx
				sub ecx,2
			jmp __m0_16_tile0_hf_draw

			__m0_16_tile0_hf_transp_setup:
				test [ablend],1
				jnz __m0_16_tile0_hf_blend
			__m0_16_tile0_hf_transp:
				cmp ecx,2
				jl __m0_16_tile0_hf_end
				dec ebx
				xor eax,eax				// Clear eax
				mov al,[ebx]			// Load pixel
				mov edx,eax
					shr eax,4
					jz __m0_16_hf_transp_4
					mov ax,[ebp + eax*2]
					mov [edi],ax
				__m0_16_hf_transp_4:
					and edx,0xF
					jz __m0_16_hf_transp_5
					mov ax,[ebp + edx*2]
					mov [edi+2],ax
				__m0_16_hf_transp_5:
				add edi,4				// Increase destination pointer
				//dec ebx
				sub ecx,2
			jmp __m0_16_tile0_hf_transp

			__m0_16_tile0_hf_blend:
				cmp ecx,2
				jl __m0_16_tile0_hf_end
				mov esi,[p_alphamap_int]
				dec ebx
				add dword ptr [p_alphamap_int],2
				xor eax,eax				// Clear eax
				mov al,[ebx]			// Load pixel
				mov edx,eax
					shr eax,4
					jz __m0_16_hf_blend_4
					mov ax,[ebp + eax*2]
					mov [edi],ax
					mov ah,0xFF
				__m0_16_hf_blend_4:
					mov [esi],ah
					and edx,0xF
					jz __m0_16_hf_blend_5
					mov ax,[ebp + edx*2]
					mov [edi+2],ax
					mov dh,0xFF
				__m0_16_hf_blend_5:
				mov [esi+1],dh
				add edi,4				// Increase destination pointer
				//dec ebx
				sub ecx,2
			jmp __m0_16_tile0_hf_blend

__m0_16_tile0_hf_end:
__m0_16_tile0_end:
			pop esi

			mov ecx,31					// number of tiles
			__m0_16_draw_tiles:
				push ecx
				mov eax,[xtile]
				lea ebx,[eax*8]
				and eax,0x1f //[hTileMask] //0x1f //[hTileMask]
				and ebx,[hScrMask]
				mov ecx,[hScrShift]
				shl ebx,cl
				lea eax,[ebx+eax*2]
				xor ebx,ebx				// ebx = 0
				mov bx,[esi+eax]		// Get tile data

				inc dword ptr [xtile]
					mov eax,ebx				// Get vflip flag
					shr eax,11
					and eax,1
				mov ebp,ebx
				mov [tileData],ebx
				and ebx,0x3FF
				and ebp,0xF000
				shl ebx,5
				shr ebp,7
				add ebx,[dummy] //edx
					add ebx,[tileOffset+eax*4]
				push esi
				add ebp,[currPalette]
				mov esi,[currPalette]
				movzx esi,word ptr [esi]

				mov ecx,4
				test dword ptr [newframe],0xF
				jnz __m0_16_draw_transp
				test dword ptr [tileData],0x400
				jnz __m0_16_draw_hflip
				__m0_16_draw:
					xor eax,eax
					mov al,[ebx]
					mov edx,eax
					shr eax,4
					mov ax,[ebp + eax*2]
					cmovz eax,esi
					shl eax,16
					and edx,0xF
					jz __m0_16_transp_3_1
					mov ax,[ebp + edx*2]
					mov [edi],eax
					add edi,4
					inc ebx
					dec ecx
				jnz __m0_16_draw
				jmp __m0_16_draw_next
				__m0_16_transp_3_1:
					mov ax,si
					mov [edi],eax
					add edi,4
					inc ebx
					dec ecx
				jnz __m0_16_draw
				jmp __m0_16_draw_next
__m0_16_draw_hflip:
				add ebx,3
				add edi,16
				mov ecx,-4
				__m0_16_draw_hflip_draw:
					xor eax,eax
					mov al,[ebx]
					dec ebx
					mov edx,eax
					and eax,0xF
					mov ax,[ebp + eax*2]
					cmovz eax,esi
					shl eax,16
					shr edx,4
					jz __m0_16_transp_3_2
					mov ax,[ebp + edx*2]
					mov [edi+ecx*4],eax
					inc ecx
				jnz __m0_16_draw_hflip_draw
				jmp __m0_16_draw_next
				__m0_16_transp_3_2:
					mov ax,si
					mov [edi+ecx*4],eax
					inc ecx
				jnz __m0_16_draw_hflip_draw
				jmp __m0_16_draw_next

			__m0_16_draw_transp:
				test [ablend],1
				jnz __m0_16_draw_blend
				test dword ptr [tileData],0x400
				jnz __m0_16_draw_transp_hflip
			__m0_16_draw_transp_draw:
					xor eax,eax
					mov al,[ebx]
					mov edx,eax
					and eax,0xF
					jz __m0_16_transp_6
					mov ax,[ebp + eax*2]
					mov [edi],ax
				__m0_16_transp_6:
					shr edx,4
					jz __m0_16_transp_7
					mov ax,[ebp + edx*2]
					mov [edi+2],ax
				__m0_16_transp_7:
					add edi,4
					inc ebx
					dec ecx
				jnz __m0_16_draw_transp_draw
				jmp __m0_16_draw_next
__m0_16_draw_transp_hflip:
				add ebx,3
				add edi,16
				mov ecx,-4
			__m0_16_draw_transp_hflip_draw:
					xor eax,eax
					mov al,[ebx]
					dec ebx
					mov edx,eax
					shr eax,4
					jz __m0_16_transp_6_2
					mov ax,[ebp + eax*2]
					mov [edi + ecx*4],ax
				__m0_16_transp_6_2:
					and edx,0xF
					jz __m0_16_transp_7_2
					mov ax,[ebp + edx*2]
					mov [edi + 2 + ecx*4],ax
				__m0_16_transp_7_2:
					//add edi,4
					//inc ebx
					inc ecx
				jnz __m0_16_draw_transp_hflip_draw
				jmp __m0_16_draw_next


//////
			__m0_16_draw_blend:
				mov esi,[p_alphamap_int]
				test dword ptr [tileData],0x400
				jnz __m0_16_draw_blend_hflip
			__m0_16_draw_blend_draw:
					xor eax,eax
					mov al,[ebx]
					mov edx,eax
					and eax,0xF
					jz __m0_16_blend_6_3
					mov ax,[ebp + eax*2]
					mov [edi],ax
					mov al,0xFF
				__m0_16_blend_6_3:
					mov [esi],al
					inc esi
					shr edx,4
					jz __m0_16_blend_7_3
					mov ax,[ebp + edx*2]
					mov [edi+2],ax
					mov al,0xFF
				__m0_16_blend_7_3:
					mov [esi],al
					add edi,4
					inc ebx
					inc esi
					dec ecx
				jnz __m0_16_draw_blend_draw
				jmp __m0_16_draw_next
__m0_16_draw_blend_hflip:
			add ebx,4
			__m0_16_draw_blend_hflip_draw:
					dec ebx
					xor eax,eax
					mov al,[ebx]
					mov edx,eax
					shr eax,4
					jz __m0_16_blend_6_4
					mov ax,[ebp + eax*2]
					mov [edi],ax
					mov al,0xFF
				__m0_16_blend_6_4:
					mov [esi],al
					inc esi
					and edx,0xF
					jz __m0_16_blend_7_4
					mov ax,[ebp + edx*2]
					mov [edi+2],ax
					mov al,0xFF
				__m0_16_blend_7_4:
					mov [esi],al
					add edi,4
					//inc ebx
					inc esi
					dec ecx
				jnz __m0_16_draw_blend_hflip_draw
//////


__m0_16_draw_next:
				add dword ptr [p_alphamap_int],8
				pop esi
				pop ecx
				dec ecx
			jnz __m0_16_draw_tiles

__m0_16_tile31:
				mov eax,[xtile]
				lea ebx,[eax*8]
				and eax,0x1f //[hTileMask] //0x1f //[hTileMask]
				and ebx,[hScrMask]
				mov ecx,[hScrShift]
				shl ebx,cl
				lea eax,[ebx+eax*2]
				xor ebx,ebx				// ebx = 0
				mov bx,[esi+eax]		// Get tile data
				inc dword ptr [xtile]
					mov eax,ebx				// Get vflip flag
					shr eax,11
					and eax,1

				mov ebp,ebx
				and ebx,0x3FF
				and ebp,0xF000
				shl ebx,5
				shr ebp,7
				add ebx,[dummy] //edx
				add ebx,[tileOffset+eax*4]
				add ebp,[currPalette]

				mov ecx,[hofs]
				test dword ptr [newframe],0xF
				jnz __m0_16_tile31_transp_setup

				and ecx,7
				jz __m0_16_tile31_end
				__m0_16_tile31_draw:
					cmp ecx,2
					jl __m0_16_tile31_end
					xor eax,eax
					mov al,[ebx]
					mov edx,eax
					shr eax,4
					and edx,0xF
					mov ax,[ebp + eax*2]
					shl eax,16
					mov ax,[ebp + edx*2]
					mov [edi],eax
					add edi,4
					inc ebx
					sub ecx,2
				jmp __m0_16_tile31_draw
__m0_16_tile31_end:
			test ecx,1
			jz __m0_done_16 //__m0_16_tile31_even
				xor eax,eax				// Clear eax
				mov al,[ebx]			// Load pixel
				and eax,0xF
				mov ax,[ebp + eax*2]	// Get 15-bit color value
				mov [edi],ax			// Store
__m0_16_tile31_even:

__m0_16_tile31_transp_setup:
				and ecx,7
				jz __m0_16_tile31t_end
				test [ablend],1
				jnz __m0_16_tile31_blend
			__m0_16_tile31_transp:
				cmp ecx,2
				jl __m0_16_tile31t_end
				xor eax,eax				// Clear eax
				mov al,[ebx]			// Load pixel
				mov edx,eax
					and eax,0xF
					jz __m0_16_transp_8
					mov ax,[ebp + eax*2]
					mov [edi],ax
				__m0_16_transp_8:
					shr edx,4
					jz __m0_16_transp_9
					mov ax,[ebp + edx*2]
					mov [edi+2],ax
				__m0_16_transp_9:
				add edi,4				// Increase destination pointer
				inc ebx
				sub ecx,2
			jmp __m0_16_tile31_transp
__m0_16_tile31t_end:
			test ecx,1
			jz __m0_done_16
				xor eax,eax				// Clear eax
				mov al,[ebx]			// Load pixel
				and eax,0xF
				jz __m0_done_16
				mov ax,[ebp + eax*2]	// Get 15-bit color value
				mov [edi],ax			// Store
			jmp __m0_done_16

////
				and ecx,7
				jz __m0_16_tile31b_end
				mov esi,[p_alphamap_int]
			__m0_16_tile31_blend:
				cmp ecx,2
				jl __m0_16_tile31b_end
				xor eax,eax				// Clear eax
				mov al,[ebx]			// Load pixel
				mov edx,eax
					and eax,0xF
					jz __m0_16_blend_8
					mov ax,[ebp + eax*2]
					mov [edi],ax
					mov al,0xFF
				__m0_16_blend_8:
					mov [esi],al
					inc esi
					shr edx,4
					jz __m0_16_blend_9
					mov ax,[ebp + edx*2]
					mov [edi+2],ax
					mov al,0xFF
				__m0_16_blend_9:
				mov [esi],al
				add edi,4				// Increase destination pointer
				inc ebx
				inc esi
				sub ecx,2
			jmp __m0_16_tile31_blend
__m0_16_tile31b_end:
			test ecx,1
			jz __m0_done_16
				xor eax,eax				// Clear eax
				mov al,[ebx]			// Load pixel
				and eax,0xF
				jz __m0_16_blend_10
				mov ax,[ebp + eax*2]	// Get 15-bit color value
				mov [edi],ax			// Store
				mov al,0xFF
__m0_16_blend_10:
				mov [esi],al
			jmp __m0_done_16
////
__m0_done_16:
			pop ebp
			popa
		}


	////////////////////////////////
	// Perform the alpha blending
	////////////////////////////////
		if (ablend)
			alpha_blend();

	newframe++;
}





#define MAX_HTILE 0x1F


void draw_rotscale_bg(int bg)
{
	/*int chrBase,scrBase,cntReg,colMode;
	int i;

	// Get the control register offset of this background
	cntReg = IOREGS_OFFSET + 8 + (bg<<1);

	// Get the horizontal and vertical screen offsets
	hofs = IOREGS[0x10 + (bg<<2)]&0xFF;
	hofs += (IOREGS[0x11 + (bg<<2)]&0x1)<<8;
	vofs = IOREGS[0x12 + (bg<<2)]&0xFF;
	vofs += (IOREGS[0x13 + (bg<<2)]&0x1)<<8;

	// Get the CHR base, SCR base and color mode
	chrBase = (VRAM[cntReg] >> 2) & 3;
	scrBase = VRAM[cntReg+1] & 0x1F;
	colMode = VRAM[cntReg] & 0x80;

	charBank = NULL;
	screenBank = NULL;
	for (i=0; i<9; i++)
	{
		if (vramBank[i].type == MAIN_BG+core)
		{
			if ((chrBase >= vramBank[i].base-bgMemBase[core]) &&
				(chrBase < vramBank[i].base+vramBank[i].size-bgMemBase[core]))
			{
				charBank = vramBase[i];
				//break;
			}
			if (((scrBase>>3) >= vramBank[i].base-bgMemBase[core]) &&
				((scrBase>>3) < vramBank[i].base+vramBank[i].size-bgMemBase[core]))
			{
				screenBank = vramBase[i];
				if (charBank!=NULL) break;
			}
		}
	}
	if ((charBank==NULL)||(screenBank==NULL)) return;

	setup_color_effects(bg);

	lastbg = bg;

	*/if (bg == 2) {
		x2 = bg2x[core];
		y2 = bg2y[core];
		deltax = BG2PA[core];
		deltay = BG2PC[core];
	} else {
		x2 = bg3x[core];
		y2 = bg3y[core];
		deltax = BG3PA[core];
		deltay = BG3PC[core];
	}/*

	hTileMask = (VRAM[cntReg+1] >> 6) & 3;
	vTileMask = rsBgMasks[hTileMask][1];
	tileShift = rsBgShifts[hTileMask];
	hTileMask = rsBgMasks[hTileMask][0];*/

		int cntReg,ofsReg,colMode;
	int i,chrBaseDec,scrBaseDec;

	// Get the control register offset of this background
	cntReg = 0x08 + (bg<<1);
	ofsReg = 0x10 + (bg<<2);

	// Get the horizontal and vertical screen offsets
	hofs = io_base[ofsReg]&0xFF;
	hofs += (io_base[ofsReg+1]&1)<<8;
	vofs = io_base[ofsReg+2]&0xFF;
	vofs += (io_base[ofsReg+3]&1)<<8;

	// Get the CHR base, SCR base and color mode
	chrBase = (io_base[cntReg] >> 2) & 0xF;
	scrBase = io_base[cntReg+1] & 0x1F;
	colMode = io_base[cntReg] & 0x80;

	mstrChar = (io_base[3]&7)*4;
	mstrScrn = (io_base[3]&0x38)*4;

	hTileMask = (io_base[cntReg+1]>>6)&3;
	tileShift = rsBgShifts[hTileMask];
	hScrShift = 7 + hTileMask;
	vScrShift = hScrShift;
	if (hScrShift>9) hScrShift = 9;
	if (vScrShift>8) vScrShift--;
	hTileMask = (1<<hScrShift)-1;
	vTileMask = (1<<vScrShift)-1;

	charBank = NULL;
	screenBank = NULL;
	chrBaseDec = 0;
	scrBaseDec = 0;

	chrBase += mstrChar;
	scrBase += mstrScrn;

	/*wsprintf(buffer,"line %d: core=%d, bg=%d, scrBase=%d, charBase=%d, C.type=%d\n",
		scanline,core,bg,scrBase,chrBase,vramBank[2].type);
	WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);*/
	for (i=0; i<9; i++)
	{
		if (vramBank[i].type == MAIN_BG+core)
		{
			if ((chrBase >= vramBank[i].base-bgMemBase[core]) &&
				(chrBase < vramBank[i].base+vramBank[i].size-bgMemBase[core]))
			{
				charBank = vramBase[i];
				//break;
			} else if (chrBase >= vramBank[i].base-bgMemBase[core])
			{
				chrBaseDec += vramBank[i].size;
			}

			if (((scrBase>>3) >= vramBank[i].base-bgMemBase[core]) &&
				((scrBase>>3) < vramBank[i].base+vramBank[i].size-bgMemBase[core]))
			{
				screenBank = vramBase[i];
				if (charBank!=NULL) break;
			} else if ((scrBase>>3) >= vramBank[i].base-bgMemBase[core])
			{
				scrBaseDec += (vramBank[i].size<<3);
			}
		}
	}
	if ((charBank==NULL)||(screenBank==NULL))
	{
		if (newframe == 0) {
			flip_palette(core*2);
			draw_backdrop();
		}
		newframe++;
		return;
	}

	//check_ext_pltt_banks(bg);
	setup_color_effects(bg);

	lastbg = bg;

	chrBase -= mstrChar;
	scrBase -= mstrScrn;
	chrBase -= chrBaseDec;
	scrBase -= scrBaseDec;
	if ((chrBase<0) || (scrBase<0)) return;
	bgcnt &= 0xE0C3;
	bgcnt |= (chrBase << 2) | (scrBase << 8);


///////////////////////////
// Handle 256x1 color BGs
///////////////////////////
	if (colMode == 0x80)
		__asm
		{
			pusha
			mov eax,[scanline]
			mov ebx,512
			dec eax
			mov edi,[vscreen_curr]
			mul ebx
			add edi,eax				// edi = vscreen+scanline*480

			mov esi,[screenBank] //[VRAM]
			mov eax,[scrBase]
			shl eax,11				// Select 2k screen base block
			mov ebx,[scanline]
			dec ebx
			add ebx,[vofs]
			and ebx,0xFF			// temp fix !!
			mov ecx,ebx
			shr ebx,3
			shl ebx,5				// ebx = (scanline/8)*32
			add esi,eax
			//add esi,ebx				// esi = nametable

			and ecx,7				// ecx = (scanline+yscroll)&7  (tile row)
			mov edx,[charBank] //[VRAM]
			mov eax,[chrBase]
			shl eax,14				// Select 16k chr base block
			shl ecx,3				// Multiply by 8 (number of bytes per row)
			add edx,eax				// edx = VRAM + chrBase
			//add edx,ecx				// edx += tile_row*8

			mov [lineaddr],edi		// Save destination address for later use

			// Is this the bottom-most layer?
			test dword ptr [newframe],0xF
			jnz __m1_not_first_bg

__m1_first_bg:

			push ebp
			mov ebp,[currPalette]

			xor ecx,ecx
__m1_hloop:
			push ecx
			mov ebx,[x2]
			mov ebp,[y2]
			sar ebx,8
			sar ebp,8
			mov eax,ebx
			mov ecx,ebp
			sar ebx,3
			sar ebp,3
			and eax,7
			and ecx,7
			and ebx,[hTileMask]
			and ebp,[vTileMask]
			push ecx
			mov ecx,[tileShift]
			shl ebp,cl
			add ebx,ebp
			pop ecx
			movzx ebx,byte ptr [esi+ebx]
			shl ecx,3
			shl ebx,6
			mov ebp,[currPalette]
			add ebx,edx
			add eax,ecx
			movzx ecx,byte ptr [ebx+eax]
			mov ax,[ebp + ecx*2]
			mov [edi],ax
			mov eax,[deltax]
			mov ebx,[deltay]
			add edi,2
			add [x2],eax
			add [y2],ebx
			pop ecx
			inc ecx
			cmp ecx,256
			jne __m1_hloop

			jmp __m1_done

/////////////////////////////////////////
// Handle semi-transparent layers
/////////////////////////////////////////
__m1_not_first_bg:
			push ebp
			mov ebp,[currPalette]

			test [ablend],1
			jnz __m1_prepare_blend

			xor ecx,ecx
__m1_hloop_transp:
			push ecx
			mov ebx,[x2]
			mov ebp,[y2]
			test [bgcnt],0x2000
			jz __nowrap_t
			sar ebx,8
			sar ebp,8
			jmp __wrap_t
__nowrap_t:
			sar ebx,8
			js __m1_transp
			cmp ebx,[hTileMask]
			jg __m1_transp
			sar ebp,8
			js __m1_transp
			cmp ebp,[vTileMask]
			jg __m1_transp
__wrap_t:
			mov eax,ebx
			mov ecx,ebp
			sar ebx,3
			sar ebp,3
			and eax,7
			and ecx,7
			and ebx,[hTileMask]
			and ebp,[vTileMask]
			push ecx
			mov ecx,[tileShift]
			shl ebp,cl
			add ebx,ebp
			pop ecx
			movzx ebx,byte ptr [esi+ebx]
			shl ecx,3
			shl ebx,6
			mov ebp,[currPalette]
			add ebx,edx
			add eax,ecx
			movzx ecx,byte ptr [ebx+eax]
			cmp ecx,0
			jz __m1_transp
			mov ax,[ebp + ecx*2]
			mov [edi],ax
__m1_transp:
			mov eax,[deltax]
			mov ebx,[deltay]
			add edi,2
			add [x2],eax
			add [y2],ebx
			pop ecx
			inc ecx
			cmp ecx,256
			jne __m1_hloop_transp

			jmp __m1_done

/////////////////////////////////////////
// Handle alpha blended layers
/////////////////////////////////////////
__m1_prepare_blend:
			mov ebp,edi
			sub edi,[vscreen]
			xor ecx,ecx
			add edi,[vscreen2]
__m1_hloop_bld:
	/*push edx
	rdtsc
	mov dword ptr [fubar],eax
	pop edx*/
			push ecx
			push ebp
			mov ebx,[x2]
			mov ebp,[y2]
			sar ebx,8
			sar ebp,8
			mov eax,ebx
			mov ecx,ebp
			sar ebx,3
			sar ebp,3
			and eax,7
			and ecx,7
			and ebx,[hTileMask]
			and ebp,[vTileMask]
			push ecx
			mov ecx,[tileShift]
			shl ebp,cl
			add ebx,ebp
			pop ecx
			movzx ebx,byte ptr [esi+ebx]
			shl ecx,3
			shl ebx,6
			mov ebp,[currPalette]
			add ebx,edx
			add eax,ecx
			movzx ecx,byte ptr [ebx+eax]
			cmp ecx,0
			jz __m1_bld_transp
			mov ax,[ebp + ecx*2]
			mov [edi],ax
			pop ebp
			jmp __m1_bld_opaque
__m1_bld_transp:
			pop ebp
			mov ax,[ebp]
			mov [edi],ax
__m1_bld_opaque:
			mov eax,[deltax]
			add ebp,2
			mov ebx,[deltay]
			add edi,2
			add [x2],eax
			add [y2],ebx
			pop ecx
			inc ecx
/*push edx
rdtsc
sub eax,dword ptr [fubar]
mov dword ptr [fubar],eax
pop edx*/
			cmp ecx,256
			jne __m1_hloop_bld
__m1_done:
			pop ebp
			popa
		}

		if (ablend)
			alpha_blend();

		newframe++;
}



void draw_rs_bg(int bg) {
	int cntReg,colMode,bmpOffs;
	int i,chrBaseDec,scrBaseDec;

	// Get the control register offset of this background
	cntReg = 8 + (bg<<1);

	bgcnt = *(unsigned short*)(io_base + cntReg);

	// Get the horizontal and vertical screen offsets
	hofs = io_base[0x10 + (bg<<2)]&0xFF;
	hofs += (io_base[0x11 + (bg<<2)]&0x1)<<8;
	vofs = io_base[0x12 + (bg<<2)]&0xFF;
	vofs += (io_base[0x13 + (bg<<2)]&0x1)<<8;

	// Get the CHR base, SCR base and color mode
	chrBase = (io_base[cntReg] >> 2) & 15;
	scrBase = io_base[cntReg+1] & 0x1F;
	colMode = io_base[cntReg] & 0x80;

	mstrChar = (io_base[3]&7)*4;
	mstrScrn = (io_base[3]&0x38)*4;

	charBank = NULL;
	chrBaseDec = 0;
	screenBank = NULL;
	scrBaseDec = 0;

	chrBase += mstrChar;
	scrBase += mstrScrn;

	for (i=0; i<9; i++)
	{
		if (vramBank[i].type == MAIN_BG+core)
		{
			if ((chrBase >= vramBank[i].base-bgMemBase[core]) &&
				(chrBase < vramBank[i].base+vramBank[i].size-bgMemBase[core]))
			{
				charBank = vramBase[i];
			} else if (chrBase >= vramBank[i].base-bgMemBase[core])
			{
				chrBaseDec += vramBank[i].size;
			}

			if (((scrBase>>3) >= vramBank[i].base-bgMemBase[core]) &&
				((scrBase>>3) < vramBank[i].base+vramBank[i].size-bgMemBase[core]))
			{
				screenBank = vramBase[i];
				if (charBank!=NULL) break;
			} else if ((scrBase>>3) >= vramBank[i].base-bgMemBase[core])
			{
				scrBaseDec += vramBank[i].size;
			}
		}
	}
	if ((charBank==NULL)||(screenBank==NULL)) {
		if (newframe == 0) {
			flip_palette(core*2);
			draw_backdrop();
		}
		newframe++;
		return;
	}

	chrBase -= chrBaseDec;
	scrBase -= scrBaseDec;
	if ((chrBase<0) || (scrBase<0)) return;

	hTileMask = (io_base[cntReg+1]>>6)&3;
	hScrShift = 7 + hTileMask;
	vScrShift = hScrShift;
	if (hScrShift>9) hScrShift = 9;
	if (vScrShift>8) vScrShift--;
	hTileMask = (1<<hScrShift)-1;
	vTileMask = (1<<vScrShift)-1;

	if (bg == 2) {
		x2 = bg2x[core];
		y2 = bg2y[core];
		deltax = BG2PA[core];
		deltay = BG2PC[core];
	} else {
		x2 = bg3x[core];
		y2 = bg3y[core];
		deltax = BG3PA[core];
		deltay = BG3PC[core];
	}

	setup_color_effects(bg);

	lastbg = bg;

	bgcnt &= 0xE0C3;
	bgcnt |= (chrBase << 2) | (scrBase << 8);

	hTileMask = (io_base[cntReg+1] >> 6) & 3;
	vTileMask = rsBgMasks[hTileMask][1];
	tileShift = rsBgShifts[hTileMask];
	hTileMask = rsBgMasks[hTileMask][0];

	if (colMode) {
		//if (newframe == 0)
			bg_line_rs_bottom();
		/*else if (ablend)
			bg_line_rs_ablend();
		else
			bg_line_rs_trans();*/
	} else {
	}

	if (ablend)
		alpha_blend();

	newframe++;
}


void draw_ext_rs_bg(int bg) {
	int cntReg,colMode,bmpOffs;
	int i,chrBaseDec,scrBaseDec;

	// Get the control register offset of this background
	cntReg = 8 + (bg<<1);

	bgcnt = *(unsigned short*)(io_base + cntReg);

	// Get the horizontal and vertical screen offsets
	hofs = io_base[0x10 + (bg<<2)]&0xFF;
	hofs += (io_base[0x11 + (bg<<2)]&0x1)<<8;
	vofs = io_base[0x12 + (bg<<2)]&0xFF;
	vofs += (io_base[0x13 + (bg<<2)]&0x1)<<8;

	// Get the CHR base, SCR base and color mode
	chrBase = (io_base[cntReg] >> 2) & 15;
	scrBase = io_base[cntReg+1] & 0x1F;
	colMode = io_base[cntReg] & 0x80;

	mstrChar = (io_base[3]&7)*4;
	mstrScrn = (io_base[3]&0x38)*4;

	charBank = NULL;
	chrBaseDec = 0;
	screenBank = NULL;
	scrBaseDec = 0;

	/*wsprintf(buffer,"line %d: core=%d, bg=%d, bank=%d, chrBase=%d, scrBase=%d, bgcnt=%X,colMode=%X, ablend=%d, layer=%d,x2=%d,y2=%d,dx=%d,dy=%d, pb=%d, pd=%d\n",
		scanline-1,core,bg,i,chrBase,scrBase,bgcnt,colMode,ablend,newframe,x2,y2,deltax,deltay,
		BG3PB[core],BG3PD[core]);
	WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);*/

	if (colMode)
	{
		//bmpOffs = ((bgcnt>>8)&31)<<14;
		for (i=0; i<9; i++)
		{
			if (vramBank[i].type == MAIN_BG+core)
			{
				if (((scrBase) >= vramBank[i].base-bgMemBase[core]) &&
					((scrBase) < vramBank[i].base+vramBank[i].size-bgMemBase[core]))
				{
					//screenBank = vramBase[i];
					//if (charBank!=NULL) break;
					charBank = vramBase[i];
					break;
				} else if ((scrBase) >= vramBank[i].base-bgMemBase[core])
				{
					scrBaseDec += vramBank[i].size;
				}
			}
		}
		/*wsprintf(buffer,"line %d: core=%d, bg=%d, bank=%d, scrBase=%d, scrBank=%X,nf=%d,ab=%d\n",
			scanline-1,core,bg,i,scrBase,charBank,newframe,ablend);
		WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);*/

		if ((charBank==NULL))
		{
			if (newframe == 0)
			{
				flip_palette(core*2);
				draw_backdrop();
			}
			newframe++;
			return;
		}
	} else
	{
		chrBase += mstrChar;
		scrBase += mstrScrn;

		for (i=0; i<9; i++)
		{
			if (vramBank[i].type == MAIN_BG+core)
			{
				if ((chrBase >= vramBank[i].base-bgMemBase[core]) &&
					(chrBase < vramBank[i].base+vramBank[i].size-bgMemBase[core]))
				{
					charBank = vramBase[i];
				} else if (chrBase >= vramBank[i].base-bgMemBase[core])
				{
					chrBaseDec += vramBank[i].size;
				}

				if (((scrBase>>3) >= vramBank[i].base-bgMemBase[core]) &&
					((scrBase>>3) < vramBank[i].base+vramBank[i].size-bgMemBase[core]))
				{
					screenBank = vramBase[i];
					if (charBank!=NULL) break;
				} else if ((scrBase>>3) >= vramBank[i].base-bgMemBase[core])
				{
					scrBaseDec += vramBank[i].size;
				}
			}
		}
		if ((charBank==NULL)||(screenBank==NULL)) {
			if (newframe == 0) {
				flip_palette(core*2);
				draw_backdrop();
			}
			newframe++;
			return;
		}
	}

	chrBase -= chrBaseDec;
	scrBase -= scrBaseDec;
	if ((chrBase<0) || (scrBase<0)) return;

	hTileMask = (io_base[cntReg+1]>>6)&3;
	hScrShift = 7 + hTileMask;
	vScrShift = hScrShift;
	if (hScrShift>9) hScrShift = 9;
	if (vScrShift>8) vScrShift--;
	hTileMask = (1<<hScrShift)-1;
	vTileMask = (1<<vScrShift)-1;

	if (bg == 2) {
		x2 = bg2x[core];
		y2 = bg2y[core];
		deltax = BG2PA[core];
		deltay = BG2PC[core];
	} else
	{
		x2 = bg3x[core];
		y2 = bg3y[core];
		deltax = BG3PA[core];
		deltay = BG3PC[core];
	}

	setup_color_effects(bg);

	lastbg = bg;

	bgcnt &= 0xE0C3;
	bgcnt |= (chrBase << 2) | (scrBase << 8);

	/*wsprintf(buffer,"line %d: core=%d, bg=%d, bank=%d, chrBase=%d, scrBase=%d, bgcnt=%X,colMode=%X, ablend=%d, layer=%d,x2=%d,y2=%d\n",
		scanline-1,core,bg,i,chrBase,scrBase,bgcnt,colMode,ablend,newframe,x2,y2);
	WriteConsole(hout,buffer,lstrlen(buffer),(LPDWORD)&dummy,NULL);*/


	if (colMode) {
		if ((chrBase&1)==0)	{
			if (ablend)
				extbg_line_256_ablend();
			else if (newframe == 0)
				extbg_line_256_bottom();
			else
				extbg_line_256_trans();

		} else
		{
			// 16-bit bitmap
			if ((useGL && glOK && (glFormat==GL_UNSIGNED_SHORT_5_5_5_1)))
			{
				if (ablend)
					extbg_line_dc_ablend_bgr2rgb_ogl();
				else if (newframe == 0)
					extbg_line_dc_bottom_bgr2rgb_ogl();
				else
					extbg_line_dc_trans_bgr2rgb_ogl();
			} else if (useD3D)
			{
				if (newframe == 0)
					extbg_line_dc_bottom_bgr2rgb_d3d();
				else if (ablend)
					extbg_line_dc_ablend_bgr2rgb_d3d();
				else
					extbg_line_dc_trans_bgr2rgb_d3d();
			} else
			{
				if (newframe == 0)
					extbg_line_dc_bottom();
				else if (ablend)
					extbg_line_dc_ablend();
				else
					extbg_line_dc_trans();
			}
		}
	} else
	{
		check_ext_pltt_banks(bg);

		hTileMask = (io_base[cntReg+1] >> 6) & 3;
		vTileMask = rsBgMasks[hTileMask][1];
		tileShift = rsBgShifts[hTileMask];
		hTileMask = rsBgMasks[hTileMask][0];

		if ((useGL && glOK && (glFormat==GL_UNSIGNED_SHORT_5_5_5_1)))
		{
			if (newframe == 0)
				extbg_line_rs_bottom_bgr2rgb_ogl();
			else if (ablend)
				extbg_line_rs_ablend_bgr2rgb_ogl();
			else
				extbg_line_rs_trans_bgr2rgb_ogl();
		} else if (useD3D)
		{
			if (newframe == 0)
				extbg_line_rs_bottom_bgr2rgb_d3d();
			else if (ablend)
				extbg_line_rs_ablend_bgr2rgb_d3d();
			else
				extbg_line_rs_trans_bgr2rgb_d3d();
		} else
		{
			if (newframe == 0)
				extbg_line_rs_bottom();
			else if (ablend)
				extbg_line_rs_ablend();
			else
				extbg_line_rs_trans();
		}
	}

	if (ablend)
		alpha_blend();

	newframe++;
}



void draw_sprites()
{
	spriteBit6 = io_base[0]&0x10; //0x40;

	if (false/*forceLayer[4]*/ || (((io_base[1] & 0x10) != 0) && enableLayer[core][4])) //enableLayer[4]))
	{
		currPalette = pal_base + 0x200;
		if ((useGL && glOK && (glFormat==GL_UNSIGNED_SHORT_5_5_5_1)) || useD3D)
			flip_palette(core*2+1);

		scanline--;

		scanline++;
	} else
		__asm
		{
			emms
		}
}



// Draw all sprites with priority prio
void draw_sprites_prio(int prio)
{
	int i;
	spriteBit6 = io_base[0]&0x10; //0x40;

	if (forceLayer[4] || (((io_base[1] & 0x10) != 0) && enableLayer[core][4])) //enableLayer[4]))
	{
		currPalette = pal_base + 0x200;

		prio_int = prio;

		// Grab the first correctly setup bank
		for (i=0; i<9; i++)
		{
			if (vramBank[i].type == MAIN_OBJ+core)
			{
				patternTbl = vramBase[i];
				break;
			}
		}
		if (i==9) { __asm { emms } return; }

		cmTileShift = 5;
		if (io_base[0]&0x10) {
			cmTileShift += (io_base[2]>>4)&3;
		}

		// 128x128 2D
		// Lots of magic numbers follow..
		tileShift = 3+0;
		hTileMask = 0x3F0;
		vTileMask = 0x00F;
		hScrShift = 7;
		vScrShift = 4;
		hofs = 0;
		vofs = 11;

		if ((io_base[0]&0x60)==0x20)
		{
			// 256x64 2D
			tileShift = 3+0;
			hTileMask = 0x3E0;
			vTileMask = 0x01F;
			hScrShift = 7;
			vScrShift = 4;
			hofs = 0;
			vofs = 12;
		} else if ((io_base[0]&0x60)==0x40)
		{
			tileShift = 3+0;
			hTileMask = 0x3FF;
			vTileMask = 0x000;
			hScrShift = 7 + ((io_base[2]>>6)&1);
			vScrShift = 0;
			hofs = 112;
			vofs = 7; // + ((io_base[2]>>6)&1);	// Need to investigate if this is a robust solution.
		}

		//if (newframe==0)
		//	draw_backdrop();
		if ((useGL && glOK && (glFormat==GL_UNSIGNED_SHORT_5_5_5_1)) || (useD3D)) {
			if (newframe==0)
			{
				flip_palette(core*2);
				draw_backdrop();
			}
			flip_palette(core*2+1);
		}

		check_ext_pltt_banks(-1);
		setup_color_effects(-1);
		scanline--;
		
		obj_line();
		scanline++;

		//if (ablend)
		//	alpha_blend();
		
		newframe++;
		return;

	} else
		__asm
		{
			emms
		}
}



void apply_master_brightness() {
	unsigned char *surface = vscreen_curr+((scanline-1)<<9);
	unsigned int brightVal,brightType;

	brightVal = io_base[0x6C]&0x1F;
	if (brightVal>16) brightVal=16;
	brightType = io_base[0x6D]&0xC0;

	if ((brightVal) && (brightType!=0x00) && (brightType!=0xC0)) {
		// Set red/blue/green shift amounts depending on what rendering mode and
		// texture format is used.
		if (useGL && glOK && (glFormat==GL_UNSIGNED_SHORT_5_5_5_1))
		{
			*(int*)&SHAMT4 = 11;
			*(int*)&SHAMT5 = 1;
			*(int*)&SHAMT6 = 6;
			*(int*)&SHAMT1 = 11;
			*(int*)&SHAMT2 = 1;
			*(int*)&SHAMT3 = 6;
		} else if (useD3D)
		{
			*(int*)&SHAMT4 = 10;
			*(int*)&SHAMT5 = 0;
			*(int*)&SHAMT6 = 5;
			*(int*)&SHAMT1 = 10;
			*(int*)&SHAMT2 = 0;
			*(int*)&SHAMT3 = 5;
		} else
		{
			*(int*)&SHAMT4 = 0;
			*(int*)&SHAMT5 = 10;
			*(int*)&SHAMT6 = 5;
			*(int*)&SHAMT1 = 0;
			*(int*)&SHAMT2 = 10;
			*(int*)&SHAMT3 = 5;
		}

		if (brightType == 0x40)
			luma_inc(surface,surface,brightVal,false);
		else
			luma_dec(surface,surface,brightVal,false);
	}
}


void blend_lines_ogl(unsigned char *dest,unsigned char *srcA,unsigned char *srcB, int factorA,int factorB) {
	if (factorA>16) factorA=16;
	if (factorB>16) factorB=16;

	__asm {
		pusha
		mov			esi,[srcA]
		mov			edx,[srcB]
		mov			edi,[dest]
		movq		mm0,[MASKBLUE]
		movd		mm6,[factorB]
		movd		mm7,[factorA]
		punpcklwd	mm6,mm6
		punpcklwd	mm7,mm7
		punpckldq	mm6,mm6
		punpckldq	mm7,mm7
		mov			ecx,64

__blend_ogl_mmx:
		movq		mm2,[esi]		// mm2 = read from SrcA (5551)
		movq		mm1,[edx]		// mm1 = read from SrcB (1555)
		movq		mm3,mm1			// RED
		movq		mm4,mm2
		pand		mm3,mm0
		psrlw		mm4,11
		pand		mm4,mm0
		pmullw		mm3,mm6
		pmullw		mm4,mm7
		paddw		mm3,mm4
		psrlw		mm3,4
		movq		mm4,mm3
		pcmpgtw		mm4,mm0
		movq		mm1,mm4
		pand		mm4,mm0
		pandn		mm1,mm3
		por			mm4,mm1
		movq		mm5,mm4
		movq		mm1,[edx]

		movq		mm3,mm1		// GREEN
		movq		mm4,mm2
		psrlw		mm3,5
		psrlw		mm4,6
		pand		mm3,mm0
		pand		mm4,mm0
		pmullw		mm3,mm6
		pmullw		mm4,mm7
		paddw		mm3,mm4
		psrlw		mm3,4
		movq		mm4,mm3
		pcmpgtw		mm4,mm0
		movq		mm1,mm4
		pand		mm4,mm0
		pandn		mm1,mm3
		por			mm4,mm1
		psllw		mm4,5
		por			mm5,mm4
		movq		mm1,[edx]

		movq		mm3,mm1		// BLUE
		movq		mm4,mm2
		psrlw		mm3,10
		psrlw		mm4,1
		pand		mm3,mm0
		pand		mm4,mm0
		pmullw		mm3,mm6
		pmullw		mm4,mm7
		paddw		mm3,mm4
		psrlw		mm3,4
		movq		mm4,mm3
		pcmpgtw		mm4,mm0
		movq		mm1,mm4
		pand		mm4,mm0
		pandn		mm1,mm3
		por			mm4,mm1
		psllw		mm4,10
		por			mm5,mm4
		por			mm5,[ALPHAON]
		
		add			esi,8
		add			edx,8
		movq		[edi],mm5
		add			edi,8
		dec			ecx
		jnz			__blend_ogl_mmx
		popa
		emms
	}
}


void blend_lines(unsigned char *dest,unsigned char *srcA,unsigned char *srcB, int factorA,int factorB) {
	__asm {
		pusha
		mov			esi,[srcA]
		mov			edx,[srcB]
		mov			edi,[dest]
		movq		mm0,[MASKBLUE]
		movd		mm6,[factorB]
		movd		mm7,[factorA]
		punpcklwd	mm6,mm6
		punpcklwd	mm7,mm7
		punpckldq	mm6,mm6
		punpckldq	mm7,mm7
		mov			ecx,64

__blend__mmx:
		movq		mm2,[esi]		// mm2 = read from SrcA (1555)
		movq		mm1,[edx]		// mm1 = read from SrcB (1555)
		movq		mm3,mm1			// RED
		movq		mm4,mm2
		pand		mm3,mm0
		pand		mm4,mm0
		pmullw		mm3,mm6
		pmullw		mm4,mm7
		paddw		mm3,mm4
		psrlw		mm3,4
		movq		mm4,mm3
		pcmpgtw		mm4,mm0
		movq		mm1,mm4
		pand		mm4,mm0
		pandn		mm1,mm3
		por			mm4,mm1
		movq		mm5,mm4
		movq		mm1,[edx]

		movq		mm3,mm1		// GREEN
		movq		mm4,mm2
		psrlw		mm3,5
		psrlw		mm4,5
		pand		mm3,mm0
		pand		mm4,mm0
		pmullw		mm3,mm6
		pmullw		mm4,mm7
		paddw		mm3,mm4
		psrlw		mm3,4
		movq		mm4,mm3
		pcmpgtw		mm4,mm0
		movq		mm1,mm4
		pand		mm4,mm0
		pandn		mm1,mm3
		por			mm4,mm1
		psllw		mm4,5
		por			mm5,mm4
		movq		mm1,[edx]

		movq		mm3,mm1		// BLUE
		movq		mm4,mm2
		psrlw		mm3,10
		psrlw		mm4,10
		pand		mm3,mm0
		pand		mm4,mm0
		pmullw		mm3,mm6
		pmullw		mm4,mm7
		paddw		mm3,mm4
		psrlw		mm3,4
		movq		mm4,mm3
		pcmpgtw		mm4,mm0
		movq		mm1,mm4
		pand		mm4,mm0
		pandn		mm1,mm3
		por			mm4,mm1
		psllw		mm4,10
		por			mm5,mm4
		por			mm5,[ALPHAON]
		
		add			esi,8
		add			edx,8
		movq		[edi],mm5
		add			edi,8
		dec			ecx
		jnz			__blend__mmx
		popa
		emms
	}
}


void blend_lines_d3d(unsigned char *dest,unsigned char *srcA,unsigned char *srcB, int factorA,int factorB) {
	__asm {
		pusha
		mov			esi,[srcA]
		mov			edx,[srcB]
		mov			edi,[dest]
		movq		mm0,[MASKBLUE]
		movd		mm6,[factorB]
		movd		mm7,[factorA]
		punpcklwd	mm6,mm6
		punpcklwd	mm7,mm7
		punpckldq	mm6,mm6
		punpckldq	mm7,mm7
		mov			ecx,64

__blend_d3d_mmx:
		movq		mm2,[esi]		// mm2 = read from SrcA (1555)
		movq		mm1,[edx]		// mm1 = read from SrcB (1555)
		movq		mm3,mm1			// RED
		movq		mm4,mm2
		pand		mm3,mm0
		psrlw		mm4,10
		pand		mm4,mm0
		pmullw		mm3,mm6
		pmullw		mm4,mm7
		paddw		mm3,mm4
		psrlw		mm3,4
		movq		mm4,mm3
		pcmpgtw		mm4,mm0
		movq		mm1,mm4
		pand		mm4,mm0
		pandn		mm1,mm3
		por			mm4,mm1
		movq		mm5,mm4
		movq		mm1,[edx]

		movq		mm3,mm1		// GREEN
		movq		mm4,mm2
		psrlw		mm3,5
		psrlw		mm4,5
		pand		mm3,mm0
		pand		mm4,mm0
		pmullw		mm3,mm6
		pmullw		mm4,mm7
		paddw		mm3,mm4
		psrlw		mm3,4
		movq		mm4,mm3
		pcmpgtw		mm4,mm0
		movq		mm1,mm4
		pand		mm4,mm0
		pandn		mm1,mm3
		por			mm4,mm1
		psllw		mm4,5
		por			mm5,mm4
		movq		mm1,[edx]

		movq		mm3,mm1		// BLUE
		movq		mm4,mm2
		psrlw		mm3,10
		pand		mm3,mm0
		pand		mm4,mm0
		pmullw		mm3,mm6
		pmullw		mm4,mm7
		paddw		mm3,mm4
		psrlw		mm3,4
		movq		mm4,mm3
		pcmpgtw		mm4,mm0
		movq		mm1,mm4
		pand		mm4,mm0
		pandn		mm1,mm3
		por			mm4,mm1
		psllw		mm4,10
		por			mm5,mm4
		por			mm5,[ALPHAON]
		
		add			esi,8
		add			edx,8
		movq		[edi],mm5
		add			edi,8
		dec			ecx
		jnz			__blend_d3d_mmx
		popa
		emms
	}
}


void perform_capture() {
	unsigned char *src,*dest;

	if (core==0) {
		if (linesCaptured!=(scanline-1))
			return;

		if (io_base[0x67]&0x80) {
			src  = vramBase[(io_base[2]>>2)&3] + ((io_base[0x66]&0xC)<<13);
			dest = vramBase[io_base[0x66]&3] + ((io_base[0x67]&0xC)<<13);
			if ((io_base[0x67]&1)==0) {							// Src A = All
				//wsprintf(buffer,"Capturing line %d to bank %d\n",(scanline-1),io_base[0x66]&3);
				//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
				if ((io_base[0x67]&2)==0) {						// Src B = VRAM
					if ((io_base[0x67]&0x60)==0x00) {			// Src A
						if (useGL) {
							flip_colors_5551_to_1555(dest+((scanline-1)<<9) ,vscreen_curr+((scanline-1)<<9), 512);
						} else {
							__asm
							{
								pusha
								mov eax,[scanline]
								dec eax
								mov esi,[vscreen_curr]
								mov edi,[dest]
								shl eax,9					// eax = scanline*512
								add esi,eax
								add edi,eax
								mov ecx,128
								rep movsd
								popa
							}
						}
					} else if ((io_base[0x67]&0x60)==0x20) {	// Src B
						if (useGL) {
							flip_colors_5551_to_1555(dest+((scanline-1)<<9), src+((scanline-1)<<9), 512);
						} else {
							__asm
							{
								pusha
								mov eax,[scanline]
								dec eax
								mov esi,[src]
								mov edi,[dest]
								shl eax,9					// eax = scanline*512
								add esi,eax
								add edi,eax
								mov ecx,128
								rep movsd
								popa
							}
						}
					} else {								// Src A + B blended
						if ((useGL) && (glFormat==GL_UNSIGNED_SHORT_5_5_5_1))
							blend_lines_ogl(dest+((scanline-1)<<9), vscreen_curr+((scanline-1)<<9),
							                src+((scanline-1)<<9), 
							                (io_base[0x64])&0x1F,(io_base[0x65])&0x1F);
						else if (useD3D)
							blend_lines_d3d(dest+((scanline-1)<<9), vscreen_curr+((scanline-1)<<9),
							                src+((scanline-1)<<9), 
							                (io_base[0x64])&0x1F,(io_base[0x65])&0x1F);
						else
							blend_lines(dest+((scanline-1)<<9), vscreen_curr+((scanline-1)<<9),
										src+((scanline-1)<<9), 
							            (io_base[0x64])&0x1F,(io_base[0x65])&0x1F);
					}
				}
			}
			linesCaptured++;
		}
	}
}


#define REG_BG0CNT 0x008
#define REG_BG1CNT 0x00a
#define REG_BG2CNT 0x00c
#define REG_BG3CNT 0x00e

int uMask,vMask;

void draw_scanline(unsigned char *ssrc,unsigned char *sdst,int nScreen) {
	int tilebias = 0,i,j,regofs, objprio;
	bool glRenderer = (useGL && glOK);

	io_base    = VRAM + IOREGS_OFFSET + (nScreen<<12);
	oam_base   = VRAM + SPRITE_TABLE_OFFSET + (nScreen<<10);
	pal_base   = VRAM + PALETTE_OFFSET + (nScreen<<10);
	patternTbl = VRAM + 0x88000 + (nScreen << 18);

	core       = nScreen;


	VRAM_curr    = ssrc;
	vscreen_curr = sdst;
	if (core == 1)
		regofs   = 0x1000;
	else
		regofs   = 0;

	numBlends = 0;

	if (nScreen < 2) { //== 0) {
		// Read/update rotation/scaling parameters
		if (scanline==1) {
			BG2X[core] = (*(unsigned int *)(io_base + 0x28)) & 0x0FFFFFFF;
			BG2Y[core] = (*(unsigned int *)(io_base + 0x2C)) & 0x0FFFFFFF;
			BG3X[core] = (*(unsigned int *)(io_base + 0x38)) & 0x0FFFFFFF;
			BG3Y[core] = (*(unsigned int *)(io_base + 0x3C)) & 0x0FFFFFFF;
			__asm {
				push ecx
				mov ecx,core
				shl ecx,2
				mov eax,[BG2X+ecx]
				shl eax,4
				sar eax,4
				mov [bg2x+ecx],eax
				mov eax,[BG2Y+ecx]
				shl eax,4
				sar eax,4
				mov [bg2y+ecx],eax
				mov eax,[BG3X+ecx]
				shl eax,4
				sar eax,4
				mov [bg3x+ecx],eax
				mov eax,[BG3Y+ecx]
				shl eax,4
				sar eax,4
				mov [bg3y+ecx],eax
				pop ecx
			}
			BG2PA[core] = (*(unsigned int *)(io_base + 0x20)) & 0xFFFF;
			BG2PB[core] = (*(unsigned int *)(io_base + 0x22)) & 0xFFFF;
			BG2PC[core] = (*(unsigned int *)(io_base + 0x24)) & 0xFFFF;
			BG2PD[core] = (*(unsigned int *)(io_base + 0x26)) & 0xFFFF;
			BG3PA[core] = (*(unsigned int *)(io_base + 0x30)) & 0xFFFF;
			BG3PB[core] = (*(unsigned int *)(io_base + 0x32)) & 0xFFFF;
			BG3PC[core] = (*(unsigned int *)(io_base + 0x34)) & 0xFFFF;
			BG3PD[core] = (*(unsigned int *)(io_base + 0x36)) & 0xFFFF;
			__asm {
				push ecx
				mov ecx,core
				shl ecx,2
				shl dword ptr [BG2PA+ecx],16
				sar dword ptr [BG2PA+ecx],16
				shl dword ptr [BG2PB+ecx],16
				sar dword ptr [BG2PB+ecx],16
				shl dword ptr [BG2PC+ecx],16
				sar dword ptr [BG2PC+ecx],16
				shl dword ptr [BG2PD+ecx],16
				sar dword ptr [BG2PD+ecx],16
				shl dword ptr [BG3PA+ecx],16
				sar dword ptr [BG3PA+ecx],16
				shl dword ptr [BG3PB+ecx],16
				sar dword ptr [BG3PB+ecx],16
				shl dword ptr [BG3PC+ecx],16
				sar dword ptr [BG3PC+ecx],16
				shl dword ptr [BG3PD+ecx],16
				sar dword ptr [BG3PD+ecx],16
				pop ecx
			}
		} else
		{

			//if ((BG2X != ((*(unsigned int *)(IOREGS + 0x28)) & 0x0FFFFFFF)) ||
			//	(BG2Y != ((*(unsigned int *)(IOREGS + 0x2C)) & 0x0FFFFFFF)))
			if (io_base[BG2X_DIRTY]!=0)	{
				BG2X[core] = (*(unsigned int *)(io_base + 0x28)) & 0x0FFFFFFF;
				BG2Y[core] = (*(unsigned int *)(io_base + 0x2C)) & 0x0FFFFFFF;
				__asm {
					push ecx
					mov ecx,core
				//xor edi,edi
				//div edi
					shl ecx,2
					mov eax,[BG2X+ecx]
					shl eax,4
					sar eax,4
					mov [bg2x+ecx],eax
					mov eax,[BG2Y+ecx]
					shl eax,4
					sar eax,4
					mov [bg2y+ecx],eax
					pop ecx
				}
			} else {
				bg2x[core] += BG2PB[core];
				bg2y[core] += BG2PD[core];
			}

			//bg2y = (scanline-1)<<8;

			BG2PA[core] = (*(unsigned int *)(io_base + 0x20)) & 0xFFFF;
			BG2PB[core] = (*(unsigned int *)(io_base + 0x22)) & 0xFFFF;
			BG2PC[core] = (*(unsigned int *)(io_base + 0x24)) & 0xFFFF;
			BG2PD[core] = (*(unsigned int *)(io_base + 0x26)) & 0xFFFF;
			__asm {
				push ecx
				mov ecx,core
				shl ecx,2
				shl dword ptr [BG2PA+ecx],16
				sar dword ptr [BG2PA+ecx],16
				shl dword ptr [BG2PB+ecx],16
				sar dword ptr [BG2PB+ecx],16
				shl dword ptr [BG2PC+ecx],16
				sar dword ptr [BG2PC+ecx],16
				shl dword ptr [BG2PD+ecx],16
				sar dword ptr [BG2PD+ecx],16
				pop ecx
			}

			//if ((BG3X != ((*(unsigned int *)(IOREGS + 0x38)) & 0x0FFFFFFF)) ||
			//	(BG3Y != ((*(unsigned int *)(IOREGS + 0x3C)) & 0x0FFFFFFF)))
			if (io_base[BG3X_DIRTY]!=0)
			{
				BG3X[core] = (*(unsigned int *)(io_base + 0x38)) & 0x0FFFFFFF;
				BG3Y[core] = (*(unsigned int *)(io_base + 0x3C)) & 0x0FFFFFFF;
				__asm {
					push ecx
					mov ecx,core
				//xor edi,edi
				//div edi
					shl ecx,2
					mov eax,[BG3X+ecx]
					shl eax,4
					sar eax,4
					mov [bg3x+ecx],eax
					mov eax,[BG3Y+ecx]
					shl eax,4
					sar eax,4
					mov [bg3y+ecx],eax
					pop ecx
				}
			} else
			{
				bg3x[core] += BG3PB[core];
				bg3y[core] += BG3PD[core];
			}

			BG3PA[core] = (*(unsigned int *)(io_base + 0x30)) & 0xFFFF;
			BG3PB[core] = (*(unsigned int *)(io_base + 0x32)) & 0xFFFF;
			BG3PC[core] = (*(unsigned int *)(io_base + 0x34)) & 0xFFFF;
			BG3PD[core] = (*(unsigned int *)(io_base + 0x36)) & 0xFFFF;
			__asm {
				push ecx
				mov ecx,core
				shl ecx,2
				shl dword ptr [BG3PA+ecx],16
				sar dword ptr [BG3PA+ecx],16
				shl dword ptr [BG3PB+ecx],16
				sar dword ptr [BG3PB+ecx],16
				shl dword ptr [BG3PC+ecx],16
				sar dword ptr [BG3PC+ecx],16
				shl dword ptr [BG3PD+ecx],16
				sar dword ptr [BG3PD+ecx],16
				pop ecx
			}
		}

		io_base[BG2X_DIRTY] = 0;
		io_base[BG3X_DIRTY] = 0;
	}


	// Have we skipped enough frames?
	if (frames != frameskip)
		return;

	isFlipped[0] = isFlipped[1] = isFlipped[2] = isFlipped[3] = false;

	// Get mosaic values for BG2
	mosaicX = mosaicY = 0;
	if ((VRAM[IOREGS_OFFSET+0x0C] & 0x40) == 0x40) {
		mosaicX = VRAM[IOREGS_OFFSET+0x4C] & 0xF;
		mosaicY = (VRAM[IOREGS_OFFSET+0x4C] >> 4) & 0xF;
	}

	currPalette = pal_base;
	if ((io_base[2]&0x3) == 1)  // Changed 0xF to 0x3 [060203]
	{
		if ((io_base[1] &
			((enableLayer[core][0]&1)+((enableLayer[core][1]&1)<<1)+
			 ((enableLayer[core][2]&1)<<2)+((enableLayer[core][3]&1)<<3)))==0)
		{
			//while (1) { }
			flip_palette(core*2);
			draw_backdrop();
			for (i=1; i<5; i++)
				draw_sprites_prio(4-i);
			newframe = 0;
			perform_capture();
			apply_master_brightness();
			return;
		}
	}

	j = (io_base[0] & 7);

	objprio = 3;

	if (j<6)
	{
		// Handle modes 0-5
		if ((io_base[2]&0x3) == 1)	// Changed 0xF to 0x3 [060203]
		{
			// Sort BGs
			for (i=0; i<4; i++)
				layers[i].prio = io_base[8+(i<<1)]&3; //io_base[8 + (i<<1)]&0x3;
			layers[4].prio = 10;
			sort_layers(5);

			lastbg = 5; // any value > 4

			for (i=1; i<5; i++)
			{
				if ((io_base[1] & (1<<sorted_layers[i].num))!=0)
					for (; objprio > sorted_layers[i].prio; objprio--)
						draw_sprites_prio(objprio);

				if (false || //forceLayer[sorted_layers[i].num] ||
					(((io_base[1] & (1<<sorted_layers[i].num))!=0) && enableLayer[core][sorted_layers[i].num]))
				{
					if ((core!=0)||(sorted_layers[i].num!=0)||((IOREGS[0]&8)==0)) {
						draw_bg = draw_bg_callbacks[j][sorted_layers[i].num];
						if (draw_bg != NULL)
							draw_bg(sorted_layers[i].num);
					}
				} 
			}

			for (; objprio >= 0; objprio--)
				draw_sprites_prio(objprio);

			newframe = 0;
			blended2 = blended;
			blended = 0;
			perform_capture();
			apply_master_brightness();
			return;
		}
	}


	currPalette = pal_base;
	//if ((j==4) && glRenderer && (glFormat==GL_UNSIGNED_SHORT_5_5_5_1))
	//	flip_palette(0);


	if (nScreen == 0)
	{
		if ((io_base[2]&0x3) != 1)		// VRAM or FIFO display mode
		{
			if ((io_base[2]&0x3) == 3)	// FIFO display mode
			{
				if (mmu_get(MMU_DISP_FIFO))
					flip_colors(vscreen+((scanline-1)<<9) ,((unsigned char*)mmu_get(MMU_DISP_FIFO))+((scanline-1)<<9), 512);
		
				apply_master_brightness();
				return;
			}

		//wsprintf(buffer,"line %d: io_base[0x%x] = 0x%x\n",scanline,0x240+((io_base[2]&0xC)>>2),io_base[VRAMCNT_A+((io_base[2]&0xC)>>2)]);
		//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

			// make sure the bank is allocated to the LCDC
			if (io_base[VRAMCNT_A + ((io_base[2]&0xC)>>2)] == 0x80)
			{
				if (useD3D) {
					flip_colors_x555(vscreen+((scanline-1)<<9) ,VRAM_A+((io_base[2]&0xC)<<15)+((scanline-1)<<9), 512);
				} else if (useGL) {
					flip_colors(vscreen+((scanline-1)<<9) ,VRAM_A+((io_base[2]&0xC)<<15)+((scanline-1)<<9), 512);
				} else {
					__asm
					{
						pusha
						mov eax,[scanline]
						dec eax
						mov edi,[vscreen]
						shl eax,9					// eax = scanline*512
						mov esi,[VRAM_A]
						mov edx,[VRAM]
						add edi,eax					// edi = vscreen+scanline*512
						mov ecx,[edx+IOREGS_OFFSET]	// DISPCNT
						and ecx,0xF0000				// Select right bank
						add esi,eax
						sub ecx,0x20000
						js __bad_bank
						shr ecx,1
						add esi,ecx
				__bad_bank:
						mov ecx,128
						rep movsd
						popa
					}
				}
				for (i=1; i<5; i++)
					draw_sprites_prio(4-i);
				perform_capture();
				apply_master_brightness();
				return;
			} else
			{
				for (i=1; i<5; i++)
					draw_sprites_prio(4-i);
			//draw_sprites();
				return;
			}
		}

		// Mode 6 ?
		switch (j)
		{
			case 6:
				// Make sure that at least bank A is allocated to BG
				if (IOREGS[VRAMCNT_A] == 0x81)
				{
					if (glFormat == GL_UNSIGNED_SHORT_5_5_5_1)
						flip_palette(core*2);

					bgcnt = *(unsigned short*)(VRAM + IOREGS_OFFSET + 8 + (2<<1));

					x2 = bg2x[0];
					y2 = bg2y[0];
					deltax = BG2PA[0];
					deltay = BG2PC[0];

					mode6_line();

					for (i=1; i<5; i++)
						draw_sprites_prio(4-i);
					return;
					break;
				}
			}
		}

	draw_sprites();
	return;

	// Is BG2 enabled?

	if (forceLayer[2] || (((VRAM[IOREGS_OFFSET+1] & 0x4) != 0) && enableLayer[0][2]))
	{

		switch (VRAM[IOREGS_OFFSET] & 7)
		{
		case 4:
			if ((mosaicX==0) && (mosaicY==0))
			{
				if ((bg2x==0)&&(bg2y==0)&&(BG2PA==0)&&(BG2PB==0)&&(BG2PC==0))
				{
					__asm
					{
						pusha
						mov eax,[scanline]
						mov ebx,480
						dec eax
						mov edi,[vscreen]
						mul ebx
						mov esi,[VRAM]
						add edi,eax				// edi = vscreen+scanline*480
						movzx ecx,byte ptr [esi+IOREGS_OFFSET]
						shr eax,1
						and ecx,0x10			// ecx = x0000 (frame buffer selection)
						add esi,eax				// esi = VRAM+scanline*240
						shl ecx,9				// ecx = x0000000000000
						lea ecx,[ecx+ecx*4]		// ecx = x0x0000000000000 (0 : A000)
						//mov ebx,[VRAM]
						add esi,ecx

						//add ebx,PALETTE_OFFSET
						mov ebx,[currPalette]
						add esi,240				// The add is done to allow negative indexing
						add edi,480				// ..dito
						mov ecx,-240
						xor eax,eax
						__bm_pixel_loop:
							mov al,[esi+ecx]	// Read one byte from VRAM
							mov dx,[ebx+eax*2]	// Look it up in the color table
							mov [edi+ecx*2],dx	// Store it in the virtual screen
							inc ecx				// Next pixel..
						jnz __bm_pixel_loop
						popa
					}
				} else
				{
					__asm
					{
						pusha
						mov eax,[scanline]
						mov ebx,480
						dec eax
						mov edi,[vscreen]
						mov ecx,eax
						mov esi,[VRAM]
						mul ebx
						add edi,eax
						movzx edx,byte ptr [esi+IOREGS_OFFSET]
						and edx,0x10			// ecx = x0000 (frame buffer selection)
						shl edx,9				// ecx = x0000000000000
						lea edx,[edx+edx*4]		// ecx = x0x0000000000000 (0 : A000)
						push ebp
						add esi,edx
						mov ecx,240
						mov eax,bg2y
						mov ebp,eax				// ebp = bg2y*240
						mov eax,[BG2PC]
						mov ebx,eax				// ebx = dy*240
						mov edx,bg2x			// edx = bg2x
						mov ecx,0
		mode_4_draw:
						push esi
						mov eax,edx
						sar eax,8
						js mode_4_blank			// u+x<0 ?
						cmp eax,240
						jge mode_4_blank		// u+x>=240 ?
						add esi,eax
						mov eax,ebp				// v
						sar eax,8				// shift out fractional bits
						js mode_4_blank			// v+y<0 ?
						cmp eax,160
						jge mode_4_blank		// v+y>240*160
						shl eax,8
						add esi,eax
						shr eax,4
						sub esi,eax
						mov al,[esi]
						and eax,255
						mov esi,[currPalette]
						mov ax,[esi+eax*2]
						mov [edi],ax
						add edx,[BG2PA]
						add ebp,ebx
						add edi,2
						inc ecx
						pop esi
						cmp ecx,240
						jne mode_4_draw
						jmp mode_4_done
		mode_4_blank:
						mov esi,[colorLUT]
						mov ax,[esi]
						mov [edi],ax
						add edx,[BG2PA]
						add ebp,ebx
						add edi,2
						inc ecx
						pop esi
						cmp ecx,240
						jne mode_4_draw
		mode_4_done:
						pop ebp
						popa
					}
				}
			} else
				__asm
				{
					pusha
					mov eax,[scanline]
					dec eax
					mov ebx,480
					mov edi,[vscreen]
					mul ebx
					mov esi,[VRAM]
					add edi,eax				// edi = vscreen+scanline*480
					movzx ecx,byte ptr [esi+IOREGS_OFFSET]
					shr eax,1
					push eax
					and ecx,0x10			// ecx = x0000 (frame buffer selection)
					mov eax,[mosaicY]
					inc eax
					mov ebx,240
					mul ebx
					mov ebx,eax
					pop eax
					cdq
					div ebx
					mul ebx
					add esi,eax				// esi = VRAM+scanline*240
					shl ecx,9				// ecx = x0000000000000
					lea ecx,[ecx+ecx*4]		// ecx = x0x0000000000000 (0 : A000)
					//mov ebx,[VRAM]
					add esi,ecx

					push ebp
					//add ebx,PALETTE_OFFSET
					mov ebx,[currPalette]
					add edi,480				// ..dito
					mov ecx,-240
					mov ebp,[mosaicX]
					xor eax,eax
					__mos_bm_pixel_loop:
						mov al,[esi]		// Read one byte from VRAM
						mov dx,[ebx+eax*2]	// Look it up in the color table
						mov [edi+ecx*2],dx	// Store it in the virtual screen
						dec ebp
						jns __mos_ok
							inc esi
							mov ebp,[mosaicX]
							add esi,ebp
						__mos_ok:
						inc ecx				// Next pixel..
					jnz __mos_bm_pixel_loop
					pop ebp
					popa
				}
			break;

		case 3:
			if (useGL && glOK && (glFormat==GL_UNSIGNED_SHORT_5_5_5_1))
				flip_colors(&vscreen[(scanline-1)*480],&VRAM[(scanline-1)*480],480);
			else
				__asm
				{
					pusha
					mov eax,[scanline]
					mov ebx,480
					dec eax
					mov edi,[vscreen]
					mul ebx
					mov esi,[VRAM]
					mov ecx,120
					add edi,eax
					add esi,eax
					rep movsd
					popa
				}
			break;
		case 5:
			if ((bg2x==0)&&(bg2y==0)&&(BG2PA==0)&&(BG2PB==0)&&(BG2PC==0))
			//if (true)
				__asm
				{
					pusha
					mov eax,[scanline]
					cmp eax,128
					jg __blank_line
					mov ebx,480
					dec eax
					mov edi,[vscreen]
					mov ecx,eax
					mov esi,[VRAM]
					mul ebx
					add edi,eax
					mov eax,320
					mul ecx
					movzx edx,byte ptr [esi+IOREGS_OFFSET]
					and edx,0x10			// ecx = x0000 (frame buffer selection)
					add esi,eax
					shl edx,9				// ecx = x0000000000000
					lea edx,[edx+edx*4]		// ecx = x0x0000000000000 (0 : A000)
					add esi,edx
					mov ecx,80
					rep movsd
					xor eax,eax
					mov ecx,40
					rep stosd
					jmp __mode_5_done
__blank_line:
					mov ebx,480
					dec eax
					mov edi,[vscreen]
					mul ebx
					add edi,eax
					mov ecx,120
					xor eax,eax
					rep stosd
__mode_5_done:
					popa
				}
			else
				__asm
				{
					pusha
					mov eax,[scanline]
					mov ebx,480
					dec eax
					mov edi,[vscreen]
					mov ecx,eax
					mov esi,[VRAM]
					mul ebx
					add edi,eax
					mov eax,160
					mul ecx
					movzx edx,byte ptr [esi+IOREGS_OFFSET]
					//mov dummy,eax			// dummy = scanline*160
					mov dummy,ecx
					and edx,0x10			// ecx = x0000 (frame buffer selection)
					shl edx,9				// ecx = x0000000000000
					lea edx,[edx+edx*4]		// ecx = x0x0000000000000 (0 : A000)
					push ebp
					add esi,edx
					mov ecx,160
					mov eax,bg2y
					//imul ecx
					mov ebp,eax				// ebp = bg2y*160
					mov eax,[BG2PC]
					//imul ecx
					mov ebx,eax				// ebx = dy*160
					mov edx,bg2x			// edx = bg2x
					mov ecx,0
	mode_5_draw:
					push esi
					mov eax,edx				// bg2x
					sar eax,8
					//add eax,ecx
					cmp eax,0
					js mode_5_blank			// u+x<0 ?
					cmp eax,160
					jge mode_5_blank		// u+x>=160 ?
					lea esi,[esi+eax*2]
					mov eax,ebp				// v
					sar eax,8				// shift out fractional bits
					//add eax,dummy			// add scanline*320
					js mode_5_blank			// v+y<0 ?
					//cmp eax,20480
					cmp eax,128
					jge mode_5_blank		// v+y>160*128
					//add esi,[div160 + eax*4]
					//add esi,[mul160 + eax*4]
					shl eax,6
					add esi,eax
					lea esi,[esi+eax*4]

					mov ax,[esi]
					mov [edi],ax
					add edx,[BG2PA]
					add ebp,ebx
					add edi,2
					inc ecx
					pop esi
					cmp ecx,240
					jne mode_5_draw
					jmp mode_5_done
	mode_5_blank:
					xor eax,eax
					mov [edi],ax
					add edx,[BG2PA]
					add ebp,ebx
					add edi,2
					inc ecx
					pop esi
					cmp ecx,240
					jne mode_5_draw
	mode_5_done:
					pop ebp
					popa
				}
			break;
		default:
			//__asm { pusha }
			break;
		}
	}
	//}

	draw_sprites();
}



D3DLOCKED_RECT d3dlr;
int pitch_int;
char *lpBits_int;

typedef struct
{
	float x,y,z;
	int c;
} tvtx;

tvtx testTri[3];

void finish_screen_d3d() {
	int *masks = &colormask[colormode][0];
	bool glRenderer = (useGL && glOK);
	float *v_t,*v_b;

	newframe = 0;
	blended = 0;

	if (++frames < (frameskip+1)) {
		zombie(TELL,GPU_FRAME_COMPLETED,0,0);
		return;
	}	
	frames = 0;

	vscreen3 = vscreen;

	if (((VRAM[IOREGS_OFFSET]&8)==0) || (!emulate3D))	{
		lpTexture->LockRect(0,&d3dlr,0,0);
		lpBits_int = (char*)d3dlr.pBits;
		pitch_int = d3dlr.Pitch;

		__asm {
			pusha
			mov edi,[lpBits_int]
			mov edx,384
			mov ebx,[pitch_int]
			mov esi,[vscreen3]
			sub ebx,512
__copy_tex_data:
			mov ecx,128
			rep movsd
			dec edx
			jnz __copy_tex_data
			popa
		}

		lpTexture->UnlockRect(0);
	}

	/* Do frame limitation */
	if (limitFps) {
		QueryPerformanceCounter(&perfCnt2);
		while (((perfCnt2.QuadPart-perfCnt1.QuadPart)<<16)/perfFreq.QuadPart < frameInterval)
		{ QueryPerformanceCounter(&perfCnt2); }
		perfCnt1 = perfCnt2;
	}


 	GetClientRect(hwnd,&rect);

	if ((VRAM[IOREGS_OFFSET + 0x304 + 1]&0x80)==0)	{
		lpVertexBufCurr = lpVertexBuf2;
	} else	{
		lpVertexBufCurr = lpVertexBuf1;
	}

	// Beginning of scene.
	if (lpD3DDevice->BeginScene() != D3D_OK) {
		MessageBox(NULL,"IDirect3DDevice8::BeginScene failed.\nTry using the OpenGL or GDI renderer instead.","Error",MB_ICONERROR);
		zombie(TELL,ZOMBIE_HALT,0,0);
		return;
	}

	// Set the stream source to our vertex buffer.
	if (lpD3DDevice->SetStreamSource(0, lpVertexBufCurr, sizeof(customVertex)) != D3D_OK) {
		MessageBox(NULL,"IDirect3DDevice8::SetStreamSource failed.\nTry using the OpenGL or GDI renderer instead.","Error",MB_ICONERROR);
		zombie(TELL,ZOMBIE_HALT,0,0);
		return;
	}


	// Draw 4 triangles (2 quads).
	if (lpD3DDevice->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 4) != D3D_OK)	{
		MessageBox(NULL,"IDirect3DDevice8::DrawPrimitive failed.\nTry using the OpenGL or GDI renderer instead.","Error",MB_ICONERROR);
		zombie(TELL,ZOMBIE_HALT,0,0);
		return;
	}

	// End of scene.
	lpD3DDevice->EndScene();

	// Swap the front and back buffers.
	if (lpD3DDevice->Present(NULL, NULL, NULL, NULL) != D3D_OK)	{
		wsprintf(buffer,"IDirect3DDevice9::Present failed\n");
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
		zombie(TELL,ZOMBIE_HALT,0,0);
	}

	//if (IOREGS[0]&8)
	//	glClear(GL_COLOR_BUFFER_BIT);
	VRAM[IOREGS_OFFSET + 0x600+3] &= 0xF7;

	zombie(TELL,GPU_FRAME_COMPLETED,0,0);
}


void finish_screen_gdi() {
	newframe = 0;
	blended = 0;

	if (++frames < (frameskip+1)) {
		// Tell the shell that another frame has been completed
		zombie(TELL,GPU_FRAME_COMPLETED,0,0);
		return;
	}
	frames = 0;

	vscreen3 = vscreen;

	// Do frame limitation
	if (limitFps) {
		QueryPerformanceCounter(&perfCnt2);
		if (((perfCnt2.QuadPart-perfCnt1.QuadPart)<<16)/perfFreq.QuadPart < 400)
			Sleep(10);
		QueryPerformanceCounter(&perfCnt2);
		while (((perfCnt2.QuadPart-perfCnt1.QuadPart)<<16)/perfFreq.QuadPart < frameInterval)
		{ QueryPerformanceCounter(&perfCnt2); }
		perfCnt1 = perfCnt2;
	}

	// Set up the BITMAPINFOHEADER if the GDI renderer is used
	memset(&bi,0,sizeof(BITMAPINFOHEADER));
 	bi.bih.biSize = sizeof(BITMAPINFOHEADER);
 	bi.bih.biWidth = width;
 	bi.bih.biHeight = -visibleLines;
 	bi.bih.biPlanes = 1;
 	bi.bih.biBitCount = 16;
  	bi.bih.biCompression = BI_BITFIELDS;

 	GetClientRect(hwnd,&rect);

	if (sizemult==1) {
		VRAM[IOREGS_OFFSET + 0x600+3] &= 0xF7;
		if (displayScreens == 3) {
			if ((VRAM[IOREGS_OFFSET + 0x304 + 1]&0x80)==0) {
				SetDIBitsToDevice(dc,0,192,256,192,0,0,0,192,vscreen3,(BITMAPINFO*)&bi,DIB_RGB_COLORS);
				SetDIBitsToDevice(dc,0,0,256,192,0,0,0,192,vscreen_s,(BITMAPINFO*)&bi,DIB_RGB_COLORS);
			} else {
				SetDIBitsToDevice(dc,0,0,256,192,0,0,0,192,vscreen3,(BITMAPINFO*)&bi,DIB_RGB_COLORS);
				SetDIBitsToDevice(dc,0,192,256,192,0,0,0,192,vscreen_s,(BITMAPINFO*)&bi,DIB_RGB_COLORS);
			}
		} else if (displayScreens == 1) {
			SetDIBitsToDevice(dc,0,0,256,192,0,0,0,192,vscreen3,(BITMAPINFO*)&bi,DIB_RGB_COLORS);
		} else if (displayScreens == 2) {
			SetDIBitsToDevice(dc,0,0,256,192,0,0,0,192,vscreen_s,(BITMAPINFO*)&bi,DIB_RGB_COLORS);
		}
	}
	zombie(TELL,GPU_FRAME_COMPLETED,0,0);
}



void finish_screen_ogl() {
	int *masks = &colormask[colormode][0];
	bool glRenderer = (useGL && glOK);
	float *v_t,*v_b;


	newframe = 0;
	blended = 0;

	if (++frames < (frameskip+1)) {
		// Tell the shell that another frame has been completed
		zombie(TELL,GPU_FRAME_COMPLETED,0,0);
		return;
	}
	frames = 0;

	vscreen3 = vscreen;

	if (vtx_begin) {
		zombie(TELL,GPU_FRAME_COMPLETED,0,0);

		return;		// To stop OpenGL from complaining about illegal ops between
					// glBegin and glEnd.
	}

	glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
	glDisable(GL_DEPTH_TEST);

	if (glRenderer)	{
		glBindTexture(GL_TEXTURE_2D,texID[0]);
		glTexSubImage2D(GL_TEXTURE_2D,0,0,0,256,384,GL_RGBA,glFormat,vscreen);
		if ( ((VRAM[IOREGS_OFFSET+2]&3)==1) &&
			 (((VRAM[IOREGS_OFFSET]&8)!=0) && (emulate3D))) {
			glBindTexture(GL_TEXTURE_2D,texID[1]);
		}
	}

	// Do frame limitation
	if (limitFps) {
		QueryPerformanceCounter(&perfCnt2);
		if (((perfCnt2.QuadPart-perfCnt1.QuadPart)<<16)/perfFreq.QuadPart < 400)
			Sleep(10);
		QueryPerformanceCounter(&perfCnt2);
		while (((perfCnt2.QuadPart-perfCnt1.QuadPart)<<16)/perfFreq.QuadPart < frameInterval)
		{ QueryPerformanceCounter(&perfCnt2); }
		perfCnt1 = perfCnt2;
	}

	GetClientRect(hwnd,&rect);
	glViewport(0,192,rect.right,192);

	// Check if glTexSubImage2D returned an error
	if ((glRenderer) && (((VRAM[IOREGS_OFFSET]&8)==0)||(!emulate3D)))
	{
		if (glGetError()!=GL_NO_ERROR)
		{
			MessageBox(hwnd,"glTexSubImage2D failed\n\nSelect another texture format or\nuse the D3D or GDI renderer.","GPU error",MB_OK | MB_ICONERROR);
			zombie(TELL,ZOMBIE_HALT,0,0);
			return;
		}
	}
	
	// Set up the BITMAPINFOHEADER if the GDI renderer is used
	memset(&bi,0,sizeof(BITMAPINFOHEADER));
 	bi.bih.biSize = sizeof(BITMAPINFOHEADER);
 	bi.bih.biWidth = width;
 	bi.bih.biHeight = -visibleLines;
 	bi.bih.biPlanes = 1;
 	bi.bih.biBitCount = 16;
  	bi.bih.biCompression = BI_BITFIELDS;


 	GetClientRect(hwnd,&rect);

	if ((VRAM[IOREGS_OFFSET + 0x304 + 1]&0x80)==0) {
		v_t = &vertexes_s[0][0];
		v_b = &vertexes[0][0];
	} else {
		v_t = &vertexes[0][0];
		v_b = &vertexes_s[0][0];
	}

	if (glRenderer)	{
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		glViewport(0, 0, rect.right, rect.bottom);
		glClear(GL_COLOR_BUFFER_BIT);

		if (((VRAM[IOREGS_OFFSET]&8)==0) ||
			(!emulate3D) ||
			((VRAM[IOREGS_OFFSET+2]&3)!=1)) {
			if (displayScreens == 3) {
				glEnable(GL_TEXTURE_2D);
				glBegin(GL_QUADS);
					glTexCoord2fv(texCoords[0]);
					glVertex2fv(&v_t[0]);
					glTexCoord2fv(texCoords[1]);
					glVertex2fv(&v_t[2]);
					glTexCoord2fv(texCoords[2]);
					glVertex2fv(&v_t[4]);
					glTexCoord2fv(texCoords[3]);
					glVertex2fv(&v_t[6]);
				glEnd();
				glDisable(GL_TEXTURE_2D);

				glEnable(GL_TEXTURE_2D);
				glBegin(GL_QUADS);
					glTexCoord2fv(texCoords_s[0]);
					glVertex2fv(&v_b[0]);
					glTexCoord2fv(texCoords_s[1]);
					glVertex2fv(&v_b[2]);
					glTexCoord2fv(texCoords_s[2]);
					glVertex2fv(&v_b[4]);
					glTexCoord2fv(texCoords_s[3]);
					glVertex2fv(&v_b[6]);
				glEnd();
				glDisable(GL_TEXTURE_2D);

			} else if (displayScreens == 1) {
				glEnable(GL_TEXTURE_2D);
				glBegin(GL_QUADS);
					glTexCoord2fv(texCoords[0]);
					glVertex2fv(&vertexes_single[0][0]);
					glTexCoord2fv(texCoords[1]);
					glVertex2fv(&vertexes_single[1][0]);
					glTexCoord2fv(texCoords[2]);
					glVertex2fv(&vertexes_single[2][0]);
					glTexCoord2fv(texCoords[3]);
					glVertex2fv(&vertexes_single[3][0]);
				glEnd();
				glDisable(GL_TEXTURE_2D);

			} else if (displayScreens == 2) {
				glEnable(GL_TEXTURE_2D);
				glBegin(GL_QUADS);
					glTexCoord2fv(texCoords_s[0]);
					glVertex2fv(&vertexes_single[0][0]);
					glTexCoord2fv(texCoords_s[1]);
					glVertex2fv(&vertexes_single[1][0]);
					glTexCoord2fv(texCoords_s[2]);
					glVertex2fv(&vertexes_single[2][0]);
					glTexCoord2fv(texCoords_s[3]);
					glVertex2fv(&vertexes_single[3][0]);
				glEnd();
				glDisable(GL_TEXTURE_2D);
			}

			if (SwapBuffers(dc)==FALSE)	{
				MessageBox(hwnd,"SwapBuffers failed","GPU error",MB_OK | MB_ICONERROR);
				frameskip = 1000000;
				return;
			}

		} else {
			glColor4f(1.0,1.0,1.0,1.0);
			if (displayScreens == 3) {
				glEnable(GL_TEXTURE_2D);
				glBegin(GL_QUADS);
					glTexCoord2fv(texCoords3D[0]);
					glVertex2fv(&v_t[0]);
					glTexCoord2fv(texCoords3D[1]);
					glVertex2fv(&v_t[2]);
					glTexCoord2fv(texCoords3D[2]);
					glVertex2fv(&v_t[4]);
					glTexCoord2fv(texCoords3D[3]);
					glVertex2fv(&v_t[6]);
				glEnd();
				glDisable(GL_TEXTURE_2D);

				glBindTexture(GL_TEXTURE_2D,texID[0]);
				glEnable(GL_TEXTURE_2D);
				glBegin(GL_QUADS);
					glTexCoord2fv(texCoords_s[0]);
					glVertex2fv(&v_b[0]);
					glTexCoord2fv(texCoords_s[1]);
					glVertex2fv(&v_b[2]);
					glTexCoord2fv(texCoords_s[2]);
					glVertex2fv(&v_b[4]);
					glTexCoord2fv(texCoords_s[3]);
					glVertex2fv(&v_b[6]);
				glEnd();
				glDisable(GL_TEXTURE_2D);
			} else if (displayScreens == 1) {
				glEnable(GL_TEXTURE_2D);
				glBegin(GL_QUADS);
					glTexCoord2fv(texCoords3D[0]);
					glVertex2fv(&vertexes_single[0][0]);
					glTexCoord2fv(texCoords3D[1]);
					glVertex2fv(&vertexes_single[1][0]);
					glTexCoord2fv(texCoords3D[2]);
					glVertex2fv(&vertexes_single[2][0]);
					glTexCoord2fv(texCoords3D[3]);
					glVertex2fv(&vertexes_single[3][0]);
				glEnd();
				glDisable(GL_TEXTURE_2D);
			} else if (displayScreens == 2) {
				glBindTexture(GL_TEXTURE_2D,texID[0]);
				glEnable(GL_TEXTURE_2D);
				glBegin(GL_QUADS);
					glTexCoord2fv(texCoords_s[0]);
					glVertex2fv(&vertexes_single[0][0]);
					glTexCoord2fv(texCoords_s[1]);
					glVertex2fv(&vertexes_single[1][0]);
					glTexCoord2fv(texCoords_s[2]);
					glVertex2fv(&vertexes_single[2][0]);
					glTexCoord2fv(texCoords_s[3]);
					glVertex2fv(&vertexes_single[3][0]);
				glEnd();
				glDisable(GL_TEXTURE_2D);
			}

			if (SwapBuffers(dc)==FALSE)	{
				MessageBox(hwnd,"SwapBuffers failed","GPU error",MB_OK | MB_ICONERROR);
				frameskip = 1000000;
				return;
			}
			glPolygonMode(GL_FRONT_AND_BACK,polymode);
			glDisable(GL_DEPTH_TEST); //Enable(GL_DEPTH_TEST);
			glDepthFunc(GL_LESS);
		}


		if ((IOREGS[0]&8) && (emulate3D))
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		VRAM[IOREGS_OFFSET + 0x603] &= 0xF7;
		glViewport(dsVPX1,dsVPY1,dsVPX2,dsVPY2);

	} else
	{
		if (sizemult==1)
		{
			VRAM[IOREGS_OFFSET + 0x603] &= 0xF7;
				/*	SwapBuffers(dc);
					glClearColor(0,0,0,0);
					glClear(GL_COLOR_BUFFER_BIT);*/
					//SwapBuffers(dc);
			//memset(vscreen_s,0xFFFFFFFF,256*48);
			if ((VRAM[IOREGS_OFFSET + 0x304 + 1]&0x80)==0)
			{
				SetDIBitsToDevice(dc_s,0,0,256,192,0,0,0,192,vscreen3,(BITMAPINFO*)&bi,DIB_RGB_COLORS);
				SetDIBitsToDevice(dc,0,0,256,192,0,0,0,192,vscreen_s,(BITMAPINFO*)&bi,DIB_RGB_COLORS);
			} else
			{
				SetDIBitsToDevice(dc,0,0,256,192,0,0,0,192,vscreen3,(BITMAPINFO*)&bi,DIB_RGB_COLORS);
				SetDIBitsToDevice(dc_s,0,0,256,192,0,0,0,192,vscreen_s,(BITMAPINFO*)&bi,DIB_RGB_COLORS);
			}
		} else
 			StretchDIBits(dc,0,0,rect.right,rect.bottom-20,
 						  0,0,width,visibleLines,vscreen3,(BITMAPINFO*)&bi,DIB_RGB_COLORS,SRCCOPY);

	}
	zombie(TELL,GPU_FRAME_COMPLETED,0,0);

}


void reset_screen() {
}



// Configuration dialog callback
BOOL CALLBACK DlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam) {
	static char stemp[64];
	int i;
	HWND hCmb,hChk,hColPC,hColGBA,hSize1,hSize2;

	hColPC = GetDlgItem(hdlg,IDC_COLPC);
	hColGBA = GetDlgItem(hdlg,IDC_COLGBA);
	hSize1 = GetDlgItem(hdlg,IDC_SIZE1);
	hSize2 = GetDlgItem(hdlg,IDC_SIZE2);

   switch(msg) {
		case WM_INITDIALOG:
			SetWindowLong(hdlg, DWL_USER, lParam);

			hCmb = GetDlgItem(hdlg,IDC_COMBO1);
			SendMessage(hCmb,CB_ADDSTRING,0,(LPARAM)"0");
			SendMessage(hCmb,CB_ADDSTRING,0,(LPARAM)"1");
			SendMessage(hCmb,CB_ADDSTRING,0,(LPARAM)"2");
			SendMessage(hCmb,CB_ADDSTRING,0,(LPARAM)"3");
			SendMessage(hCmb,CB_ADDSTRING,0,(LPARAM)"4");
			SendMessage(hCmb,CB_ADDSTRING,0,(LPARAM)"5");
			SendMessage(hCmb,CB_SETCURSEL,frameskip,0);

			hCmb = GetDlgItem(hdlg,IDC_COMBO2);
			SendMessage(hCmb,CB_ADDSTRING,0,(LPARAM)"GDI");
			SendMessage(hCmb,CB_ADDSTRING,0,(LPARAM)"OpenGL");
			SendMessage(hCmb,CB_SETCURSEL,(glOK&&useGL)?1:0,0);

			hCmb = GetDlgItem(hdlg,IDC_COMBO3);
			SendMessage(hCmb,CB_ADDSTRING,0,(LPARAM)"5551");
			SendMessage(hCmb,CB_ADDSTRING,0,(LPARAM)"1555_REV");
			SendMessage(hCmb,CB_SETCURSEL,(glFormat==GL_UNSIGNED_SHORT_5_5_5_1)?0:1,0);
			EnableWindow(hCmb,useGL);

			hCmb = GetDlgItem(hdlg,IDC_COMBO4);
			SendMessage(hCmb,CB_ADDSTRING,0,(LPARAM)"NEAREST");
			SendMessage(hCmb,CB_ADDSTRING,0,(LPARAM)"LINEAR");
			SendMessage(hCmb,CB_SETCURSEL,(filter==GL_LINEAR)?1:0,0);
			EnableWindow(hCmb,useGL);

			hChk = GetDlgItem(hdlg,IDC_CHECK1);
			if (limitFps)
				SendMessage(hChk, BM_SETCHECK, BST_CHECKED, 0);
			else
				SendMessage(hChk, BM_SETCHECK, BST_UNCHECKED, 0);

			hChk = GetDlgItem(hdlg,IDC_OBJ);
			if (enableObj)
				SendMessage(hChk, BM_SETCHECK, BST_CHECKED, 0);
			else
				SendMessage(hChk, BM_SETCHECK, BST_UNCHECKED, 0);

			for (i=0; i<4; i++)
			{
				hChk = GetDlgItem(hdlg,IDC_BG3+i);
				SendMessage(hChk,BM_SETCHECK,enableLayer[3-i]?BST_CHECKED:BST_UNCHECKED,0);
				hChk = GetDlgItem(hdlg,IDC_BG3F+i);
				SendMessage(hChk,BM_SETCHECK,forceLayer[3-i]?BST_CHECKED:BST_UNCHECKED,0);
			}

			SendMessage(hColPC, BM_SETCHECK, 1-colormode, 0);
			SendMessage(hColGBA, BM_SETCHECK, colormode, 0);

			SendMessage(hSize1, BM_SETCHECK, (sizemult==1)?1:0, 0);
			SendMessage(hSize2, BM_SETCHECK, (sizemult==2)?1:0, 0);


			return TRUE;

		case WM_COMMAND:
			switch(LOWORD(wParam)) {
				case IDOK:
					hCmb = GetDlgItem(hdlg,IDC_COMBO1);
					frameskip = SendMessage(hCmb,CB_GETCURSEL,0,0);

					hCmb = GetDlgItem(hdlg,IDC_COMBO4);
					filter = (SendMessage(hCmb,CB_GETCURSEL,0,0)==0)?GL_NEAREST:GL_LINEAR;
					if (glIsInit)
					{
 						glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter);
  						glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter);
					}

					hCmb = GetDlgItem(hdlg,IDC_COMBO2);
					useGL = (SendMessage(hCmb,CB_GETCURSEL,0,0)==1);
					if (useGL && !glIsInit)
						glOK = setup_gl(dc);

					hCmb = GetDlgItem(hdlg,IDC_COMBO3);
					glFormat = (SendMessage(hCmb,CB_GETCURSEL,0,0)==0)?GL_UNSIGNED_SHORT_5_5_5_1:GL_UNSIGNED_SHORT_1_5_5_5_REV;

					hChk = GetDlgItem(hdlg,IDC_CHECK1);
					limitFps = (SendMessage(hChk,BM_GETCHECK,0,0)==BST_CHECKED);

					hChk = GetDlgItem(hdlg,IDC_OBJ);
					enableObj = (SendMessage(hChk,BM_GETCHECK,0,0)==BST_CHECKED);

					for (i=0; i<4; i++)	{
						hChk = GetDlgItem(hdlg,IDC_BG3+i);
						enableLayer[0][3-i] = (SendMessage(hChk,BM_GETCHECK,0,0)==BST_CHECKED);
						hChk = GetDlgItem(hdlg,IDC_BG3F+i);
						//forceLayer[3-i] = (SendMessage(hChk,BM_GETCHECK,0,0)==BST_CHECKED);
					}

					if (SendMessage(hColPC,BM_GETCHECK,0,0)==BST_CHECKED)
						colormode = 0;
					else
						colormode = 1;

					if (SendMessage(hSize1,BM_GETCHECK,0,0)==BST_CHECKED) {
						sizemult = 1;
						vertexes[2][1] = vertexes[3][1] = -0.78f;
					} else {
						sizemult = 2;
						vertexes[2][1] = vertexes[3][1] = -0.9f;
					}
					zombie(TELL,GPU_SIZE_MULTIPLIER,(void*)sizemult,0);

					EndDialog(hdlg, 0);
					return TRUE;

				case IDCANCEL:
					EndDialog(hdlg, 1);
					return FALSE;

				case IDC_COMBO2:
					if (HIWORD(wParam)==CBN_SELCHANGE) {
						if (SendMessage((HWND)lParam,CB_GETCURSEL,0,0)==1) {
							hCmb = GetDlgItem(hdlg,IDC_COMBO3);
							EnableWindow(hCmb,1);
							hCmb = GetDlgItem(hdlg,IDC_COMBO4);
							EnableWindow(hCmb,1);
						} else {
							hCmb = GetDlgItem(hdlg,IDC_COMBO3);
							EnableWindow(hCmb,0);
							hCmb = GetDlgItem(hdlg,IDC_COMBO4);
							EnableWindow(hCmb,0);
						}
					}
					return TRUE;

				default:
					break;
			}
			break;
   }

	return FALSE;
}
